/*
 * Copyright (c) 2014 Jochem Bakker, Grass Valley
 *
 * This file is part of Libav.
 *
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
/*
 * This file was modified by Grass Valley (2013/2014).
 * See libav-gv-changes.txt file for modification history.
 */


#include "avcodec.h"
#include "bytestream.h"
#include "mpegvideo.h"
#include "mpeguserdata.h"
#include "data_auxiliary.h"
#include "mpegvideo.h"

#define CHECK_USERID(a, b, c, d, method) \
if (buf_ptr[0] == a && buf_ptr[1] == b && buf_ptr[2] == c && buf_ptr[3] == d) method( avctx, dataAuxPt, buf_ptr, buf_end );


/*
 * Add an afd packet to the auxiliary data stream
 */
static void mpeg_userdata_add_afd_data( AVCodecContext *avctx,  int afdCode, DataAuxTp* dataAuxPt)
{
	int idx = dataAuxPt->entryCt;
	int offset = 0;

	// To support 8 (bit) digit afd encoding, check if the afdcode is already in 8 bit code:
	if ( afdCode > 0x0F) {
		// Afd code was alreayd in 8 bits..
		afdCode = afdCode & 0x7C; // Mask all reserved bits that should be zero
	}
	else
	{
		// Encode the AFD code for transport as ancillary data according toe SMPTE 2016-1:2009 table 4
		afdCode = (afdCode << 3) & 0x78;

		// Add aspect ratio bit (16:9 or 4:3) on bit 2...
		if ((int)(av_q2d(avctx->sample_aspect_ratio) * avctx->width / avctx->height * 10) >= 17) { /* 16:9 */
			afdCode |= 0x04;
		}
	}

	/*
	 * Add the AFD information to an auxiliary data package
	 */
	if ( dataAuxPt->entryCt < DATA_AUX_MAX_ENTRIES ) {

		// Check if the AFD is already present in the aux data list
		for( int n = 0; n < dataAuxPt->entryCt; n++ ) {
			if ( dataAuxPt->entryVc[n].type == DataAuxIdAfdEn ) {
				idx = n; // Simply overwrite any existing data
				break;
			}
		}


		// Determine the offset of the auxiliary data
		if ( idx > 0 )
			offset = dataAuxPt->entryVc[idx - 1].offset + dataAuxPt->entryVc[idx - 1].length;

		// Add the new entry & data
		dataAuxPt->entryVc[idx].type = DataAuxIdAfdEn;
		dataAuxPt->entryVc[idx].offset = offset;
		dataAuxPt->entryVc[idx].length = 1; // we only have a single nibble :(
		dataAuxPt->dataVc[ offset ] = afdCode; // Store the AFD data
		dataAuxPt->entryCt++;
	}
}


/*
 * Decode the AFD information
 */
static void mpeg_userdata_decode_dtg( AVCodecContext *avctx,  DataAuxTp* dataAuxPt, const uint8_t* buf_ptr,  const uint8_t* buf_end)
{
	int afdCode = 0;
	int flags = buf_ptr[4];

	buf_ptr += 5;
	if ( flags & 0x80 ) {
		/* skip event id */
		buf_ptr += 2;
	}
	if ( flags & 0x40 ) {
		if ( buf_end - buf_ptr < 1 )
			return;

		afdCode = buf_ptr[0] & 0x0f; // Get the AFD data

		/*
		 * Add the AFD information to an auxiliary data package
		 */
		mpeg_userdata_add_afd_data( avctx, afdCode, dataAuxPt );
	}
}


/*
 * Decode the mpeg user data.
 * - This method expects the entire user_data to be present in the supplied packet.
 * 	 User data that is divided over several avpackets is not supported by this method.
 */
static int mpeg_userdata_decoder_process(AVCodecContext *avctx, void *outdata, int *got_data, AVPacket *avpkt)
{
	DataAuxTp* 			dataAuxPt = 0;
	uint8_t*			buf_in = avpkt->data;
	int 				buf_in_size = avpkt->size;
	const uint8_t*		buf_ptr = buf_in;
    const uint8_t*		buf_end = buf_in + buf_in_size;
	uint32_t 			start_code = -1;
	//MpegUserDataContext *mpegUDCtx = (MpegUserDataContext*) avctx->priv_data;


	/* -------> Sanity check outgoing data address. */
	if ( outdata == 0 ) return 0;


	/* -------> Check integrity incoming data. */
	if ( buf_in == 0 ) return 0;
	if ( buf_in_size < 6 ) return 0;

	/* -------> No full packet yet... */
	if ( got_data )
		*got_data = 0;

	/* -------> Reset output entry counter. */
	dataAuxPt 			= (DataAuxTp*)outdata;
	dataAuxPt->entryCt 	= 0; /* -------> WATCH OUT, this might not be correct for teletext, 2 fields contain 1 packet! */


	/* -------> New packet or append to already received data */
	/* -------> Find user data start code. */
	for(;;) {
		/* find next start code */
		buf_ptr = avpriv_mpv_find_start_code( buf_ptr, buf_end, &start_code );
		if ( start_code == USER_START_CODE ){
			break;
		}
		if ( buf_ptr >= buf_end ){
			break; /* -------> Start code not found, simply ignore this packet */
		}
	}

	if ( start_code == USER_START_CODE ) {
		/* YES, user code found, parse the user code */
		/* we parse the DTG active format information */
		if ( buf_end - buf_ptr >= 5 ) // Check for at least the user ID
		{
			CHECK_USERID('D', 'T', 'G', '1', mpeg_userdata_decode_dtg);
			// <Add other services here>

			/* -------> TODO Integrate the VBI.c decoder here since they both do the same. Both decoders
			 			expect user data with a specific ID. When integrating the vbi.c code, vbi data
			 			inside mxf headers will work as well. */
		}
	}

	/*
	 * AFD information can also be found in a MXF metadata, when present, the dtg_active_format variable has been set.
	 * When this var is set, generate the data aux packets.
	 * Note that the mpeg_userdata_add_afd_data function will overwrite the afd data if it was already present
	 */
	if ( avctx->dtg_active_format != 0 ) {
		mpeg_userdata_add_afd_data( avctx, avctx->dtg_active_format, dataAuxPt );
	}

	/* When we have data present in our data aux structure, notify our caller. */
	if ( dataAuxPt->entryCt > 0 && got_data )
		*got_data = 1;

	return buf_in_size; // Indicate that we parsed the entire buffer
}



/*
 * Initialize this decoder
 */
static int mpeg_userdata_decoder_init( AVCodecContext *avctx )
{
	// MpegUserDataContext *mpegUDCtx = (MpegUserDataContext*) avctx->priv_data;
	// TODO Store something useful in the mpegUDCtx
	return 0;
}


/*******************************************************************************************/	
/*******************************************************************************************/	
/*******************************************************************************************/	
/*******************************************************************************************/	
AVCodec ff_mpeg_userdata_decoder = {
    .name= "MPEG User data",
    .type = AVMEDIA_TYPE_DATA,
    .id = AV_CODEC_ID_MPEG_USERDATA,
    .priv_data_size = sizeof(MpegUserDataContext),
    .init = mpeg_userdata_decoder_init,
    .decode = mpeg_userdata_decoder_process,
    .long_name = NULL_IF_CONFIG_SMALL("Grass Valley : MPEG2 header userdata decoder - DataAuxTp compatible"),
};


