/*
 * MXF Ancillary essence data_int8 decoder, decodes to DataAuxTp struct (see data_auxiliary.h).
 * SMPTE 436m-2006 - 6.1 ANC frame element
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
/*
 * This file was modified by Grass Valley (2013/2014).
 * See libav-gv-changes.txt file for modification history.
 */

#include "avcodec.h"
#include "libavutil/internal.h"
#include "data_auxiliary.h"
#include <string.h>


//#define DEBUG_MXF_ANC_ESSENCE_DECODER


static DataAuxIdEnTp getDataAuxIdEn( int16_t did, int16_t sdid )
{
	if ( did == 0x061 && sdid == 0x001 ) { return DataAuxIdCdp708En; }
	if ( did == 0x041 && sdid == 0x005 ) { return DataAuxIdAfdEn; }
	if ( did == 0x043 && sdid == 0x002 ) { return DataAuxIdOP47En; }
	//if ( did == 0x0FF && sdid == 0x0FF ) { return DataAuxIdVbiTxtEn; }
	return DataAuxIdUnknownEn;
}


/**
 Decodes an MXF ANC frame element (SMPTE 436m-2006 - 6.1) to the DataAuxTp struct (see data_auxiliary.h)

 @param		got_data[out] 		zero if the vbi aux data packet is incomplete, 1 if complete

 @return	bytes read from demuxed packet
*/
static int mxfanc_decode(AVCodecContext *avctx, void *outdata, int *got_data, AVPacket *avpkt)
                              
{
	DataAuxTp* dataAuxPt = 0;
	int32_t ancPacketsCt = 0;
	uint32_t lineNr = 0;
	int32_t dataIx = 0;
	int32_t outputOffset = 0;
	int32_t ancPacketIx = 0;
	uint32_t wrapTypeId = 0;
	uint32_t payloadCodingId = 0;
	uint32_t payloadSz = 0;
	int16_t did  = 0;
	int16_t sdid = 0;
	int16_t size = 0;
	DataAuxIdEnTp dataAuxIdEn = 0;
	uint16_t outputPacketIx = 0;
	int32_t alignSize = 0;
	int32_t paddingSz = 0;
	//const AVClass *av_class = avctx->av_class;
	uint8_t *dataInPt = avpkt->data;
	int dataInSize = avpkt->size;
	OP47PacketTp *op47DataPacketPt = 0;

	/* -------> Sanity check outgoing data address. */
	dataAuxPt = (DataAuxTp*) outdata;
	if ( dataAuxPt == NULL ) { return 0; }

	/* -------> No full vbi packet yet... */
	if (got_data)
		*got_data = 0;

	/* -------> Reset output entry counter. */
	dataAuxPt->entryCt 	= 0;

	/* -------> Sanity check incoming data size (demuxed packet size). */
	if ( dataInSize < 1 ) { return 0; }

	/* -------> Get number of ANC packets in this demuxed packet. */
	ancPacketsCt = ((int32_t)(dataInPt[0] << 8)) | dataInPt[1];
	dataIx = 2;

#ifdef DEBUG_MXF_ANC_ESSENCE_DECODER
	av_log(NULL, AV_LOG_ERROR, "mxfanc_decode(): ancPacketsCt:%d\n", ancPacketsCt);
#endif

	for ( ancPacketIx = 0; ancPacketIx < ancPacketsCt; ++ancPacketIx ) {

		lineNr = ((int32_t)(dataInPt[dataIx] << 8)) | dataInPt[dataIx + 1];
		dataIx += 2;

		wrapTypeId = dataInPt[dataIx++];
			/* -------> Wrapping type of the ANC Packet payload (SMPTE 436m-2006 : Table 7):
			 0x001		VANC Frame (Interlaced or segmented progressive frame)
			 0x002		VANC Field 1
			 0x003		VANC Field 2
			 0x004		VANC Progressive Frame
			 0x011		HANC Frame (Interlaced or segmented progressive frame)
			 0x012		HANC Field 1
			 0x013		HANC Field 2
			 0x014		HANC Progressive Frame
			*/
		payloadCodingId = dataInPt[dataIx++];
			/* -------> An enumerated Sample Coding:
			 4		8-bit  luma samples
			 5		8-bit  color difference samples
			 6		8-bit  luma and color difference samples
			 7		10-bit luma samples
			 8		10-bit color difference samples
			 9		10-bit luma and color difference samples
			 10		8-bit  luma samples, with parity
			 11		8-bit  color difference samples, with parity
			 12		8-bit  luma and color difference samples, with parity
			*/
		payloadSz = ((int32_t)(dataInPt[dataIx] << 8)) | dataInPt[dataIx + 1];
		//av_log(NULL, AV_LOG_ERROR, "mxfanc_decode(): PAYLOAD SIZE[%d] %x [%x] - %x\n",ancPacketIx,dataInPt[dataIx],(dataInPt[dataIx] << 8) , dataInPt[dataIx+1]);
		dataIx += 2;

		/* -------> Skip the 8 padding bytes always present before payload
					(officially this is the payload already, but this is not given in the given payload size (payloadSz)). */
		dataIx += 8;

		/* -------> Start of payload. */

		/* -------> Get DID/SDID & size. */
		did  = (int16_t)dataInPt[dataIx + 0];
		sdid = (int16_t)dataInPt[dataIx + 1];
		size = (int16_t)dataInPt[dataIx + 2];
		dataIx += 3;
		dataAuxIdEn = getDataAuxIdEn( did, sdid );


#ifdef DEBUG_MXF_ANC_ESSENCE_DECODER
	av_log(NULL, AV_LOG_ERROR, "mxfanc_decode(): dataAuxIdEn:%d (unknown=%d) [DID:%x, SDID:%x]\n", dataAuxIdEn, DataAuxIdUnknownEn, did, sdid);
#endif

		if ( dataAuxIdEn == DataAuxIdOP47En ) {

			/* -------> Extract teletext packets from OP47 packet. Multiple packets possible (max 5). */
			op47DataPacketPt = (OP47PacketTp*) &dataInPt[dataIx];

			for ( int8_t Ix = 0; Ix < 5; Ix++ ) {
				
				if ( op47DataPacketPt->vbiDesc[Ix] != 0 ) {
					
					/* -------> Sanity check output struct header overflow. */
					if ( dataAuxPt->entryCt < DATA_AUX_MAX_ENTRIES ) {

						outputOffset = dataAuxPt->entryCt * 47;
						dataAuxPt->entryVc[dataAuxPt->entryCt].type = DataAuxIdVbiTxtEn;
						dataAuxPt->entryVc[dataAuxPt->entryCt].length = 47; // Field,LineNm, 45 bytes data
						dataAuxPt->entryVc[dataAuxPt->entryCt].offset = outputOffset;

						/* -------> Put teletext data belonging information in the Data out container */
						dataAuxPt->dataVc[outputOffset]	=  (op47DataPacketPt->vbiDesc[Ix] >> 7 ) & 0x01; //Field Odd or Even
						dataAuxPt->dataVc[outputOffset+1] = ( op47DataPacketPt->vbiDesc[Ix] & 0x1F); // Vbi lineNumber
						memcpy(( &dataAuxPt->dataVc[outputOffset + 2]), &op47DataPacketPt->vibPacketVc[Ix][0], 45);

						dataAuxPt->entryCt++;
					}
					else {
						av_log(NULL, AV_LOG_WARNING, "mxf_anc_essence_decoder: dataEntryCt:%d, output struct header overflow\n", dataAuxPt->entryCt );
					}
				}
			}

		}

		else if ( ( dataAuxIdEn == DataAuxIdCdp708En ) || ( dataAuxIdEn == DataAuxIdAfdEn ) ) {

			outputPacketIx = dataAuxPt->entryCt;

			/* -------> Sanity check output struct header overflow. */
			if ( dataAuxPt->entryCt < DATA_AUX_MAX_ENTRIES ) {

				/* -------> Write to output data buffer. */
				dataAuxPt->entryCt++;
				dataAuxPt->entryVc[outputPacketIx].type	  = dataAuxIdEn;
				dataAuxPt->entryVc[outputPacketIx].offset = outputOffset;
				dataAuxPt->entryVc[outputPacketIx].length = size;

				/* Sanity check void* alignment. */
				if ( outputOffset % sizeof(void*) != 0 ) {
					av_log(NULL, AV_LOG_ERROR, "mxf_anc_essence_decoder: outputPacketIx:%d, outputOffset not aligned to void*\n", outputPacketIx);
					return 0;
				}

				/* -------> Sanity check buffer overflow. */
				if ( (size + outputOffset) > sizeof(dataAuxPt->dataVc) ) {
					av_log(NULL, AV_LOG_ERROR, "mxf_anc_essence_decoder: outputPacketIx:%d, data write will cause overflow in output struct data\n", outputPacketIx);
					return 0;
				}

				/* -------> Copy payload, and optionally memset the alignment padding bytes. */
				memcpy( &(dataAuxPt->dataVc[outputOffset]), &(dataInPt[dataIx]), (size_t) size );

				/* -------> Align size to void*. */
				alignSize = 0;
				if ( size % sizeof(void*) != 0 ) {
					alignSize = sizeof(void*) - (size % sizeof(void*));
				}

				/* -------> Increment the output offset to next entry payload (void* aligned). */
				outputOffset += dataAuxPt->entryVc[outputPacketIx].length + alignSize;
			}
			else {
				av_log(NULL, AV_LOG_WARNING, "mxf_anc_essence_decoder: dataEntryCt:%d, output struct header overflow\n", dataAuxPt->entryCt );
			}
		}
		else {
			/* -------> Not supported anc received. */
			
		}

		/* -------> Calculate MXF padding bytes for 32 bits alignment of payload. */
		if ( payloadSz % 4 ) {

			paddingSz = 4 - (payloadSz % 4);
		}
		else {

			/* -------> Payload is aligned, so no need for padding. */
			paddingSz = 0;
		}


		/* -------> Goto next ANC packet by skipping the payload, minus the IDs and size. */
		dataIx += payloadSz + paddingSz - 3; // The minus 3 -> DID SDID DC. // + paddingSz - 3;
	}


	/* Sanity check full input packet handled. */
	if ( dataIx > dataInSize ) {
#ifdef DEBUG_MXF_ANC_ESSENCE_DECODER
		av_log(NULL, AV_LOG_ERROR, "mxf_anc_essence_decoder: dataIx [%d] != dataInSize [%d], full input packet not handled?,size %d, payl %d, padding %d\n", dataIx, dataInSize, size ,payloadSz,paddingSz);
#endif
		//return 0;
	}

	/* -------> Set size of total structure. */
	if ( dataAuxPt->entryCt > 0 ) {
		*got_data = 1; // We have got data!
	}

#ifdef DEBUG_MXF_ANC_ESSENCE_DECODER
	av_log(NULL, AV_LOG_ERROR, "mxfanc_decode(): normal return\n", dataAuxIdEn, DataAuxIdUnknownEn, did, sdid);
#endif

	return dataInSize;
}


AVCodec ff_mxfanc_decoder = {
    .name            = "Auxiliary - DataAuxTp compatible",
    .type            = AVMEDIA_TYPE_DATA,
    .id              = AV_CODEC_ID_MXFANC,
    .priv_data_size  = sizeof(DataAuxTp),
    .decode          = mxfanc_decode,
    .capabilities    = CODEC_CAP_DR1,
    .long_name       = NULL_IF_CONFIG_SMALL("Grass Valley: Auxiliary - DataAuxTp compatible"),
};


