/*
 *
 * This file is part of Libav.
 *
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
/*
 * This file was created by Grass Valley (2013/2014).
 * Created by Harold (2010-11-27)
 */

#include "avcodec.h"
#include "bytestream.h"
#include "mpegvideo.h"
#include "vbi.h"
#include "data_auxiliary.h"

#define SIN_OFFSET		0 // was 7
#define PAL_PIX_WIDTH	720
#define RAW_LENGTH      384

typedef struct { // struct used in NosDecFFStremTp.cpp, ffmpeg:vbi.c
    	int filledBo;
    	int fieldNm; // 0..1
    	int lineNm;	 // 7.22
    	int sizeBt;	 // number of bytes in buffer, now 720
    	unsigned char yBf[PAL_PIX_WIDTH];
} STRUCT_VBI_LINE;


/* -------> Conversion tables used for inserting a sine pattern into the SDI pixel stream. */
static unsigned char SinePulse0[PAL_PIX_WIDTH];
static unsigned char SinePulse1[PAL_PIX_WIDTH];

int vbi_tt_decoder_init( AVCodecContext *avctx );
unsigned char reverse_byte( unsigned char val );
unsigned char deham8( unsigned char ham8);
void ShiftLeftByOne(uint8_t * arr, int len);




/*******************************************************************************************/	
/*** decoder LUT init **********************************************************************/	
/*******************************************************************************************/	
int vbi_tt_decoder_init( AVCodecContext *avctx )
{
	/* -------> Initilize the sine lookup tables. */
	float freq = 6.9375/2.0; // in MHz	  
	unsigned char val;
	
	int j = 0; 
	int ix;
	j -= SIN_OFFSET;

	for ( ix = 0; ix < PAL_PIX_WIDTH; ix++ ) {
		val = (unsigned char)(16.0 + 75.0 + 75.0*-sin((double)j*2.0*3.141592653*freq / 13.5));
		SinePulse0[ix] = val;

		val = (unsigned char)(16.0 + 75.0 + 75.0* + sin((double)j*2.0*3.141592653*freq / 13.5));
		SinePulse1[ix] = val;

		j++;
	}
	return 0;
}


#if 0
/*******************************************************************************************/	
/*** Decoder for EBU TT data in PES packet *************************************************/	
/*******************************************************************************************/	
static void convert_tt_to_yline( unsigned char *srcDataPt, unsigned char *linePt ) 
{
	int i;
	unsigned char *lut0, *lut1;
	unsigned int pos, bit, last_bit, ofs, cur;


	/*----------------------------------------------------------------------------------------*/
	/*                                       Blank line                                       */
	/*----------------------------------------------------------------------------------------*/
 
	/* -------> Null packets result in blank lines. */
	if ( srcDataPt==NULL ) {
		for ( i = 0; i < PAL_PIX_WIDTH; i++ ) {
			linePt[i] = 0x10;
		}
	}

	/*----------------------------------------------------------------------------------------*/
	/*                                       Filled line                                      */
	/*----------------------------------------------------------------------------------------*/

	else {

		/* -------> Get the TT packet data pointer. Note that this code may access the data
					vector at offsets 45 and 46, and expect a null byte there. */
		pos		 = 0;			   
		cur		 = 0x0;
		last_bit = 100;         
		
		for ( i = 0;i < SIN_OFFSET; i++ ) { 
			linePt[i] = 0x10;
		}
		
		for (;i < PAL_PIX_WIDTH; i++) {
			
			bit = pos / 144144; // divide by teletext period in ps (picoseconds)
			ofs = pos % 144144;
			
			if (last_bit!=bit) {
				cur  = (cur<<1)&0x6;
				cur |= (srcDataPt[bit >> 3] >> (bit & 7)) & 0x1;
				//cur |= (bit&1);
			}                    
			
			last_bit = bit;
			if (bit&1) {
				lut0 = SinePulse1;
				lut1 = SinePulse0;
			} else {
				lut0 = SinePulse0;
				lut1 = SinePulse1;
			}
			
			switch (cur) {
				case 0x0: // 000		
					linePt[i]=0x10;
					break;				
					
				case 0x1: // 001		
					if (ofs<72072) {
						linePt[i]=0x10;
					} else {
						linePt[i]=(unsigned int)lut0[i];
					}
					break;				
					
				case 0x2: // 010		
					linePt[i]=(unsigned int)lut1[i];
					break;				
					
				case 0x3: // 011		
					if (ofs<72072) {
						linePt[i]=(unsigned int)lut1[i];
					} else {
						linePt[i]=0xA6;
					}
					break;				
					
				case 0x4: // 100		
					if (ofs<72072) {
						linePt[i]=(unsigned int)lut0[i];
					} else {
						linePt[i]=0x10;
					}
					break;				
					
				case 0x5: // 101		
					linePt[i]=(unsigned int)lut0[i];
					break;				
					
				case 0x6: // 110		
					if (ofs<72072) {
						linePt[i]=0xA6;
					} else {
						linePt[i]=(unsigned int)lut1[i];
					}
					break;				
					
				case 0x7: // 111		
					linePt[i]=0xA6;
					break;				
			}
			pos += 74074; // add pixel in ps (picoseconds)
		}                                             
	}
}
#endif

unsigned char reverse_byte( unsigned char val ) {
	return (((val   )&1)<<7) |
	       (((val>>1)&1)<<6) |
	       (((val>>2)&1)<<5) |
	       (((val>>3)&1)<<4) |
	       (((val>>4)&1)<<3) |
	       (((val>>5)&1)<<2) |
	       (((val>>6)&1)<<1) |
	       (((val>>7)&1)   );
}


/*******************************************************************************************/	
/*******************************************************************************************/	
/*******************************************************************************************/	
unsigned char deham8( unsigned char ham8) {
	/* -------> NEEDS ATTENTION: Use a lookup table. */

	unsigned char res = 0;
	res += (ham8 >> 1) & 0x01;									// Bit 1 << 0 (bitrange = 0-7).
	res += (((ham8 >> 3) & 0x01) << 1);							// Bit 3 << 1.
	res += (((ham8 >> 5) & 0x01) << 2);							// Bit 5 << 2.
	res += (((ham8 >> 7) & 0x01) << 3);							// Bit 7 << 3.
	return res;
}

void ShiftLeftByOne(uint8_t * arr, int len) 
{     
	int i;     
	for (i = 0;  i < len - 1;  ++i)     
	{         
		arr[i] = (arr[i] << 1) | ((arr[i+1] >> 7) & 1);     
	}     
	arr[len-1] = arr[len-1] << 1; 
} 


/*******************************************************************************************/	
/*** Decoder for EBU TT data in PES packet *************************************************/	
/*******************************************************************************************/	
static int vbi_tt_decoder_process(AVCodecContext *avctx, void *outdata, int *outdata_size, AVPacket *avpkt)
{
	int Ct=0;
	int i;
	int data_identifier;
	int data_unit_id;
	int data_unit_length;
	int val;
	int field;
	int line;
	//int mag;
	//int pack;
	//int k1,k2,k;
	//int pageTens, pageUnits, pageNm;
	unsigned char ttDataVc[45];
	uint8_t *buf = avpkt->data;
	int buf_size = avpkt->size;
	DataAuxTp *dataAuxPt = (DataAuxTp *)outdata;
	
	// PDF: "ITU EBU teletext in mpeg stream.pdf" in SMPTE directory HV
	// EBU TT data (the data identifier is already read in the caller)
	
   	data_identifier  = buf[Ct++]; 


	/* -------> Reset output entry counter. */
	if ( outdata_size == 0 ){ 	dataAuxPt->entryCt = 0;}

   	while ( Ct < buf_size ) {
	   	data_unit_id     = buf[Ct++];
   		data_unit_length = buf[Ct++];
   		val              = buf[Ct++];
   		
   		field            = (((val>>5)&1)==1)?0:1;
   		line             = val & 0x1F;
   		
   		ttDataVc[0] = 0x55; // clock run-in
   		ttDataVc[1] = 0x55; // clock run-in
   		
   		ttDataVc[2] = reverse_byte( buf[Ct++] );  // framing code
		   		
   		ttDataVc[3] = reverse_byte( buf[Ct++] );  // mag
   		ttDataVc[4] = reverse_byte( buf[Ct++] );  // pack

 		// 40 data bytes
   		for ( i = 0; i < 320/8; i++ ) {
   			 ttDataVc[5+i] = reverse_byte( buf[Ct++] ); 
   		}

   		if ( ((data_unit_id==0x02) || (data_unit_id==0x03))  ){
			
			/* -------> Sanity check output struct header overflow. */
			if ( dataAuxPt->entryCt < DATA_AUX_MAX_ENTRIES ) {
				dataAuxPt->entryVc[dataAuxPt->entryCt].type = DataAuxIdVbiTxtEn;
				dataAuxPt->entryVc[dataAuxPt->entryCt].offset = dataAuxPt->entryCt * DATA_AUX_TXT_PACKET_SIZE;
				dataAuxPt->entryVc[dataAuxPt->entryCt].length = DATA_AUX_TXT_PACKET_SIZE;


				dataAuxPt->dataVc[dataAuxPt->entryCt * DATA_AUX_TXT_PACKET_SIZE] = field;
				dataAuxPt->dataVc[(dataAuxPt->entryCt * DATA_AUX_TXT_PACKET_SIZE) + 1] = line;

				memcpy(dataAuxPt->dataVc + dataAuxPt->entryVc[dataAuxPt->entryCt].offset + 2 , ttDataVc, DATA_AUX_TXT_PACKET_DATA_SIZE);

				dataAuxPt->entryCt++;
			}
			else {
				av_log(NULL, AV_LOG_WARNING, "vbi decoder: dataEntryCt:%d, output struct header overflow\n", dataAuxPt->entryCt );
			}

			*outdata_size = sizeof(DataAuxTp);
		}
		else if  ( (data_unit_id==0xFF) && (!dataAuxPt->entryCt) ) {
			
			/* -------> Sanity check output struct header overflow. */
			if ( dataAuxPt->entryCt < DATA_AUX_MAX_ENTRIES ) {
				/* -------> Add dummy data to keep stream alive when no real vbi data is available.
							in the service stream a dummy packet is added. */

				dataAuxPt->entryVc[dataAuxPt->entryCt].type = DataAuxIdVbiTxtEn;
				dataAuxPt->entryVc[dataAuxPt->entryCt].offset = dataAuxPt->entryCt * DATA_AUX_TXT_PACKET_SIZE;
				dataAuxPt->entryVc[dataAuxPt->entryCt].length = 0;


				dataAuxPt->dataVc[dataAuxPt->entryCt * DATA_AUX_TXT_PACKET_SIZE] = 0;
				dataAuxPt->dataVc[(dataAuxPt->entryCt * DATA_AUX_TXT_PACKET_SIZE) + 1] = 0;

				memset(dataAuxPt->dataVc + dataAuxPt->entryVc[dataAuxPt->entryCt].offset + 2 , 0x00, DATA_AUX_TXT_PACKET_DATA_SIZE);

				dataAuxPt->entryCt++;
			}
			else {
				av_log(NULL, AV_LOG_WARNING, "vbi decoder: dataEntryCt:%d, output struct header overflow\n", dataAuxPt->entryCt );
			}

			*outdata_size = sizeof(DataAuxTp);
		}
   	}

	return buf_size;

}


/*******************************************************************************************/	
/*** decoder type for RAW VBI lines in PES *************************************************/	
/*******************************************************************************************/	
static int vbi_decoder_line_process(AVCodecContext *avctx, void *outdata, int *outdata_size, AVPacket *avpkt)
{
    //VBIContext *s = avctx->priv_data;
    int i;
    int val;
    int Ct=0;
    int type;
	STRUCT_VBI_LINE *vbiPt = outdata;
	const uint8_t *buf = avpkt->data;
	int buf_size = avpkt->size;
	const uint8_t *debug_buf = buf; 
	
	*outdata_size = sizeof(STRUCT_VBI_LINE);


	av_log(avctx, AV_LOG_ERROR, "vbi.c : Decode Raw Teletext decoder called\n");
	
   // align to the end of the buffer
    for (i=0;i<buf_size-(PAL_PIX_WIDTH+2);i++) {
    	bytestream_get_byte(&buf);
    	Ct++;
    }
    
    val = bytestream_get_byte(&buf);
    Ct++;

    type = val>>4;
    vbiPt->fieldNm = ((val&0xF)-1);

    val = bytestream_get_byte(&buf);
    Ct++;

    vbiPt->lineNm  = val + 7;
    
    if (type!=0) {
    	av_log(avctx, AV_LOG_ERROR, "unexpected VBI type!=0 (type=%d, field=%d, line=%d, size=%d)\n", type, vbiPt->fieldNm, vbiPt->lineNm, buf_size-Ct );
	    for (i=0;i<buf_size-Ct;i++) {
    		av_log(avctx, AV_LOG_ERROR, "%02X ", bytestream_get_byte(&debug_buf) );
   	 	}	
    	av_log(avctx, AV_LOG_ERROR, "\n");
    	return buf_size;
    }
    if ((buf_size-Ct)!=PAL_PIX_WIDTH) {
    	av_log(avctx, AV_LOG_ERROR, "unexpected VBI size!=%d (type=%d, field=%d, line=%d, size=%d)\n", PAL_PIX_WIDTH, type, vbiPt->fieldNm, vbiPt->lineNm, buf_size-Ct );
	    for (i=0;i<buf_size-Ct;i++) {
    		av_log(avctx, AV_LOG_ERROR, "%02X ", bytestream_get_byte(&debug_buf) );
   	 	}	
    	av_log(avctx, AV_LOG_ERROR, "\n");
    	return buf_size;
    }
    //av_log(avctx, AV_LOG_ERROR, "VBI OK (type=%d, field=%d, line=%d, size=%d)\n", type, vbiPt->fieldNm, vbiPt->lineNm, buf_size-Ct );
    for (i=0;i<PAL_PIX_WIDTH;i++) {
	    vbiPt->yBf[i] = bytestream_get_byte(&buf);
    }
    vbiPt->filledBo = -1;
    vbiPt->sizeBt   = buf_size-Ct;



	/* -------> If we have the raw Y values then we can create a 45 bytes teletext packet. */




    return buf_size;
}


static void vbi_userdata_demultiplex_data( AVCodecContext *avctx, uint8_t lineCt, void *rawDataIn, int rawDataSz, DataAuxTp *dataAuxPt) 
{
	uint8_t rawBytesVc[388];
	uint8_t *rawBytesPt =  rawBytesVc;
	uint8_t extractedBytes[360]; // 9 blocks of 40 bytes -> OP47 = 328 bytes
	int  rawBytesLength = 384;
	uint16_t vbiLineNm =0;
	uint8_t *ccDataPt;
	uint16_t op47PacketRawVc[262];

	for (char Ix=0; Ix < lineCt; Ix++) {
			rawBytesLength = 384;
		/****
		 * Demultiplex one line
		 */

			memcpy(rawBytesVc, (uint8_t *) rawDataIn + (Ix * 5100), 388);

			/* -------> Copy first set of raw VBI data */
			vbiLineNm = (uint16_t)(((uint16_t)(rawBytesPt[0]) << 8) | rawBytesPt[1]);
			rawBytesPt += 4;

			/* Demultiplex data (remove the marker bit. |16 bits| marker | 16 bits | marker ... */
			for (uint16_t Iy = 0; Iy < 384; Iy+=2) {

				/* Copy 16 bits of data */
				extractedBytes[Iy]	 = rawBytesPt[0];
				extractedBytes[Iy+1] = rawBytesPt[1];

				/* decrease length by 2 bytes */
				rawBytesLength-=2;
				memmove(rawBytesPt, rawBytesPt+2, rawBytesLength);
		
				/* Shift array by 1 bit (remove marker) */
				ShiftLeftByOne(rawBytesPt, rawBytesLength);

			}

		/****
		 * Create OP47 SDP packet
		 */

			ccDataPt = extractedBytes;
				
			/* -------> Each raw YUV op47 packet is for now 360 bytes.
						The data is split up in blocks of 40 bytes.
						40 Data bytes == 32 MSB + 8 LSB. In 1 LSB there are 4 * 2  LS bits. 
						These 2 bits combined with one MSB gives 10 bit. */
										
			for (uint8_t blockNr=0; blockNr < 360/40; blockNr++) {
				uint16_t vbi10BitVc[32];
				uint8_t  rawLsbVc[8];
				uint8_t  rawMsbVc[32];
				uint8_t yuvIndex = 0;
	
				memcpy(rawMsbVc, ccDataPt, 32); 	 // Copy the most significant bytes
				memcpy(rawLsbVc, ccDataPt + 32, 8);  // Copy the least significant bytes
	
	
				for (uint8_t Ix=0; Ix < 32; Ix++) {
					vbi10BitVc[Ix] = rawMsbVc[Ix]; 
					vbi10BitVc[Ix] = (vbi10BitVc[Ix] << 2) | ((rawLsbVc[Ix/4] >> ((Ix % 4)*2)) & 0x3);
				}
	
				ccDataPt+=40; //Next block of 32 + 8 samples
	
				/* -------> Strip u/v component */
				
				for(uint8_t Ix = 0; Ix < 16; Ix++) {
					op47PacketRawVc[Ix + (blockNr*16)] = vbi10BitVc[yuvIndex];
					yuvIndex+=2;
				}
			}

			/****
			 * OP47 SDP to teletext packet
			 */

		
			/* -------> Check if our decoded data contains an OP47 SDP (Subtitle Distribution Packet) */
			if ((op47PacketRawVc[3] == 0x143) && (op47PacketRawVc[4] == 0x102)) {
				
				/* -------> At this stage we will have two OP47 packets odd and even field. 
							Both seem only to carry one txt packet. */
		
				uint8_t vbi8BitTxtPacketVc[DATA_AUX_TXT_PACKET_DATA_SIZE];
				uint8_t offset	= 0;

				/* -------> Sanity check output struct header overflow. */
				if ( dataAuxPt->entryCt < DATA_AUX_MAX_ENTRIES ) {

					offset = ( dataAuxPt->entryCt ) * DATA_AUX_TXT_PACKET_SIZE;
					dataAuxPt->entryVc[dataAuxPt->entryCt].type = DataAuxIdVbiTxtEn;
					dataAuxPt->entryVc[dataAuxPt->entryCt].length = DATA_AUX_TXT_PACKET_SIZE; // Field,LineNm, 45 bytes data
					dataAuxPt->entryVc[dataAuxPt->entryCt].offset = offset;
		
					/* -------> Strip parity and inverted parity. */
					for(uint8_t Ix=0; Ix < DATA_AUX_TXT_PACKET_SIZE; Ix++) {
						vbi8BitTxtPacketVc[Ix] = (uint8_t)(op47PacketRawVc[15+Ix] & 0xFF);
					}

					/* -------> Put teletext data belonging information in the Data out container */
					dataAuxPt->dataVc[offset]	= (op47PacketRawVc[10] & 0x80) ? 1 : 0; //Field Odd or Even
					dataAuxPt->dataVc[offset+1] = 21; // Vbi lineNumber
					memcpy(( &dataAuxPt->dataVc[offset + 2]), vbi8BitTxtPacketVc, DATA_AUX_TXT_PACKET_DATA_SIZE);

					dataAuxPt->entryCt++;
				}
				else {
					av_log(NULL, AV_LOG_WARNING, "vbi-decoder: dataEntryCt:%d, output struct header overflow\n", dataAuxPt->entryCt );
				}
			}
			else {
				//av_log(avctx, AV_LOG_ERROR, "### VBI.C: Received userdata, expected SCM2 captions packed in OP47 but got DID:%04x,SDID:%04x\n",
				//	op47PacketRawVc[3], op47PacketRawVc[4]);
			}


	}


}


//
// Decode the VBI userdata.
// Note that vbi data can be stored per field thus requiring multiple packets before a single vbi data packet is complete.
// - got_data is set to 1 when the vbi data is complete
static int vbi_userdata_decoder_process(AVCodecContext *avctx, void *outdata, int *got_data, AVPacket *avpkt)
{
	DataAuxTp* dataAuxPt = 0;
	uint8_t *buf_in = avpkt->data;
	int buf_in_size = avpkt->size;
	const uint8_t *buf_ptr = buf_in;
    const uint8_t *buf_end = buf_in + buf_in_size;
	uint8_t  lineCt = 0;
	uint8_t *rawOP47DataPt = 0;

	uint16_t rawOP47DataSz = 0;
	uint32_t start_code = -1;
	uint16_t expectedBufLength = 0;
	VBIContext *vbiCtxPt = (VBIContext*) avctx->priv_data;

	/* -------> Sanity check outgoing data address. */
	if ( outdata == 0 ) return 0;

	/* -------> Check integrity incoming data. */
	if ( buf_in == 0 ) return 0;
	if ( buf_in_size < 10 ) return 0;

	/* -------> Data is not yet complete. */
	if (got_data != NULL)
		*got_data = 0;

	/* -------> Reset output entry counter. */
	dataAuxPt 			= (DataAuxTp*)outdata;


	/* -------> New packet or append to allready received data */
	if ( vbiCtxPt->decoderStateEn == CREATE_NEW_PACKET ) {

		/* -------> New packet, restart the counter */
		dataAuxPt->entryCt 	= 0;

		/* -------> Find user data start code. */
		for(;;) {
			 /* find next start code */
			 buf_ptr = avpriv_mpv_find_start_code(buf_ptr,buf_end, &start_code);
			 if (start_code == 0x0001B2){
				break;
			 }
			 if (buf_ptr >= buf_end){
			 	/* -------> Start code not found */
				//av_log(avctx, AV_LOG_ERROR, "### VBI.C: User data start code not found\n");
				break;
				// FAIL, EXPECTED OR IGNORE ???? */
			 }
		}	

		if (start_code == 0x0001B2) {

/* -------> !!!! Note expected buffer length should come from video format. */
			
			lineCt = buf_ptr[5]; // Get number of lines present in user data
			//av_log(avctx, AV_LOG_ERROR, "### VBI.C: Width %d; Height %d, Lines : %d\n", avctx->width, avctx->height, lineCt);
			//expectedBufLength = lineCt * ((avctx->width == 1920) ? 5100 : 3400);
			expectedBufLength = lineCt *  5100 ;

			/* -------> See if we received enough information to process */
			if (expectedBufLength >  ( buf_end - ( buf_ptr + 4 ) ) ) {
				vbiCtxPt->decoderStateEn = WAIT_FOR_MORE_DATA;

				if(vbiCtxPt->internalDataPt == 0){
					vbiCtxPt->internalDataSz = buf_end - ( buf_ptr + 4 );
					vbiCtxPt->internalDataPt = av_malloc( vbiCtxPt->internalDataSz );
					if (vbiCtxPt->internalDataPt) memcpy( vbiCtxPt->internalDataPt, &buf_ptr[5], vbiCtxPt->internalDataSz );
    			}
				else {
					// Buffer in use uhoh troubles...
					av_log(avctx, AV_LOG_ERROR, "vbi.c : New packet : Internal data buffer Allready defined \n");
				}
			}
			else {
				rawOP47DataPt = av_malloc( buf_in_size );
				memcpy(rawOP47DataPt, buf_in, buf_in_size );
				vbiCtxPt->decoderStateEn = PROCESS_USERDATA;
			}
		}
	}
	else if ( vbiCtxPt->decoderStateEn == WAIT_FOR_MORE_DATA) {		

/* -------> WE SHOULD CHECK IF WE NEED MORE THAN 2 PACKETS!!! UNLIKELY */		

		if ( vbiCtxPt->internalDataPt ) {
			/* -------> Create storing space */
			rawOP47DataPt = av_malloc( vbiCtxPt->internalDataSz + buf_in_size);

			/* -------> Copy previous received data */
			memcpy(rawOP47DataPt, vbiCtxPt->internalDataPt, vbiCtxPt->internalDataSz );
			rawOP47DataSz += vbiCtxPt->internalDataSz;
			av_free( vbiCtxPt->internalDataPt );
			vbiCtxPt->internalDataPt = 0;
			vbiCtxPt->internalDataSz = 0;
			memcpy(rawOP47DataPt + rawOP47DataSz, buf_in, buf_in_size );
			vbiCtxPt->decoderStateEn = PROCESS_USERDATA;
			rawOP47DataSz += buf_in_size;
		}
		else {
			vbiCtxPt->decoderStateEn = CREATE_NEW_PACKET;
			return 0;
		}
	}


	/* -------> Process the userdata */
	if ( vbiCtxPt->decoderStateEn == PROCESS_USERDATA ) {

		/* -------> Get number of lines. */
		lineCt = *rawOP47DataPt;
		
		/* -------> Move to start of data. */
		vbi_userdata_demultiplex_data( avctx, lineCt, rawOP47DataPt+1, rawOP47DataSz-2, dataAuxPt);
		vbiCtxPt->decoderStateEn = CREATE_NEW_PACKET;
		
		av_free( rawOP47DataPt );
		rawOP47DataPt = 0;
		rawOP47DataSz = 0;

		/* -------> Got a full vbi packet, notify the caller ! */
		if (got_data != NULL)
			*got_data = 1;
	}

	return buf_in_size;
}


static int init_vbi_userdata_decoder( AVCodecContext *avctx ) 
{
	VBIContext *vbiCtxPt = (VBIContext*) avctx->priv_data;

	vbiCtxPt->decoderStateEn = CREATE_NEW_PACKET;
	vbiCtxPt->internalDataPt = 0;
	vbiCtxPt->internalDataSz = 0;

	return 0;
}


/*******************************************************************************************/	
/*******************************************************************************************/	
/*******************************************************************************************/	
AVCodec ff_vbi_tt_decoder = {
    .name 				= "vbi_tt",
    .type 				= AVMEDIA_TYPE_DATA,
    .id 				= AV_CODEC_ID_VBI_TT,
    .priv_data_size 	= sizeof(VBIContext),
    .init 				= vbi_tt_decoder_init, 
    .decode 			= vbi_tt_decoder_process,
    .long_name 			= NULL_IF_CONFIG_SMALL("Grass Valley : VBI_TT EBU teletext - DataAuxTp compatible"),
};


/*******************************************************************************************/	
/*******************************************************************************************/	
/*******************************************************************************************/	
/*******************************************************************************************/	
AVCodec ff_vbi_decoder = {
    .name = "vbi",
    .type = AVMEDIA_TYPE_DATA,
    .id = AV_CODEC_ID_VBI,
    .priv_data_size = sizeof(VBIContext),
    .decode = vbi_decoder_line_process, 
    .long_name = NULL_IF_CONFIG_SMALL("Grass Valley : VBI lines - DataAuxTp compatible"),
};

/*******************************************************************************************/	
/*******************************************************************************************/	
/*******************************************************************************************/	
/*******************************************************************************************/	
AVCodec ff_vbi_userdata_decoder = {
    .name= "vbi_userdata",
    .type = AVMEDIA_TYPE_DATA,
    .id = AV_CODEC_ID_VBI_USERDATA,
    .priv_data_size = sizeof(VBIContext),
    .init = init_vbi_userdata_decoder,
    .decode = vbi_userdata_decoder_process, 
    .long_name = NULL_IF_CONFIG_SMALL("Grass Valley : VBI lines within MPEG2 header userdata - DataAuxTp compatible"),
};


