/*
 * AVF Demuxer
 * Copyright (c) 2014 Grass Valley, Jochem Bakker <jochem.bakker@grassvalley.com>
 *
 * This file is part of Libav.
 *
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */
/*
 * This file was modified by Grass Valley (2013/2014).
 * See libav-gv-changes.txt file for modification history.
 */

#include <limits.h>
#include "libavutil/opt.h"
#include "avformat.h"
#include "internal.h"
#include "isom.h"
#include "libavutil/intreadwrite.h"

//#include "mov.c"



#undef NDEBUG
#include <assert.h>

//
// Definition of the AVF header
#define AVF_MAGIC			"PT AVF"  // Shouldn't this be GV AVF in the future :) ???
#define MAX_ESSENCES		32
#define MAX_PATH			128


//
// Definition of a single file in the .avf
typedef struct {
	int64_t 			size;
	int64_t 			offset;
	char 				filename[MAX_PATH];
} STRUCT_AVF_ESSENCE;


//
// AVF header definition
typedef struct {
	char				magicStr[8];
	int32_t				version;
	char				subtypeStr[3];
	int32_t				status;
	int32_t				essenceNm;
	STRUCT_AVF_ESSENCE	essence[MAX_ESSENCES];
} STRUCT_HEADER_AVF;


//
// The AVF context is the context of this demuxer (storage of local variables)
typedef struct AVFContext {
	MOVContext 			movContext;	// Original MOV struct MUST be the first part!!!

	// AVF stuff:
	STRUCT_HEADER_AVF 	avfHdrBf;

	// The specific requested filename
	char* 				extendedFilename;

	// The offset of the real data file in this file
    int64_t 			file_offset;

    // For options:
	const AVClass*		class;     	// Class for private options.
} AVFContext;


//
// Probe method should return max score when the memory struct contains the AVF header
static int avf_probe(AVProbeData *p)
{
	if (memcmp( p->buf, AVF_MAGIC, strlen( AVF_MAGIC ) )==0) {
    	return AVPROBE_SCORE_MAX;
    }
    return 0;
}


//
// Helper method to dump the headers to the log
static void avf_dump_header(AVFormatContext *s)
{
	AVFContext *avfPt = s->priv_data;

	av_log(s, AV_LOG_DEBUG, "The avf file contains the following item(s):\n");
	for( int cntIx = 0; cntIx < avfPt->avfHdrBf.essenceNm; cntIx++ ) {
		av_log( s, AV_LOG_DEBUG, "- %s [%ld]\n", avfPt->avfHdrBf.essence[cntIx].filename, avfPt->avfHdrBf.essence[cntIx].size );
	}
}


//
// Read the Grass Valley AVF header format
// The format consists of a simple header followed by a list of filenames and their sizes and offsets
// The filename of the file that should be played is given as extra option (an AVOption) in the AVDictionary
// When opening an .avf file using the avformat_open_input method, supply an extra AVDictionary with the 'extended_filename' option set.
// When the 'extended_filename' is empty, the first .m2v or .dv file will be automatically selected!
static int avf_read_header(AVFormatContext *s)
{
    AVFContext*			avfPt = s->priv_data;
    AVIOContext*		pb = s->pb;
    MOVStreamContext* 	pmsc = NULL;
    AVStream* 			st = NULL;
    int 				retValNm = 0;
    int					requestedStreamNumberIx = -1;
    char				currentFilenameSt[MAX_PATH] = "\0";

    // Set the mov-structure to avf
    avfPt->movContext.is_avf_extension = 1;

    // Copy the entire header data into the AVF struct
	if ( avio_size(pb) >= sizeof(STRUCT_HEADER_AVF) ) {
		memcpy( (void*)&avfPt->avfHdrBf, (void*)pb->buf_ptr ,sizeof(STRUCT_HEADER_AVF) );

		// The headers are now in memory, in the struct
		avf_dump_header(s);

		// Is an extended filename supplied?
		if ( avfPt->extendedFilename != NULL)
			strcpy( currentFilenameSt, avfPt->extendedFilename ); // Copy the name to a local buffer (since it can be set later on)


		// Now loop through all headers (with added safety)
		for( int cnt = 0; cnt < avfPt->avfHdrBf.essenceNm && cnt < MAX_ESSENCES; cnt++ )
		{
			// Set the file pointer to the start of the .mov file
			// (this is required for reading the .mov header by the mov code)
			if ( av_match_ext( avfPt->avfHdrBf.essence[cnt].filename, "mov" ) ) {
				avio_seek( pb, avfPt->avfHdrBf.essence[cnt].offset, SEEK_SET );
			}

			// Get the offset of the file that should be played
			if ( avfPt->extendedFilename == NULL ) {
				if ( av_match_ext( avfPt->avfHdrBf.essence[cnt].filename, "m2v,dv" ) ) {
					avfPt->file_offset = avfPt->avfHdrBf.essence[cnt].offset;
					av_log( s, AV_LOG_WARNING, "Extended filename not specified, using first found: %s\n", avfPt->avfHdrBf.essence[cnt].filename );
					strcpy( currentFilenameSt, avfPt->avfHdrBf.essence[cnt].filename );
					break;
				}
			} else {
				if ( strcmp( avfPt->avfHdrBf.essence[cnt].filename, avfPt->extendedFilename) == 0 ) {
					avfPt->file_offset = avfPt->avfHdrBf.essence[cnt].offset;
					av_log( s, AV_LOG_INFO, "Extended filename '%s' found, using offset: %ld\n", avfPt->extendedFilename, avfPt->file_offset );
				}
			}
		}

		// If we still do NOT have an extended file name, quit!
		if ( avfPt->file_offset == 0 )	{
			if ( avfPt->extendedFilename != NULL )
				av_log( s, AV_LOG_ERROR, "Extended filename '%s' is NOT found in the avf file!\n", avfPt->extendedFilename );
			else
				av_log( s, AV_LOG_ERROR, "Extended filename not specified and no default .m2v or .dv found in the avf file!\n" );
			return AVERROR_INVALIDDATA;
		}

		// Forward the call to the mov code that will read the .mov header:
		retValNm = mov_read_header( s, 1 );

		// Destroy all streams except the requested one
		if ( avfPt->movContext.fc->nb_streams > 0 ){

			for( int cnt = 0; cnt < avfPt->movContext.fc->nb_streams; cnt++ ){

				// Get a pointer to the MOV stream context
				pmsc = (MOVStreamContext*) avfPt->movContext.fc->streams[cnt]->priv_data;

				if ( pmsc != NULL ){
					if ( pmsc->drefs_count > 0 && pmsc->drefs ){
						if ( strcmp( pmsc->drefs[0].filename, currentFilenameSt ) != 0 )
						{
							// Not the desired file, clean it up

							/* free all data in a stream component */
							st = avfPt->movContext.fc->streams[cnt];
							if (st->parser)
								av_parser_close(st->parser);
							if (st->attached_pic.data)
								av_free_packet(&st->attached_pic);
							av_dict_free(&st->metadata);
							av_freep(&st->probe_data.buf);
							av_free(st->index_entries);
							av_free(st->codec->extradata);
							av_free(st->codec->subtitle_header);
							av_free(st->codec);
							av_free(st->priv_data);
							av_free(st->info);
							av_free(st);

							// Clear the stream pointer
							avfPt->movContext.fc->streams[cnt] = NULL;
						}
						else
							requestedStreamNumberIx = cnt;
					}
				}
			}
			if ( requestedStreamNumberIx < 0 ) {
				av_log( s, AV_LOG_ERROR, "Unable to locate stream with the name '%s'!\n", currentFilenameSt);
				return AVERROR_INVALIDDATA; // Stream not found, quit immediately
			}

			// Make sure the requested stream is this first stream
			avfPt->movContext.fc->nb_streams = 1; // Force to only a single stream
			if ( requestedStreamNumberIx != 0 )
			{
				// Move the requested stream number to the first entry
				avfPt->movContext.fc->streams[0] = avfPt->movContext.fc->streams[requestedStreamNumberIx];
				avfPt->movContext.fc->streams[requestedStreamNumberIx] = NULL;
			}

			// Get a pointer to the MOV stream structure
			pmsc = (MOVStreamContext*) avfPt->movContext.fc->streams[0]->priv_data;
			pmsc->pb = pb; // Let the buffer pointer point to our own file data buffer
			pmsc->ffindex = 0; // Make sure it uses the first stream

			// Now set the file offset to the file that should be played
			avio_seek( pb, avfPt->file_offset, SEEK_SET);
		}
		return retValNm; // Return the original return code
	}
	else
	{
		// Not enough data, generate an error??
		av_log(s, AV_LOG_ERROR, "Invalid AVF header found!\n");
		mov_read_close(s);
		return AVERROR_INVALIDDATA;
	}
}


//
// Read a packet
// JLB: This method had to be overwritten to make sure the offset points to the correct sub-file
//      All code should be the same  as the code in the mov_read_packet method with the exception of the
//		seek where the offset is added:
//			if (avio_seek(sc->pb, avf->file_offset + sample->pos, SEEK_SET) != avf->file_offset + sample->pos)
//
static int avf_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    AVFContext *avf = s->priv_data;
    MOVContext *mov = &(avf->movContext);
    MOVStreamContext *sc;
    AVIndexEntry *sample;
    AVStream *st = NULL;
    int ret;
 retry:
    sample = mov_find_next_sample(s, &st);
    if (!sample) {
        mov->found_mdat = 0;
        if (!mov->next_root_atom)
            return AVERROR_EOF;
        avio_seek(s->pb, mov->next_root_atom, SEEK_SET); // JLB Searching inside the mov file???
        mov->next_root_atom = 0;
        if (mov_read_default(mov, s->pb, (MOVAtom){ AV_RL32("root"), INT64_MAX }) < 0 ||
            s->pb->eof_reached)
            return AVERROR_EOF;
        av_dlog(s, "read fragments, offset 0x%"PRIx64"\n", avio_tell(s->pb));
        goto retry;
    }
    sc = st->priv_data;
    /* must be done just before reading, to avoid infinite loop on sample */
    sc->current_sample++;

    if (st->discard != AVDISCARD_ALL) {
        if (avio_seek(sc->pb, avf->file_offset + sample->pos, SEEK_SET) != avf->file_offset + sample->pos) { // JLB: Added file_offset
            av_log(mov->fc, AV_LOG_ERROR, "stream %d, offset 0x%"PRIx64": partial file\n",
                   sc->ffindex, sample->pos);
            return AVERROR_INVALIDDATA;
        }
        ret = av_get_packet(sc->pb, pkt, sample->size);
        if (ret < 0)
            return ret;
        if (sc->has_palette) {
            uint8_t *pal;

            pal = av_packet_new_side_data(pkt, AV_PKT_DATA_PALETTE, AVPALETTE_SIZE);
            if (!pal) {
                av_log(mov->fc, AV_LOG_ERROR, "Cannot append palette to packet\n");
            } else {
                memcpy(pal, sc->palette, AVPALETTE_SIZE);
                sc->has_palette = 0;
            }
        }
#if CONFIG_DV_DEMUXER
        if (mov->dv_demux && sc->dv_audio_container) {
            avpriv_dv_produce_packet(mov->dv_demux, pkt, pkt->data, pkt->size);
            av_free(pkt->data);
            pkt->size = 0;
            ret = avpriv_dv_get_packet(mov->dv_demux, pkt);
            if (ret < 0)
                return ret;
        }
#endif
    }

    pkt->stream_index = sc->ffindex;
    pkt->dts = sample->timestamp;

#ifdef GV_MODIFICATIONS
    // JLB: Addition to support atc start for this mov demuxer
//    if ( mov->start_time_code_ms != 0 ) {
//    	pkt->dts += ((mov->start_time_code_ms / av_q2d(st->time_base)) / 1000L );
//    }
#endif

    if (sc->ctts_data && sc->ctts_index < sc->ctts_count) {
        pkt->pts = pkt->dts + sc->dts_shift + sc->ctts_data[sc->ctts_index].duration;
        /* update ctts context */
        sc->ctts_sample++;
        if (sc->ctts_index < sc->ctts_count &&
            sc->ctts_data[sc->ctts_index].count == sc->ctts_sample) {
            sc->ctts_index++;
            sc->ctts_sample = 0;
        }
        if (sc->wrong_dts)
            pkt->dts = AV_NOPTS_VALUE;
    } else {
        int64_t next_dts = (sc->current_sample < st->nb_index_entries) ?
            st->index_entries[sc->current_sample].timestamp : st->duration;
        pkt->duration = next_dts - pkt->dts;
        pkt->pts = pkt->dts;
    }
    if (st->discard == AVDISCARD_ALL)
        goto retry;
    pkt->flags |= sample->flags & AVINDEX_KEYFRAME ? AV_PKT_FLAG_KEY : 0;
    pkt->pos = sample->pos;
    av_dlog(s, "stream %d, pts %"PRId64", dts %"PRId64", pos 0x%"PRIx64", duration %d\n",
            pkt->stream_index, pkt->pts, pkt->dts, pkt->pos, pkt->duration);
    return 0;
}


//
// Definition of this Demuxer and its options
//

#define OFFSET(x) offsetof(AVFContext, x)
#define DEC AV_OPT_FLAG_DECODING_PARAM

// I define 1 options, the extended filename, that is the name of the requested embedded file
static const AVOption avf_options[] = {
    { "extended_filename", "The filename of the AV file inside this .avf", OFFSET(extendedFilename), AV_OPT_TYPE_STRING, {.str = NULL}, 0, 0, DEC },
    { NULL },
};

//
// Definition of this demuxer class and options
static const AVClass avf_demuxer_class = {
    .class_name = "avf demuxer",
    .item_name  = av_default_item_name,
    .option     = avf_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

//
// This structure is registered in the allformats.c class using the REGISTER_DEMUXER (AVF, avf) macro
AVInputFormat ff_avf_demuxer = {
    .name           = "avf",
    .long_name      = NULL_IF_CONFIG_SMALL("AVF (Grass Valley)"),
    .priv_data_size = sizeof(AVFContext),
    .read_probe     = avf_probe,
    .read_header    = avf_read_header, 	// Overwrite the function using our own function
    .read_packet    = avf_read_packet,	// Overwrite the function using our own function
    .read_close     = mov_read_close, 	// JLB Point to the original code
    .read_seek      = mov_read_seek,  	// JLB Point to the original code
    .priv_class     = &avf_demuxer_class,
};
