/*
 * LXF demuxer
 * Copyright (c) 2010 Tomas Härdin
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "libavutil/intreadwrite.h"
#include "libavcodec/bytestream.h"
#include "avformat.h"
#include "internal.h"
#include "riff.h"

#define LXF_MAX_PACKET_HEADER_SIZE 256
#define LXF_HEADER_DATA_SIZE    120
#define LXF_IDENT               "LEITCH\0"
#define LXF_IDENT_LENGTH        8
#define LXF_SAMPLERATE          48000
#define LXF_MAX_AUDIO_PACKET    (8008*15*4) ///< 15-channel 32-bit NTSC audio frame
#define LXF_TIMECODE_TYPE_NONE  0xff
#define LXF_TIMECODE_TYPE_NTSC  0
#define LXF_TIMECODE_TYPE_PAL   2
#define LXF_PAL_DIVIDER_V1      28800
#define LXF_NTSC_DIVIDER_V1     24024

static const AVCodecTag lxf_tags[] = {
    { CODEC_ID_MJPEG,       0 },
    { CODEC_ID_MPEG1VIDEO,  1 },
    { CODEC_ID_MPEG2VIDEO,  2 },    //MpMl, 4:2:0
    { CODEC_ID_MPEG2VIDEO,  3 },    //MpPl, 4:2:2
    { CODEC_ID_DVVIDEO,     4 },    //DV25
    { CODEC_ID_DVVIDEO,     5 },    //DVCPRO
    { CODEC_ID_DVVIDEO,     6 },    //DVCPRO50
    { CODEC_ID_RAWVIDEO,    7 },    //PIX_FMT_ARGB, where alpha is used for chroma keying
    { CODEC_ID_RAWVIDEO,    8 },    //16-bit chroma key
    { CODEC_ID_MPEG2VIDEO,  9 },    //4:2:2 CBP ("Constrained Bytes per Gop")
    { CODEC_ID_NONE,        0 },
};

typedef struct {
    int channels;                       ///< number of audio channels. zero means no audio
    uint8_t temp[LXF_MAX_AUDIO_PACKET]; ///< temp buffer for de-planarizing the audio data
    uint32_t version;
    uint32_t video_format;
    uint32_t packet_type;
    uint32_t extended_size;
	int64_t  current_pts;
	uint8_t timecode_type;
} LXFDemuxContext;

static int lxf_probe(AVProbeData *p)
{
    if (!memcmp(p->buf, LXF_IDENT, LXF_IDENT_LENGTH))
        return AVPROBE_SCORE_MAX;

    return 0;
}

/**
 * Verify the checksum of an LXF packet header
 *
 * @param[in] header the packet header to check
 * @return zero if the checksum is OK, non-zero otherwise
 */
static int check_checksum(const uint8_t *header, int size)
{
    int x;
    uint32_t sum = 0;

    for (x = 0; x < size; x += 4)
        sum += AV_RL32(&header[x]);

    return sum;
}

/**
 * Read input until we find the next ident. If found, copy it to the header buffer
 *
 * @param[out] header where to copy the ident to
 * @return 0 if an ident was found, < 0 on I/O error
 */
static int sync(AVFormatContext *s, uint8_t *header)
{
    uint8_t buf[LXF_IDENT_LENGTH];
    int ret;

    if ((ret = avio_read(s->pb, buf, LXF_IDENT_LENGTH)) != LXF_IDENT_LENGTH)
        return ret < 0 ? ret : AVERROR_EOF;

    while (memcmp(buf, LXF_IDENT, LXF_IDENT_LENGTH)) {
        if (s->pb->eof_reached)
            return AVERROR_EOF;

        memmove(buf, &buf[1], LXF_IDENT_LENGTH-1);
        buf[LXF_IDENT_LENGTH-1] = avio_r8(s->pb);
    }

    memcpy(header, LXF_IDENT, LXF_IDENT_LENGTH);

    return 0;
}

/**
 * Read and checksum the next packet header
 *
 * @return the size of the payload following the header or < 0 on failure
 */
static int get_packet_header(AVFormatContext *s)
{
    LXFDemuxContext *lxf = s->priv_data;
    AVIOContext   *pb  = s->pb;
    int track_size, samples, ret;
    uint32_t audio_format, header_size, channels, tmp;
    AVStream *st;
    uint8_t header[LXF_MAX_PACKET_HEADER_SIZE];
    const uint8_t *p;

    //find and read the ident
    if ((ret = sync(s, header)) < 0)
        return ret;

    ret = avio_read(pb, header + LXF_IDENT_LENGTH, 8);
    if (ret != 8)
        return ret < 0 ? ret : AVERROR_EOF;

    p = header + LXF_IDENT_LENGTH;
    lxf->version = bytestream_get_le32(&p);
    header_size = bytestream_get_le32(&p);
    if (lxf->version > 1)
        av_log_ask_for_sample(s, "Unknown format version %i\n", lxf->version);
    if (header_size < (lxf->version ? 72 : 60) ||
        header_size > LXF_MAX_PACKET_HEADER_SIZE ||
        (header_size & 3)) {
        av_log(s, AV_LOG_ERROR, "Invalid header size 0x%x\n", header_size);
        return AVERROR_INVALIDDATA;
    }

    //read the rest of the packet header
    if ((ret = avio_read(pb, header + (p - header),
                          header_size - (p - header))) !=
                          header_size - (p - header)) {
        return ret < 0 ? ret : AVERROR_EOF;
    }

    if (check_checksum(header, header_size))
        av_log(s, AV_LOG_ERROR, "checksum error\n");

    lxf->packet_type = bytestream_get_le32(&p);
    p+=4; // skip stream id

   if (!lxf->version) {
    	lxf->current_pts = ((int64_t)bytestream_get_le32(&p)) >> 1; // fields
    	p+=4; // skip packet duration
	} else {
    	lxf->current_pts = bytestream_get_le64(&p);
		if (lxf->timecode_type == LXF_TIMECODE_TYPE_PAL) {
    		lxf->current_pts /= LXF_PAL_DIVIDER_V1;
		} else {
    		lxf->current_pts /= LXF_NTSC_DIVIDER_V1;
		}
    	p+=8; // skip packet duration
	}
 
    lxf->extended_size = 0;
    switch (lxf->packet_type) {
    case 0:
        //video
        lxf->video_format = bytestream_get_le32(&p);
        ret               = bytestream_get_le32(&p);
        //skip VBI data and metadata
        avio_skip(pb, (int64_t)(uint32_t)AV_RL32(p + 4) +
                      (int64_t)(uint32_t)AV_RL32(p + 12));
        break;
    case 1:
        //audio
        if (!(st = s->streams[1])) {
            av_log(s, AV_LOG_INFO, "got audio packet, but no audio stream present\n");
            break;
        }

        if (lxf->version == 0) p += 8;
        audio_format = bytestream_get_le32(&p);
        channels     = bytestream_get_le32(&p);
        track_size   = bytestream_get_le32(&p);

        //set codec based on specified audio bitdepth
        //we only support tightly packed 16-, 20-, 24- and 32-bit PCM at the moment
        st->codec->bits_per_coded_sample = (audio_format >> 6) & 0x3F;

        if (st->codec->bits_per_coded_sample != (audio_format & 0x3F)) {
            av_log(s, AV_LOG_WARNING, "only tightly packed PCM currently supported\n");
            return AVERROR_PATCHWELCOME;
        }

        switch (st->codec->bits_per_coded_sample) {
        case 16: st->codec->codec_id = CODEC_ID_PCM_S16LE; break;
        case 20: st->codec->codec_id = CODEC_ID_PCM_LXF;   break;
        case 24: st->codec->codec_id = CODEC_ID_PCM_S24LE; break;
        case 32: st->codec->codec_id = CODEC_ID_PCM_S32LE; break;
        default:
            av_log(s, AV_LOG_WARNING,
                   "only 16-, 20-, 24- and 32-bit PCM currently supported\n");
            return AVERROR_PATCHWELCOME;
        }

		if (lxf->timecode_type == LXF_TIMECODE_TYPE_NONE) {
            samples = track_size * 8 / st->codec->bits_per_coded_sample;
 
            //use audio packet size to determine video standard
            //for NTSC we have one 8008-sample audio frame per five video frames
            if (samples == LXF_SAMPLERATE * 5005 / 30000) {
 	   			lxf->timecode_type = LXF_TIMECODE_TYPE_NTSC;
            } else {
 	   			lxf->timecode_type = LXF_TIMECODE_TYPE_PAL;
            }
		}

        //TODO: warning if track mask != (1 << channels) - 1?
        ret = av_popcount(channels) * track_size;

        break;

    default:
        tmp = bytestream_get_le32(&p);
        ret = bytestream_get_le32(&p);
        if (tmp == 1)
            lxf->extended_size = bytestream_get_le32(&p);
        break;
    }

    return ret;
}

static int lxf_read_header(AVFormatContext *s)
{
    LXFDemuxContext *lxf = s->priv_data;
    AVIOContext   *pb  = s->pb;
    uint8_t header_data[LXF_HEADER_DATA_SIZE];
    int ret, cnt;
    AVStream *st;
    uint64_t store_pos = 0;
    uint32_t video_params, disk_params, num, den;
    uint16_t record_date, expiration_date;
    
	lxf->timecode_type = LXF_TIMECODE_TYPE_NONE;;
    if ((ret = get_packet_header(s)) < 0)
        return ret;

    if (ret != LXF_HEADER_DATA_SIZE) {
        av_log(s, AV_LOG_ERROR, "expected %d B size header, got %d\n",
               LXF_HEADER_DATA_SIZE, ret);
        return AVERROR_INVALIDDATA;
    }

    if ((ret = avio_read(pb, header_data, LXF_HEADER_DATA_SIZE)) != LXF_HEADER_DATA_SIZE)
        return ret < 0 ? ret : AVERROR_EOF;

    if (!(st = avformat_new_stream(s, NULL)))
        return AVERROR(ENOMEM);

    st->duration          = AV_RL32(&header_data[32]);
    video_params          = AV_RL32(&header_data[40]);
    record_date           = AV_RL16(&header_data[56]);
    expiration_date       = AV_RL16(&header_data[58]);
    disk_params           = AV_RL32(&header_data[116]);

    st->codec->codec_type = AVMEDIA_TYPE_VIDEO;
    st->codec->bit_rate   = 1000000 * ((video_params >> 14) & 0xFF);
    st->codec->codec_tag  = video_params & 0xF;
    st->codec->codec_id   = ff_codec_get_id(lxf_tags, st->codec->codec_tag);

    if ((video_params >> 22) & 1)
        av_log(s, AV_LOG_WARNING, "VBI data not yet supported\n");

    if ((lxf->channels = (disk_params >> 2) & 0xF)) {
        if (!(st = avformat_new_stream(s, NULL)))
            return AVERROR(ENOMEM);

        st->codec->codec_type  = AVMEDIA_TYPE_AUDIO;
        st->codec->sample_rate = LXF_SAMPLERATE;
        st->codec->channels    = lxf->channels;
    }

    avio_skip(s->pb, lxf->extended_size);

	// Store file pointer
	store_pos = avio_tell( s->pb );

	// Hopefully we discover the timecode type via the audio samples.
    for (cnt=0; cnt < 99 && lxf->timecode_type == LXF_TIMECODE_TYPE_NONE; cnt++) {
    	if ((ret = get_packet_header(s)) < 0)
       		return ret;

		avio_skip( s->pb, ret );
	} 

	// Restore file pointer
	if ((ret = avio_seek(s->pb, store_pos, SEEK_SET)) < 0)
		return ret;

    if (lxf->timecode_type == LXF_TIMECODE_TYPE_NONE) {
		if (lxf->version) {
			if ((lxf->current_pts % LXF_NTSC_DIVIDER_V1) == 0) {
  				lxf->timecode_type = LXF_TIMECODE_TYPE_NTSC;
			} else {
  				lxf->timecode_type = LXF_TIMECODE_TYPE_PAL;
			}
		} else {
    		av_log(s, AV_LOG_WARNING, "Could find timecode type, assume PAL.\n");
  			lxf->timecode_type = LXF_TIMECODE_TYPE_PAL;
		}
	}

    if (lxf->timecode_type == LXF_TIMECODE_TYPE_PAL) {
		num=1; den=25;
	} else {
		num=1001; den=30000;
	}

    for (cnt=0; cnt < s->nb_streams; cnt++) {
        avpriv_set_pts_info(s->streams[cnt], 64, num, den);
	}

    return 0;
}

/**
 * De-planerize the PCM data in lxf->temp
 * FIXME: remove this once support for planar audio is added to libavcodec
 *
 * @param[out] out where to write the de-planerized data to
 * @param[in] bytes the total size of the PCM data
 */
static void deplanarize(LXFDemuxContext *lxf, AVStream *ast, uint8_t *out, int bytes)
{
    int x, y, z, i, bytes_per_sample = ast->codec->bits_per_coded_sample >> 3;

    for (z = i = 0; z < lxf->channels; z++)
        for (y = 0; y < bytes / bytes_per_sample / lxf->channels; y++)
            for (x = 0; x < bytes_per_sample; x++, i++)
                out[x + bytes_per_sample*(z + y*lxf->channels)] = lxf->temp[i];
}

static int lxf_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    LXFDemuxContext *lxf = s->priv_data;
    AVIOContext   *pb  = s->pb;
    uint8_t *buf;
    AVStream *ast = NULL;
    uint32_t stream;
    int ret, ret2;

    if ((ret = get_packet_header(s)) < 0)
        return ret;

    stream = lxf->packet_type;

    if (stream > 1) {
        av_log(s, AV_LOG_WARNING, "got packet with illegal stream index %u\n", stream);
        return AVERROR(EAGAIN);
    }

    if (stream == 1 && !(ast = s->streams[1])) {
        av_log(s, AV_LOG_ERROR, "got audio packet without having an audio stream\n");
        return AVERROR_INVALIDDATA;
    }

    //make sure the data fits in the de-planerization buffer
    if (ast && ret > LXF_MAX_AUDIO_PACKET) {
        av_log(s, AV_LOG_ERROR, "audio packet too large (%i > %i)\n",
            ret, LXF_MAX_AUDIO_PACKET);
        return AVERROR_INVALIDDATA;
    }

    if ((ret2 = av_new_packet(pkt, ret)) < 0)
        return ret2;

    //read non-20-bit audio data into lxf->temp so we can deplanarize it
    buf = ast && ast->codec->codec_id != CODEC_ID_PCM_LXF ? lxf->temp : pkt->data;

    if ((ret2 = avio_read(pb, buf, ret)) != ret) {
        av_free_packet(pkt);
        return ret2 < 0 ? ret2 : AVERROR_EOF;
    }

    pkt->pts = AV_NOPTS_VALUE;
    pkt->dts = lxf->current_pts;
    pkt->stream_index = stream;

    if (ast) {
    	//pkt->pts = lxf->current_pts;
        if(ast->codec->codec_id != CODEC_ID_PCM_LXF)
            deplanarize(lxf, ast, pkt->data, ret);
    } else {
        //picture type (0 = closed I, 1 = open I, 2 = P, 3 = B)
        if (((lxf->video_format >> 22) & 0x3) < 2)
            pkt->flags |= AV_PKT_FLAG_KEY;

    }

    return ret;
}

/** --------------------------------------------------------------------------------------
 Jump/Seek a position in the file
 - This method works by calculating the average frame size in the file and jumping to the
   calculated position (I assume all frames are equal size which is of course not the real case).
 - Read the header, check it's PTS,
 - When the distance between the new PTS and the requested PTS is too large, jump again
   with a calculated gues to the front (both negative jump and positive jump).
 - When almost there, use steps of +1. Note that this is only done with the new ID is higher then
   the current ID since we cannot easily jump back a single step.

 @param[in] s pointer to the format context
 @param[in] stream_index the index of the stream
 @param[in] sample_time the requested new pts (in stream units, i.e. frames)
 @param[in] flags search flags (not used)
 @return 0 upon success, < 0 for error

 @author    Jochem Bakker
 ----------------------------------------------------------------------------------------- */
static int lxf_read_seek(AVFormatContext *s, int stream_index, int64_t sample_time, int flags)
{
    LXFDemuxContext *lxf = s->priv_data;
	AVStream*		st = s->streams[stream_index];
	int64_t			sample_time_request = 0;
	int64_t 		seek_pos = 0;
	int64_t			avgFrameSzBt = 0;
	int64_t   		ret = 0;
	int64_t			pts = 0;
	int				bailOutCt = 1000; // Do a maximum of X steps


	if ( st == NULL || st->duration == 0 )
		return 0; // Safety valve: Do nothing

	/* ------->  Subtract the stream start code of the stream when absolute time is requested */
	sample_time_request = sample_time;
	if (st->start_time != AV_NOPTS_VALUE && st->start_time != 0 && (flags & AVSEEK_FLAG_ABSOLUTE_TIME))
	{
		sample_time -= st->start_time;
		av_log(s, AV_LOG_INFO, "LxfSeek, sample_time_request %ld corrected to sample_time %ld\n", sample_time_request, sample_time );
	}
	else
		av_log(s, AV_LOG_INFO, "LxfSeek, sample_time_request %ld\n", sample_time_request);


	if ( sample_time == 0 ) {
		avio_seek( s->pb, 0, SEEK_SET ); // Start at the begin
		ff_update_cur_dts(s, st, sample_time_request);
		return 0;
	}

	/* clamp above zero */
	sample_time = FFMAX(sample_time, 0);

	/* -------> Calculate the average size per frame */
	avgFrameSzBt =  avio_size( s->pb )/ st->duration;

	/* -------> Calculate the new position */
	seek_pos = avgFrameSzBt * sample_time;
	if ((ret = avio_seek(s->pb, seek_pos, SEEK_SET)) < 0)
		return ret;

	// Do calculated-guesses-algo
	do
	{
        if ((ret = get_packet_header(s)) < 0)
            return ret;
 
		/* -------> Get the current position (probably shifted due to locating the header) */
		seek_pos = avio_tell( s->pb );

		pts = lxf->current_pts;

		if ( pts == sample_time ) {
			/* -------> SKip not video packets. */
            if (lxf->packet_type != 0) {
				/* -------> Skip this frame and jump to the next one */
				avio_skip( s->pb, ret );
				bailOutCt--; // Safety valve...
                continue;
            }
			/* -------> Seek to the start of the marker (filepos is pointing to the data now, we want it to point to the header) */
			avio_seek(s->pb, seek_pos, SEEK_SET);
			ff_update_cur_dts(s, st, sample_time_request);

			return 0;

		} else if ( pts > sample_time ) {
			//
			/* -------> Seek back a lot of frames */
			seek_pos -= avgFrameSzBt * ((pts - sample_time) + 2); // Seek back some extra frames
			seek_pos = FFMAX(seek_pos, 0);

			/* -------> Jump to some frames before */
			if ((ret = avio_seek(s->pb, seek_pos, SEEK_SET)) < 0)
				return ret;
		} else if ( pts < sample_time ) {
			if ( pts < sample_time - 4) {
				//
				/* -------> still a long way to go, use a large step */
				seek_pos += avgFrameSzBt * ((sample_time - pts) - 2); // Take a large step, but land before the requested frame
				seek_pos = FFMAX(seek_pos, 0);

				/* ------->  Skip a lot of frames */
				if ((ret = avio_seek(s->pb, seek_pos, SEEK_SET)) < 0)
					return ret;
			} else {
				//
				/* -------> Skip this frame and jump to the next one */
				avio_skip( s->pb, ret );
			}
		}
		bailOutCt--; // Safety valve...
	} while ( bailOutCt > 0 );

	return 0; // Success!
}


AVInputFormat ff_lxf_demuxer = {
    .name           = "lxf",
    .long_name      = NULL_IF_CONFIG_SMALL("VR native stream format (LXF)"),
    .priv_data_size = sizeof(LXFDemuxContext),
    .read_probe     = lxf_probe,
    .read_header    = lxf_read_header,
    .read_packet    = lxf_read_packet,
    .read_seek      = lxf_read_seek,
    .codec_tag      = (const AVCodecTag* const []){lxf_tags, 0},
};
