/*
 * MXF demuxer.
 * Copyright (c) 2006 SmartJog S.A., Baptiste Coudurier <baptiste dot coudurier at smartjog dot com>
 *
 * This file is part of Libav.
 *
 * Libav is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * Libav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/*
 * References
 * SMPTE 336M KLV Data Encoding Protocol Using Key-Length-Value
 * SMPTE 377M MXF File Format Specifications
 * SMPTE 378M Operational Pattern 1a
 * SMPTE 379M MXF Generic Container
 * SMPTE 381M Mapping MPEG Streams into the MXF Generic Container
 * SMPTE 382M Mapping AES3 and Broadcast Wave Audio into the MXF Generic Container
 * SMPTE 383M Mapping DV-DIF Data to the MXF Generic Container
 *
 * Principle
 * Search for Track numbers which will identify essence element KLV packets.
 * Search for SourcePackage which define tracks which contains Track numbers.
 * Material Package contains tracks with reference to SourcePackage tracks.
 * Search for Descriptors (Picture, Sound) which contains codec info and parameters.
 * Assign Descriptors to correct Tracks.
 *
 * Metadata reading functions read Local Tags, get InstanceUID(0x3C0A) then add MetaDataSet to MXFContext.
 * Metadata parsing resolves Strong References to objects.
 *
 * Simple demuxer, only OP1A supported and some files might not work at all.
 * Only tracks with associated descriptors will be decoded. "Highly Desirable" SMPTE 377M D.1
 */
/*
 * This file was modified by Grass Valley (2013/2014).
 * See libav-gv-changes.txt file for modification history.
 */

//#define DEBUG

#include "libavutil/aes.h"
#include "libavutil/mathematics.h"
#include "libavutil/timecode.h"
#include "libavcodec/bytestream.h"
#include "avformat.h"
#include "internal.h"
#include "mxf.h"

#include "libavcodec/mpegvideo.h"

typedef enum {
    Header,
    BodyPartition,
    Footer
} MXFPartitionType;

typedef enum {
    OP1a = 1,
    OP1b,
    OP1c,
    OP2a,
    OP2b,
    OP2c,
    OP3a,
    OP3b,
    OP3c,
    OPAtom,
    OPSonyOpt,  /* FATE sample, violates the spec in places */
} MXFOP;

typedef struct {
    int closed;
    int complete;
    MXFPartitionType type;
    uint64_t previous_partition;
    int index_sid;
    int body_sid;
    int64_t this_partition;
    int64_t essence_offset;         ///< absolute offset of essence
    int64_t essence_length;
    int32_t kag_size;
    int64_t header_byte_count;
    int64_t index_byte_count;
    int pack_length;

    int64_t body_offset;         ///< JLB the body-offset as found in the partition header data
} MXFPartition;


// JLB Added for reading absolute seek
typedef struct {
    UID uid;
    enum MXFMetadataSetType type;
    UID data_definition;
    int64_t start_time;	// in frames
    int16_t rounded_timecode_base;
    int8_t drop_frame;
    int source_track_id;
} MXFTimecodeComponent;


// JLB Added for essence container data
typedef struct {
    UID uid;
    enum MXFMetadataSetType type;
    UID linked_package_id;
    int32_t index_sid;
    int32_t body_sid;
} MXFEssenceContainerData;

typedef struct {
    UID uid;
    enum MXFMetadataSetType type;
    UID source_container_ul;
} MXFCryptoContext;

typedef struct {
    UID uid;
    enum MXFMetadataSetType type;
    UID source_package_uid;
    UID data_definition_ul;
    int64_t duration;
    int64_t start_position;
    int source_track_id;
} MXFStructuralComponent;

typedef struct {
    UID uid;
    enum MXFMetadataSetType type;
    UID data_definition_ul;
    UID *structural_components_refs;
    int structural_components_count;
    int64_t duration;
} MXFSequence;

typedef struct {
    UID uid;
    enum MXFMetadataSetType type;
    MXFSequence *sequence; /* mandatory, and only one */
    UID sequence_ref;
    int track_id;
    uint8_t track_number[4];
    AVRational edit_rate;
    int intra_only;
} MXFTrack;

typedef struct {
    UID uid;
    enum MXFMetadataSetType type;
    UID essence_container_ul;
    UID essence_codec_ul;
    AVRational sample_rate;
    AVRational aspect_ratio;
    int width;
    int height; /* Field height, not frame height */
    int frame_layout; /* See MXFFrameLayout enum */
    int channels;
    int bits_per_sample;
    unsigned int component_depth;
    unsigned int horiz_subsampling;
    unsigned int vert_subsampling;
    UID *sub_descriptors_refs;
    int sub_descriptors_count;
    int linked_track_id;
    uint8_t *extradata;
    int extradata_size;
    enum AVPixelFormat pix_fmt;

    int active_format; // JLB Added
} MXFDescriptor;

typedef struct {
    UID uid;
    enum MXFMetadataSetType type;
    int edit_unit_byte_count;
    int index_sid;
    int body_sid;
    AVRational index_edit_rate;
    uint64_t index_start_position;
    uint64_t index_duration;
    int8_t *temporal_offset_entries;
    int *flag_entries;
    uint64_t *stream_offset_entries;
    int nb_index_entries;

    int64_t this_partition; /* Partition offset where this index table is located */
    int nb_delta_entries;
    int8_t *delta_pos_table_entries;
    //uint32_t *delta_element_entries;
    int8_t *key_frame_offset_entries;
} MXFIndexTableSegment;

typedef struct {
    UID uid;
    enum MXFMetadataSetType type;
    UID package_uid;
    UID *tracks_refs;
    int tracks_count;
    MXFDescriptor *descriptor; /* only one */
    UID descriptor_ref;
} MXFPackage;

typedef struct {
    UID uid;
    enum MXFMetadataSetType type;
} MXFMetadataSet;

/* decoded index table */
typedef struct {
    int index_sid;
    int body_sid;
    int nb_ptses;               /* number of PTSes or total duration of index */
    int64_t first_dts;          /* DTS = EditUnit + first_dts */
    int64_t *ptses;             /* maps EditUnit -> PTS */
    int nb_segments;
    MXFIndexTableSegment **segments;    /* sorted by IndexStartPosition */
    AVIndexEntry *fake_index;   /* used for calling ff_index_search_timestamp() */
} MXFIndexTable;

typedef struct {
    MXFPartition *partitions;
    unsigned partitions_count;
    MXFOP op;
    UID *packages_refs;
    int packages_count;
    MXFMetadataSet **metadata_sets;
    int metadata_sets_count;
    AVFormatContext *fc;
    struct AVAES *aesc;
    uint8_t *local_tags;
    int local_tags_count;
    uint64_t footer_partition;
    KLVPacket current_klv_data;
    int current_klv_index;
    int run_in;
    MXFPartition *current_partition;
    int parsing_backward;
    int64_t last_forward_tell;
    int last_forward_partition;
    int current_edit_unit;
    int nb_index_tables;
    MXFIndexTable *index_tables;
    int edit_units_per_packet;      ///< how many edit units to read at a time (PCM, OPAtom)
    int		m_indexSID;				// JLB: Added the Index SID of the current playing package
    uint64_t first_essence_offset;  // JLB: The offset of the first essence in the file
    uint64_t overall_start_time_ms;    ///< Global start time of the data (in ms)
} MXFContext;

enum MXFWrappingScheme {
    Frame,
    Clip,
};

/* NOTE: klv_offset is not set (-1) for local keys */
typedef int MXFMetadataReadFunc(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset);

typedef struct {
    const UID key;
    MXFMetadataReadFunc *read;
    int ctx_size;
    enum MXFMetadataSetType type;
} MXFMetadataReadTableEntry;

/* partial keys to match */
static const uint8_t mxf_header_partition_pack_key[]       = { 0x06,0x0e,0x2b,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x02 };
static const uint8_t mxf_essence_element_key[]             = { 0x06,0x0e,0x2b,0x34,0x01,0x02,0x01,0x01,0x0d,0x01,0x03,0x01 };
static const uint8_t mxf_avid_essence_element_key[]        = { 0x06,0x0e,0x2b,0x34,0x01,0x02,0x01,0x01,0x0e,0x04,0x03,0x01 };
static const uint8_t mxf_system_item_key[]                 = { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0D,0x01,0x03,0x01,0x04 };
static const uint8_t mxf_klv_key[]                         = { 0x06,0x0e,0x2b,0x34 };
/* complete keys to match */
static const uint8_t mxf_crypto_source_container_ul[]      = { 0x06,0x0e,0x2b,0x34,0x01,0x01,0x01,0x09,0x06,0x01,0x01,0x02,0x02,0x00,0x00,0x00 };
static const uint8_t mxf_encrypted_triplet_key[]           = { 0x06,0x0e,0x2b,0x34,0x02,0x04,0x01,0x07,0x0d,0x01,0x03,0x01,0x02,0x7e,0x01,0x00 };
static const uint8_t mxf_encrypted_essence_container[]     = { 0x06,0x0e,0x2b,0x34,0x04,0x01,0x01,0x07,0x0d,0x01,0x03,0x01,0x02,0x0b,0x01,0x00 };
static const uint8_t mxf_sony_mpeg4_extradata[]            = { 0x06,0x0e,0x2b,0x34,0x04,0x01,0x01,0x01,0x0e,0x06,0x06,0x02,0x02,0x01,0x00,0x00 };

// JLB added index table key
static const uint8_t mxf_index_table[]            		   = { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x10,0x01,0x00 };

// Forward declaration
static int64_t mxf_find_offset(AVFormatContext *s, int stream_index, int64_t sample_time, int flags, int64_t* psample_time_request, int64_t* pend_of_essencecontainer, int showInfoLogging);


#define IS_KLV_KEY(x, y) (!memcmp(x, y, sizeof(y)))

static int64_t klv_decode_ber_length(AVIOContext *pb)
{
    uint64_t size = avio_r8(pb);
    if (size & 0x80) { /* long form */
        int bytes_num = size & 0x7f;
        /* SMPTE 379M 5.3.4 guarantee that bytes_num must not exceed 8 bytes */
        if (bytes_num > 8)
            return AVERROR_INVALIDDATA;
        size = 0;
        while (bytes_num--)
            size = size << 8 | avio_r8(pb);
    }
    return size;
}

static int mxf_read_sync(AVIOContext *pb, const uint8_t *key, unsigned size)
{
    int i, b;
    for (i = 0; i < size && !pb->eof_reached; i++) {
        b = avio_r8(pb);
        if (b == key[0])
            i = 0;
        else if (b != key[i])
            i = -1;
    }
    return i == size;
}

static int klv_read_packet(KLVPacket *klv, AVIOContext *pb)
{
    if (!mxf_read_sync(pb, mxf_klv_key, 4))
        return AVERROR_INVALIDDATA;
    klv->offset = avio_tell(pb) - 4;
    memcpy(klv->key, mxf_klv_key, 4);
    avio_read(pb, klv->key + 4, 12);
    klv->length = klv_decode_ber_length(pb);
    return klv->length == -1 ? -1 : 0;
}

static int mxf_get_stream_index(AVFormatContext *s, KLVPacket *klv)
{
    int i;

    for (i = 0; i < s->nb_streams; i++) {
        MXFTrack *track = s->streams[i]->priv_data;
        /* SMPTE 379M 7.3 */
        if (!memcmp(klv->key + sizeof(mxf_essence_element_key), track->track_number, sizeof(track->track_number)))
            return i;
    }
    /* return 0 if only one stream, for OP Atom files with 0 as track number */
    return s->nb_streams == 1 ? 0 : -1;
}

/* XXX: use AVBitStreamFilter */
static int mxf_get_d10_aes3_packet(AVIOContext *pb, AVStream *st, AVPacket *pkt, int64_t length)
{
    const uint8_t *buf_ptr, *end_ptr;
    uint8_t *data_ptr;
    int i;

    if (length > 61444) /* worst case PAL 1920 samples 8 channels */
        return AVERROR_INVALIDDATA;
    length = av_get_packet(pb, pkt, length);
    if (length < 0)
        return length;
    data_ptr = pkt->data;
    end_ptr = pkt->data + length;
    buf_ptr = pkt->data + 4; /* skip SMPTE 331M header */
    for (; buf_ptr + st->codec->channels*4 < end_ptr; ) {
        for (i = 0; i < st->codec->channels; i++) {
            uint32_t sample = bytestream_get_le32(&buf_ptr);
            if (st->codec->bits_per_coded_sample == 24)
                bytestream_put_le24(&data_ptr, (sample >> 4) & 0xffffff);
            else
                bytestream_put_le16(&data_ptr, (sample >> 12) & 0xffff);
        }
        buf_ptr += 32 - st->codec->channels*4; // always 8 channels stored SMPTE 331M
    }
    av_shrink_packet(pkt, data_ptr - pkt->data);
    return 0;
}

static int mxf_decrypt_triplet(AVFormatContext *s, AVPacket *pkt, KLVPacket *klv)
{
    static const uint8_t checkv[16] = {0x43, 0x48, 0x55, 0x4b, 0x43, 0x48, 0x55, 0x4b, 0x43, 0x48, 0x55, 0x4b, 0x43, 0x48, 0x55, 0x4b};
    MXFContext *mxf = s->priv_data;
    AVIOContext *pb = s->pb;
    int64_t end = avio_tell(pb) + klv->length;
    int64_t size;
    uint64_t orig_size;
    uint64_t plaintext_size;
    uint8_t ivec[16];
    uint8_t tmpbuf[16];
    int index;

    if (!mxf->aesc && s->key && s->keylen == 16) {
        mxf->aesc = av_aes_alloc();
        if (!mxf->aesc)
            return AVERROR(ENOMEM);
        av_aes_init(mxf->aesc, s->key, 128, 1);
    }
    // crypto context
    avio_skip(pb, klv_decode_ber_length(pb));
    // plaintext offset
    klv_decode_ber_length(pb);
    plaintext_size = avio_rb64(pb);
    // source klv key
    klv_decode_ber_length(pb);
    avio_read(pb, klv->key, 16);
    if (!IS_KLV_KEY(klv, mxf_essence_element_key))
        return AVERROR_INVALIDDATA;
    index = mxf_get_stream_index(s, klv);
    if (index < 0)
        return AVERROR_INVALIDDATA;
    // source size
    klv_decode_ber_length(pb);
    orig_size = avio_rb64(pb);
    if (orig_size < plaintext_size)
        return AVERROR_INVALIDDATA;
    // enc. code
    size = klv_decode_ber_length(pb);
    if (size < 32 || size - 32 < orig_size)
        return AVERROR_INVALIDDATA;
    avio_read(pb, ivec, 16);
    avio_read(pb, tmpbuf, 16);
    if (mxf->aesc)
        av_aes_crypt(mxf->aesc, tmpbuf, tmpbuf, 1, ivec, 1);
    if (memcmp(tmpbuf, checkv, 16))
        av_log(s, AV_LOG_ERROR, "probably incorrect decryption key\n");
    size -= 32;
    size = av_get_packet(pb, pkt, size);
    if (size < 0)
        return size;
    else if (size < plaintext_size)
        return AVERROR_INVALIDDATA;
    size -= plaintext_size;
    if (mxf->aesc)
        av_aes_crypt(mxf->aesc, &pkt->data[plaintext_size],
                     &pkt->data[plaintext_size], size >> 4, ivec, 1);
    av_shrink_packet(pkt, orig_size);
    pkt->stream_index = index;
    avio_skip(pb, end - avio_tell(pb));
    return 0;
}

static int mxf_read_partition_pack(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFContext *mxf = arg;
    MXFPartition *partition, *tmp_part;
    UID op;
    uint64_t footer_partition;
    uint32_t nb_essence_containers;

    if (mxf->partitions_count+1 >= UINT_MAX / sizeof(*mxf->partitions))
        return AVERROR(ENOMEM);

    av_dlog(mxf->fc, AV_LOG_DEBUG, "Reading PartitionPack: %d", mxf->partitions_count);

    tmp_part = av_realloc(mxf->partitions, (mxf->partitions_count + 1) * sizeof(*mxf->partitions));
    if (!tmp_part)
        return AVERROR(ENOMEM);
    mxf->partitions = tmp_part;

    if (mxf->parsing_backward) {
        /* insert the new partition pack in the middle
         * this makes the entries in mxf->partitions sorted by offset */
        memmove(&mxf->partitions[mxf->last_forward_partition+1],
                &mxf->partitions[mxf->last_forward_partition],
                (mxf->partitions_count - mxf->last_forward_partition)*sizeof(*mxf->partitions));
        partition = mxf->current_partition = &mxf->partitions[mxf->last_forward_partition];
    } else {
        mxf->last_forward_partition++;
        partition = mxf->current_partition = &mxf->partitions[mxf->partitions_count];
    }

    memset(partition, 0, sizeof(*partition));
    mxf->partitions_count++;
    partition->pack_length = avio_tell(pb) - klv_offset + size;

    switch(uid[13]) {
    case 2:
        partition->type = Header;
        break;
    case 3:
        partition->type = BodyPartition;
        break;
    case 4:
        partition->type = Footer;
        break;
    default:
        av_log(mxf->fc, AV_LOG_ERROR, "unknown partition type %i\n", uid[13]);
        return AVERROR_INVALIDDATA;
    }

    /* consider both footers to be closed (there is only Footer and CompleteFooter) */
    partition->closed = partition->type == Footer || !(uid[14] & 1);
    partition->complete = uid[14] > 2;
    avio_skip(pb, 4);
    partition->kag_size = avio_rb32(pb);
    partition->this_partition = avio_rb64(pb);
    partition->previous_partition = avio_rb64(pb);
    footer_partition = avio_rb64(pb);
    partition->header_byte_count = avio_rb64(pb);
    partition->index_byte_count = avio_rb64(pb);
    partition->index_sid = avio_rb32(pb);
    partition->body_offset = avio_rb64(pb); // JLB Added this line instead of a skip
    partition->body_sid = avio_rb32(pb);
    avio_read(pb, op, sizeof(UID));
    nb_essence_containers = avio_rb32(pb);

    /* some files don'thave FooterPartition set in every partition */
    if (footer_partition) {
        if (mxf->footer_partition && mxf->footer_partition != footer_partition) {
            av_log(mxf->fc, AV_LOG_ERROR,
                   "inconsistent FooterPartition value: %"PRIu64" != %"PRIu64"\n",
                   mxf->footer_partition, footer_partition);
        } else {
            mxf->footer_partition = footer_partition;
        }
    }

    av_dlog(mxf->fc,
            "PartitionPack: ThisPartition = 0x%"PRIX64
            ", PreviousPartition = 0x%"PRIX64", "
            "FooterPartition = 0x%"PRIX64", IndexSID = %i, BodySID = %i\n",
            partition->this_partition,
            partition->previous_partition, footer_partition,
            partition->index_sid, partition->body_sid);

    /* sanity check PreviousPartition if set */
    if (partition->previous_partition &&
        mxf->run_in + partition->previous_partition >= klv_offset) {
        av_log(mxf->fc, AV_LOG_ERROR,
               "PreviousPartition points to this partition or forward\n");
        return AVERROR_INVALIDDATA;
    }

    if      (op[12] == 1 && op[13] == 1) mxf->op = OP1a;
    else if (op[12] == 1 && op[13] == 2) mxf->op = OP1b;
    else if (op[12] == 1 && op[13] == 3) mxf->op = OP1c;
    else if (op[12] == 2 && op[13] == 1) mxf->op = OP2a;
    else if (op[12] == 2 && op[13] == 2) mxf->op = OP2b;
    else if (op[12] == 2 && op[13] == 3) mxf->op = OP2c;
    else if (op[12] == 3 && op[13] == 1) mxf->op = OP3a;
    else if (op[12] == 3 && op[13] == 2) mxf->op = OP3b;
    else if (op[12] == 3 && op[13] == 3) mxf->op = OP3c;
    else if (op[12] == 64&& op[13] == 1) mxf->op = OPSonyOpt;
    else if (op[12] == 0x10) {
        /* SMPTE 390m: "There shall be exactly one essence container"
         * The following block deals with files that violate this, namely:
         * 2011_DCPTEST_24FPS.V.mxf - two ECs, OP1a
         * abcdefghiv016f56415e.mxf - zero ECs, OPAtom, output by Avid AirSpeed */
        if (nb_essence_containers != 1) {
            MXFOP op = nb_essence_containers ? OP1a : OPAtom;

            /* only nag once */
            if (!mxf->op)
                av_log(mxf->fc, AV_LOG_WARNING,
                       "\"OPAtom\" with %u ECs - assuming %s\n",
                       nb_essence_containers,
                       op == OP1a ? "OP1a" : "OPAtom");

            mxf->op = op;
        } else
            mxf->op = OPAtom;
    } else {
        av_log(mxf->fc, AV_LOG_ERROR, "unknown operational pattern: %02xh %02xh - guessing OP1a\n", op[12], op[13]);
        mxf->op = OP1a;
    }

    if (partition->kag_size <= 0 || partition->kag_size > (1 << 20)) {
        av_log(mxf->fc, AV_LOG_WARNING, "invalid KAGSize %i - guessing ", partition->kag_size);

        if (mxf->op == OPSonyOpt)
            partition->kag_size = 512;
        else
            partition->kag_size = 1;

        av_log(mxf->fc, AV_LOG_WARNING, "%i\n", partition->kag_size);
    }

    return 0;
}

//
// JLB Added new method
//
static int mxf_read_random_index_pack(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFContext		*mxf = arg;
    int 			idx, idz;
    int				newPartitionCount;
	MXFPartition 	*newPartitions;
	uint32_t		bodySid;
	uint64_t		partitionOffset;

	newPartitionCount = (size - 4) / 12; // Subtract the last length (4 bytes)

	// Make sure we have enough space for all partitions
	newPartitions = av_malloc( newPartitionCount * sizeof( *mxf->partitions ) );
	memset(newPartitions, 0, newPartitionCount * sizeof( *mxf->partitions ));
	if (!newPartitions)
		return AVERROR(ENOMEM);

	av_log(mxf->fc, AV_LOG_INFO, "RIP, Found %d items\n", newPartitionCount);
    for ( idx = 0; idx < newPartitionCount; idx++ ) {
    	bodySid = avio_rb32(pb);
    	partitionOffset = avio_rb64(pb);

    	// Skip the header since it is probably already filled in
		newPartitions[idx].body_sid = bodySid;
		newPartitions[idx].this_partition = partitionOffset;
		newPartitions[idx].index_sid = 0;
		newPartitions[idx].essence_length = 0;
		newPartitions[idx].essence_offset = partitionOffset + size; // JLB, It could actually be little bit further, but this will be filled in when reading the partition pack header
		newPartitions[idx].header_byte_count = 0;
		newPartitions[idx].index_byte_count = 0;
		newPartitions[idx].pack_length = -1;
		newPartitions[idx].kag_size = 0;
		newPartitions[idx].type = BodyPartition; // Probably a body partition

		if (idx > 0)
			newPartitions[idx].previous_partition = newPartitions[idx - 1].this_partition;
		//av_log(mxf->fc, AV_LOG_INFO, "- RIP, partition %d, offset %ld\n", newPartitions[idx].index_sid, newPartitions[idx].this_partition);
    }

    // Overwrite all partitions that have been read into our new table
    for( idx = 0; idx < mxf->partitions_count; idx++  )    {
    	for ( idz = 0; idz < newPartitionCount; idz++ ){
    		if ( mxf->partitions[idx].this_partition == newPartitions[idz].this_partition){
    		    memcpy( &(newPartitions[idz]), &(mxf->partitions[idx]), sizeof(MXFPartition));
    		}
    	}
    }

    // Let the partitions point to our new partitions table
    av_freep(&mxf->partitions);
	mxf->partitions = newPartitions;
	mxf->partitions_count = newPartitionCount;

    return 0;
}

static int mxf_read_primer_pack(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFContext *mxf = arg;
    int item_num = avio_rb32(pb);
    int item_len = avio_rb32(pb);

    if (item_len != 18) {
        av_log_ask_for_sample(pb, "unsupported primer pack item length %d\n",
                              item_len);
        return AVERROR_PATCHWELCOME;
    }
    if (item_num > UINT_MAX / item_len)
        return AVERROR_INVALIDDATA;
    mxf->local_tags_count = item_num;
    mxf->local_tags = av_malloc(item_num*item_len);
    if (!mxf->local_tags)
        return AVERROR(ENOMEM);
    avio_read(pb, mxf->local_tags, item_num*item_len);
    return 0;
}


/*
 * JLB Test if there is a RIP present
 */
static int mxf_locate_RIP_gv(MXFContext* mxf, KLVPacket* klv)
{
	static unsigned char RIP_key[16] = { 0x06, 0x0e, 0x2b, 0x34, 0x02, 0x05, 0x01, 0x01, 0x0d, 0x01, 0x02, 0x01, 0x01, 0x11, 0x01, 0x00 };
	AVIOContext*	pb = mxf->fc->pb;
	int64_t			pos;
	int64_t 		fileSizeBt;
	int64_t			ripSizeBt;

	/* -------> Get and check file size. If file is less than 20 bytes long, give up right now. */
	fileSizeBt = avio_size( pb );
	if ( fileSizeBt < 20 ) {
		return AVERROR_INVALIDDATA;
	}

	/* -------> Seek back 4 bytes to read the 32-bit big-endion RIP (Random Index Pack)
				object size, and do a sanity check on the result. This object is not very
				large, think 60 bytes. It contains one or more partition offsets, and one
				of them should be the index body partition, holding the index tables. */
	pos = fileSizeBt - 4;
	avio_seek( pb, pos, SEEK_SET );
	ripSizeBt = avio_rb32( pb );
	if ( ripSizeBt > fileSizeBt ) {
		return AVERROR_INVALIDDATA;
	}

	/* -------> Seek back to the start of the assumed RIP object in the file. */
	pos = fileSizeBt - ripSizeBt;
	avio_seek( pb, pos, SEEK_SET );


	/* -------> Read in the KLV header and check if we have the RIP key.
				If not, this MFX file probably has the RIP stored somewhere else, and we
				give up here. */
	if (klv_read_packet(klv, pb) < 0){
		return AVERROR_INVALIDDATA;
	}
	if (!IS_KLV_KEY(klv->key, RIP_key)){
		return AVERROR_INVALIDDATA;
	}
	return 1; // Found
}

/*
 * JLB Read RIP packet from the end of the file (if present)
 */
static int mxf_read_RIP_gv(MXFContext* mxf)
{
	AVIOContext*	pb = mxf->fc->pb;
	KLVPacket		klv;

	if (mxf_locate_RIP_gv(mxf, &klv) >= 0) {
		// Yes, RIP found
		av_log(mxf->fc, AV_LOG_INFO, "READ RIP table near end of file....\n" );

		/* -------> We found the RIP object. Decode the BER section. This gives us the length
					of the RIP object. Again, think of a small value. */
		mxf_read_random_index_pack(mxf, pb, 0, klv.length, klv.key, klv.offset);
	}
	else{
		av_log(mxf->fc, AV_LOG_WARNING, "No RIP table found. Searches might be slow!\n" );
	}
	return 0;
}


static int mxf_add_metadata_set(MXFContext *mxf, void *metadata_set)
{
    MXFMetadataSet **tmp;
    if (mxf->metadata_sets_count+1 >= UINT_MAX / sizeof(*mxf->metadata_sets))
        return AVERROR(ENOMEM);
    tmp = av_realloc(mxf->metadata_sets, (mxf->metadata_sets_count + 1) * sizeof(*mxf->metadata_sets));
    if (!tmp)
        return AVERROR(ENOMEM);
    mxf->metadata_sets = tmp;
    mxf->metadata_sets[mxf->metadata_sets_count] = metadata_set;
    mxf->metadata_sets_count++;
    return 0;
}

static int mxf_add_timecode_metadata(AVDictionary **pm, const char *key, AVTimecode *tc)
{
    char buf[AV_TIMECODE_STR_SIZE];
    av_dict_set(pm, key, av_timecode_make_string(tc, buf, 0), 0);

    return 0;
}

static int mxf_read_cryptographic_context(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFCryptoContext *cryptocontext = arg;
    if (size != 16)
        return AVERROR_INVALIDDATA;
    if (IS_KLV_KEY(uid, mxf_crypto_source_container_ul))
        avio_read(pb, cryptocontext->source_container_ul, 16);
    return 0;
}

static int mxf_read_content_storage(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFContext *mxf = arg;
    switch (tag) {
    case 0x1901:
        mxf->packages_count = avio_rb32(pb);
        if (mxf->packages_count >= UINT_MAX / sizeof(UID))
            return AVERROR_INVALIDDATA;
        mxf->packages_refs = av_malloc(mxf->packages_count * sizeof(UID));
        if (!mxf->packages_refs)
            return AVERROR(ENOMEM);
        avio_skip(pb, 4); /* useless size of objects, always 16 according to specs */
        avio_read(pb, (uint8_t *)mxf->packages_refs, mxf->packages_count * sizeof(UID));
        break;
    }
    return 0;
}

static int mxf_read_source_clip(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFStructuralComponent *source_clip = arg;
    switch(tag) {
    case 0x0202:
        source_clip->duration = avio_rb64(pb);
        break;
    case 0x1201:
        source_clip->start_position = avio_rb64(pb);
        break;
    case 0x1101:
        /* UMID, only get last 16 bytes */
        avio_skip(pb, 16);
        avio_read(pb, source_clip->source_package_uid, 16);
        break;
    case 0x1102:
        source_clip->source_track_id = avio_rb32(pb);
        break;
    }
    return 0;
}

static int mxf_read_material_package(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFPackage *package = arg;
    switch(tag) {
    case 0x4403:
        package->tracks_count = avio_rb32(pb);
        if (package->tracks_count >= UINT_MAX / sizeof(UID))
            return AVERROR_INVALIDDATA;
        package->tracks_refs = av_malloc(package->tracks_count * sizeof(UID));
        if (!package->tracks_refs)
            return AVERROR(ENOMEM);
        avio_skip(pb, 4); /* useless size of objects, always 16 according to specs */
        avio_read(pb, (uint8_t *)package->tracks_refs, package->tracks_count * sizeof(UID));
        break;
    }
    return 0;
}

static int mxf_read_track(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFTrack *track = arg;
    switch(tag) {
    case 0x4801:
        track->track_id = avio_rb32(pb);
        break;
    case 0x4804:
        avio_read(pb, track->track_number, 4);
        break;
    case 0x4B01:
        track->edit_rate.num = avio_rb32(pb);
        track->edit_rate.den = avio_rb32(pb);
        break;
    case 0x4803:
        avio_read(pb, track->sequence_ref, 16);
        break;
    }
    return 0;
}

static int mxf_read_sequence(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFSequence *sequence = arg;
    switch(tag) {
    case 0x0202:
        sequence->duration = avio_rb64(pb);
        break;
    case 0x0201:
        avio_read(pb, sequence->data_definition_ul, 16);
        break;
    case 0x1001:
        sequence->structural_components_count = avio_rb32(pb);
        if (sequence->structural_components_count >= UINT_MAX / sizeof(UID))
            return AVERROR_INVALIDDATA;
        sequence->structural_components_refs = av_malloc(sequence->structural_components_count * sizeof(UID));
        if (!sequence->structural_components_refs)
            return AVERROR(ENOMEM);
        avio_skip(pb, 4); /* useless size of objects, always 16 according to specs */
        avio_read(pb, (uint8_t *)sequence->structural_components_refs, sequence->structural_components_count * sizeof(UID));
        break;
    }
    return 0;
}

static int mxf_read_source_package(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFPackage *package = arg;
    switch(tag) {
    case 0x4403:
        package->tracks_count = avio_rb32(pb);
        if (package->tracks_count >= UINT_MAX / sizeof(UID))
            return AVERROR_INVALIDDATA;
        package->tracks_refs = av_malloc(package->tracks_count * sizeof(UID));
        if (!package->tracks_refs)
            return AVERROR(ENOMEM);
        avio_skip(pb, 4); /* useless size of objects, always 16 according to specs */
        avio_read(pb, (uint8_t *)package->tracks_refs, package->tracks_count * sizeof(UID));
        break;
    case 0x4401:
        /* UMID, only get last 16 bytes */
        avio_skip(pb, 16);
        avio_read(pb, package->package_uid, 16);
        break;
    case 0x4701:
        avio_read(pb, package->descriptor_ref, 16);
        break;
    }
    return 0;
}

static int mxf_read_index_entry_array(AVIOContext *pb, MXFIndexTableSegment *segment)
{
    int i, length;

    segment->nb_index_entries = avio_rb32(pb);
    if (!segment->nb_index_entries)
        return 0;
    else if (segment->nb_index_entries < 0 ||
             segment->nb_index_entries >
             (INT_MAX / sizeof(*segment->stream_offset_entries)))
        return AVERROR(ENOMEM);

    length = avio_rb32(pb);

    segment->temporal_offset_entries = av_mallocz(segment->nb_index_entries * sizeof(*segment->temporal_offset_entries));
    segment->flag_entries            = av_mallocz(segment->nb_index_entries * sizeof(*segment->flag_entries));
    segment->stream_offset_entries   = av_mallocz(segment->nb_index_entries * sizeof(*segment->stream_offset_entries));

    // JLB Added
    segment->key_frame_offset_entries = av_mallocz(segment->nb_index_entries * sizeof(*segment->key_frame_offset_entries));

    if (!segment->flag_entries || !segment->stream_offset_entries ||
        !segment->temporal_offset_entries
        || !segment->key_frame_offset_entries
        ) {
        av_freep(&segment->flag_entries);
        av_freep(&segment->stream_offset_entries);
        av_freep(&segment->temporal_offset_entries);
        av_freep(&segment->key_frame_offset_entries);
        return AVERROR(ENOMEM);
    }

    for (i = 0; i < segment->nb_index_entries; i++) {
        segment->temporal_offset_entries[i] = avio_r8(pb);
        // JLB Added
        segment->key_frame_offset_entries[i] = avio_r8(pb); /* KeyFrameOffset */
        segment->flag_entries[i] = avio_r8(pb);
        segment->stream_offset_entries[i] = avio_rb64(pb);
        avio_skip(pb, length - 11);
    }
    return 0;
}


/*
 * JLB added delta entry array
 */
static int mxf_read_delta_entry_array(AVIOContext *pb, MXFIndexTableSegment *segment)
{
    segment->nb_delta_entries = avio_rb32(pb);
    if (!segment->nb_delta_entries)
        return 0;

    avio_rb32(pb);

    if ( segment->delta_pos_table_entries ) {
    	av_freep(&segment->delta_pos_table_entries);
    }

    segment->delta_pos_table_entries = av_mallocz(segment->nb_delta_entries * sizeof(*segment->delta_pos_table_entries));

    if (!segment->delta_pos_table_entries) {
        av_freep(&segment->delta_pos_table_entries);
        return AVERROR(ENOMEM);
    }

    for (int i = 0; i < segment->nb_delta_entries; i++) {
        segment->delta_pos_table_entries[i] = avio_r8(pb);
        avio_r8(pb);	/* Slice */
        avio_rb32(pb);	/* ElementDelta */
    }
    return 0;
}


static int mxf_read_index_table_segment(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFIndexTableSegment *segment = arg;

    switch(tag) {
    case 0x3F05:
        segment->edit_unit_byte_count = avio_rb32(pb);
        av_dlog(NULL, "EditUnitByteCount %d\n", segment->edit_unit_byte_count);
        break;
    case 0x3F06:
        segment->index_sid = avio_rb32(pb);
        av_dlog(NULL, "IndexSID %d\n", segment->index_sid);
        break;
    case 0x3F07:
        segment->body_sid = avio_rb32(pb);
        av_dlog(NULL, "BodySID %d\n", segment->body_sid);
        break;

    case 0x3F09: // JLB Added delta entry array
        //av_log(NULL, AV_LOG_INFO, "DeltaEntryArray found!\n");
    	return mxf_read_delta_entry_array(pb, segment);

    case 0x3F0A:
        av_dlog(NULL, "IndexEntryArray found\n");
        return mxf_read_index_entry_array(pb, segment);

    case 0x3F0B:
        segment->index_edit_rate.num = avio_rb32(pb);
        segment->index_edit_rate.den = avio_rb32(pb);
        av_dlog(NULL, "IndexEditRate %d/%d\n", segment->index_edit_rate.num,
                segment->index_edit_rate.den);
        break;
    case 0x3F0C:
        segment->index_start_position = avio_rb64(pb);
        av_dlog(NULL, "IndexStartPosition %"PRId64"\n", segment->index_start_position);
        break;
    case 0x3F0D:
        segment->index_duration = avio_rb64(pb);
        av_dlog(NULL, "IndexDuration %"PRId64"\n", segment->index_duration);
        break;
    }
    return 0;
}

static void mxf_read_pixel_layout(AVIOContext *pb, MXFDescriptor *descriptor)
{
    int code, value, ofs = 0;
    char layout[16] = {0};

    do {
        code = avio_r8(pb);
        value = avio_r8(pb);
        av_dlog(NULL, "pixel layout: code %#x\n", code);

        if (ofs < 16) {
            layout[ofs++] = code;
            layout[ofs++] = value;
        }
    } while (code != 0); /* SMPTE 377M E.2.46 */

    ff_mxf_decode_pixel_layout(layout, &descriptor->pix_fmt);
}

static int mxf_read_generic_descriptor(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFDescriptor *descriptor = arg;
    descriptor->pix_fmt = AV_PIX_FMT_NONE;
    switch(tag) {
    case 0x3F01:
        descriptor->sub_descriptors_count = avio_rb32(pb);
        if (descriptor->sub_descriptors_count >= UINT_MAX / sizeof(UID))
            return AVERROR_INVALIDDATA;
        descriptor->sub_descriptors_refs = av_malloc(descriptor->sub_descriptors_count * sizeof(UID));
        if (!descriptor->sub_descriptors_refs)
            return AVERROR(ENOMEM);
        avio_skip(pb, 4); /* useless size of objects, always 16 according to specs */
        avio_read(pb, (uint8_t *)descriptor->sub_descriptors_refs, descriptor->sub_descriptors_count * sizeof(UID));
        break;
    case 0x3004:
        avio_read(pb, descriptor->essence_container_ul, 16);
        break;
    case 0x3006:
        descriptor->linked_track_id = avio_rb32(pb);
        break;
    case 0x3201: /* PictureEssenceCoding */
        avio_read(pb, descriptor->essence_codec_ul, 16);
        break;
    case 0x3203:
        descriptor->width = avio_rb32(pb);
        break;
    case 0x3202:
        descriptor->height = avio_rb32(pb);
        break;
    case 0x320C:
        descriptor->frame_layout = avio_r8(pb);
        break;
    case 0x320E:
        descriptor->aspect_ratio.num = avio_rb32(pb);
        descriptor->aspect_ratio.den = avio_rb32(pb);
        break;
    case 0x3301:
        descriptor->component_depth = avio_rb32(pb);
        break;
    case 0x3302:
        descriptor->horiz_subsampling = avio_rb32(pb);
        break;
    case 0x3308:
        descriptor->vert_subsampling = avio_rb32(pb);
        break;
    case 0x3D03:
        descriptor->sample_rate.num = avio_rb32(pb);
        descriptor->sample_rate.den = avio_rb32(pb);
        break;
    case 0x3D06: /* SoundEssenceCompression */
        avio_read(pb, descriptor->essence_codec_ul, 16);
        break;
    case 0x3D07:
        descriptor->channels = avio_rb32(pb);
        break;
    case 0x3D01:
        descriptor->bits_per_sample = avio_rb32(pb);
        break;
    case 0x3401:
        mxf_read_pixel_layout(pb, descriptor);
        break;
    case 0x3218: // Added for active format description
		descriptor->active_format = avio_r8(pb);
		break;

    default:
        /* Private uid used by SONY C0023S01.mxf */
        if (IS_KLV_KEY(uid, mxf_sony_mpeg4_extradata)) {
            descriptor->extradata = av_malloc(size + FF_INPUT_BUFFER_PADDING_SIZE);
            if (!descriptor->extradata)
                return AVERROR(ENOMEM);
            descriptor->extradata_size = size;
            avio_read(pb, descriptor->extradata, size);
        }
        break;
    }
    return 0;
}


/*
 * JLB Added reading method to read the time code descriptor
 * (Required for absolute seek)
 */
static int mxf_read_timecode_descriptor_gv(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
    MXFTimecodeComponent *timecode = arg;
    switch(tag){
    case 0x0201:
    	avio_read(pb, timecode->data_definition, 16);
    	break;
	case 0x1501:
		timecode->start_time = avio_rb64(pb);
		break;
	case 0x1502:
		timecode->rounded_timecode_base = avio_rb16(pb);
		break;
	case 0x1503:
		timecode->drop_frame = avio_r8(pb);
		break;
	}
    return 0;
}

/*
 * JLB Read the essence container data, it defines the link between the index sid and the package
 */
static int mxf_read_essence_container_data_gv(void *arg, AVIOContext *pb, int tag, int size, UID uid, int64_t klv_offset)
{
	MXFEssenceContainerData *ecd = arg;
	switch(tag){
	case 0x2701:
        avio_read(pb, ecd->linked_package_id, 16);
		break;
	case 0x3F06:
		ecd->index_sid = avio_rb32(pb);
		break;
	case 0x3F07:
		ecd->body_sid = avio_rb32(pb);
		break;
	}
	return 0;
}


/*
 * Match an uid independently of the version byte and up to len common bytes
 * Returns: boolean
 */
static int mxf_match_uid(const UID key, const UID uid, int len)
{
    int i;
    for (i = 0; i < len; i++) {
        if (i != 7 && key[i] != uid[i])
            return 0;
    }
    return 1;
}

static const MXFCodecUL *mxf_get_codec_ul(const MXFCodecUL *uls, UID *uid)
{
    while (uls->uid[0]) {
        if(mxf_match_uid(uls->uid, *uid, uls->matching_len))
            break;
        uls++;
    }
    return uls;
}

static void *mxf_resolve_strong_ref(MXFContext *mxf, UID *strong_ref, enum MXFMetadataSetType type)
{
    int i;

    if (!strong_ref)
        return NULL;
    for (i = 0; i < mxf->metadata_sets_count; i++) {
        if (!memcmp(*strong_ref, mxf->metadata_sets[i]->uid, 16) &&
            (type == AnyType || mxf->metadata_sets[i]->type == type)) {
            return mxf->metadata_sets[i];
        }
    }
    return NULL;
}

static const MXFCodecUL mxf_picture_essence_container_uls[] = {
    // video essence container uls
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x02,0x0D,0x01,0x03,0x01,0x02,0x04,0x60,0x01 }, 14, AV_CODEC_ID_MPEG2VIDEO }, /* MPEG-ES Frame wrapped */
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x01,0x0D,0x01,0x03,0x01,0x02,0x02,0x41,0x01 }, 14,    AV_CODEC_ID_DVVIDEO }, /* DV 625 25mbps */
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x01,0x0D,0x01,0x03,0x01,0x02,0x05,0x00,0x00 }, 14,   AV_CODEC_ID_RAWVIDEO }, /* Uncompressed Picture */
    { { 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 },  0,      AV_CODEC_ID_NONE },
};

/* EC ULs for intra-only formats */
static const MXFCodecUL mxf_intra_only_essence_container_uls[] = {
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x01,0x0D,0x01,0x03,0x01,0x02,0x01,0x00,0x00 }, 14, AV_CODEC_ID_MPEG2VIDEO }, /* MXF-GC SMPTE D-10 Mappings */
    { { 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 },  0,       AV_CODEC_ID_NONE },
};

/* intra-only PictureEssenceCoding ULs, where no corresponding EC UL exists */
static const MXFCodecUL mxf_intra_only_picture_essence_coding_uls[] = {
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x0A,0x04,0x01,0x02,0x02,0x01,0x32,0x00,0x00 }, 14,       AV_CODEC_ID_H264 }, /* H.264/MPEG-4 AVC Intra Profiles */
    { { 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 },  0,       AV_CODEC_ID_NONE },
};

static const MXFCodecUL mxf_sound_essence_container_uls[] = {
    // sound essence container uls
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x01,0x0D,0x01,0x03,0x01,0x02,0x06,0x01,0x00 }, 14, AV_CODEC_ID_PCM_S16LE }, /* BWF Frame wrapped */
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x02,0x0D,0x01,0x03,0x01,0x02,0x04,0x40,0x01 }, 14,       AV_CODEC_ID_MP2 }, /* MPEG-ES Frame wrapped, 0x40 ??? stream id */
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x01,0x0D,0x01,0x03,0x01,0x02,0x01,0x05,0x01 }, 14, AV_CODEC_ID_PCM_S16LE }, /* D-10 Mapping 30Mbps PAL Extended Template */
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x01,0x0D,0x01,0x03,0x01,0x02,0x01,0x03,0x01 }, 14, AV_CODEC_ID_PCM_S16LE }, /* D-10 Mapping 40Mbps PAL Extended Template */
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x01,0x0D,0x01,0x03,0x01,0x02,0x01,0x01,0x01 }, 14, AV_CODEC_ID_PCM_S16LE }, /* D-10 Mapping 50Mbps PAL Extended Template */
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x01,0x0D,0x01,0x03,0x01,0x02,0x01,0x06,0x01 }, 14, AV_CODEC_ID_PCM_S16LE }, /* D-10 Mapping 30Mbps NTSC Extended Template */
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x01,0x0D,0x01,0x03,0x01,0x02,0x01,0x04,0x01 }, 14, AV_CODEC_ID_PCM_S16LE }, /* D-10 Mapping 40Mbps NTSC Extended Template */
    { { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x01,0x0D,0x01,0x03,0x01,0x02,0x01,0x02,0x01 }, 14, AV_CODEC_ID_PCM_S16LE }, /* D-10 Mapping 50Mbps NTSC Extended Template */
    { { 0x06,0x0E,0x2B,0x34,0x01,0x01,0x01,0xFF,0x4B,0x46,0x41,0x41,0x00,0x0D,0x4D,0x4F }, 14, AV_CODEC_ID_PCM_S16LE }, /* 0001GL00.MXF.A1.mxf_opatom.mxf */
    { { 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 },  0,      AV_CODEC_ID_NONE },
};

/* JLB Added the following block */
static const MXFCodecUL mxf_data_container_uls[] = {
    // data essence container uls
	{ { 0x06,0x0E,0x2B,0x34,0x04,0x01,0x01,0x09,0x0D,0x01,0x03,0x01,0x02,0x0E,0x00,0x00 }, 14,    AV_CODEC_ID_MXFANC }, /* MXF-GC Generic ANC Data Mapping Undefined Payload */
    { { 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 },  0,      AV_CODEC_ID_NONE },
};


/**
 * Computes the absolute file offset of the given essence container offset
 */
static int mxf_absolute_bodysid_offset(MXFContext *mxf, int body_sid, int64_t offset, int64_t *offset_out)
{
    int x;
    int64_t offset_in = offset;     /* for logging */

    for (x = 0; x < mxf->partitions_count; x++) {
        MXFPartition *p = &mxf->partitions[x];

        if (p->body_sid != body_sid)
            continue;

        if (offset < p->essence_length || !p->essence_length) {
            *offset_out = p->essence_offset + offset;
            return 0;
        }

        offset -= p->essence_length;
    }

    av_log(mxf->fc, AV_LOG_ERROR,
           "failed to find absolute offset of %"PRIX64" in BodySID %i - partial file?\n",
           offset_in, body_sid);

    return AVERROR_INVALIDDATA;
}

/**
 * Returns the end position of the essence container with given BodySID, or zero if unknown
 */
//static int64_t mxf_essence_container_end(MXFContext *mxf, int body_sid)
//{
//    int x;
//    int64_t ret = 0;
//
//    for (x = 0; x < mxf->partitions_count; x++) {
//        MXFPartition *p = &mxf->partitions[x];
//
//        if (p->body_sid != body_sid)
//            continue;
//
//        if (!p->essence_length)
//            return 0;
//
//        ret = p->essence_offset + p->essence_length;
//    }
//
//    return ret;
//}

/* EditUnit -> absolute offset */
static int mxf_edit_unit_absolute_offset(MXFContext *mxf, MXFIndexTable *index_table, int64_t edit_unit, int64_t *edit_unit_out, int64_t *offset_out, int nag)
{
    int i;
    int64_t offset_temp = 0;

    for (i = 0; i < index_table->nb_segments; i++) {
        MXFIndexTableSegment *s = index_table->segments[i];

        edit_unit = FFMAX(edit_unit, s->index_start_position);  /* clamp if trying to seek before start */

        if (edit_unit < s->index_start_position + s->index_duration) {
            int64_t index = edit_unit - s->index_start_position;

            if (s->edit_unit_byte_count)
                offset_temp += s->edit_unit_byte_count * index;
            else if (s->nb_index_entries) {
                if (s->nb_index_entries == 2 * s->index_duration + 1)
                    index *= 2;     /* Avid index */

                if (index < 0 || index >= s->nb_index_entries) {
                    av_log(mxf->fc, AV_LOG_ERROR, "IndexSID %i segment at %"PRId64" IndexEntryArray too small\n",
                           index_table->index_sid, s->index_start_position);
                    return AVERROR_INVALIDDATA;
                }

                offset_temp = s->stream_offset_entries[index];
            } else {
                av_log(mxf->fc, AV_LOG_ERROR, "IndexSID %i segment at %"PRId64" missing EditUnitByteCount and IndexEntryArray\n",
                       index_table->index_sid, s->index_start_position);
                return AVERROR_INVALIDDATA;
            }

            if (edit_unit_out)
                *edit_unit_out = edit_unit;

            return mxf_absolute_bodysid_offset(mxf, index_table->body_sid, offset_temp, offset_out);
        } else {
            /* EditUnitByteCount == 0 for VBR indexes, which is fine since they use explicit StreamOffsets */
            offset_temp += s->edit_unit_byte_count * s->index_duration;
        }
    }

    if (nag)
        av_log(mxf->fc, AV_LOG_ERROR, "failed to map EditUnit %"PRId64" in IndexSID %i to an offset\n", edit_unit, index_table->index_sid);

    return AVERROR_INVALIDDATA;
}


static int mxf_is_intra_only(MXFDescriptor *d)
{
    return mxf_get_codec_ul(mxf_intra_only_essence_container_uls,
                            &d->essence_container_ul)->id != AV_CODEC_ID_NONE ||
           mxf_get_codec_ul(mxf_intra_only_picture_essence_coding_uls,
                            &d->essence_codec_ul)->id     != AV_CODEC_ID_NONE;
}

static int mxf_parse_structural_metadata(MXFContext *mxf)
{
    MXFPackage *material_package = NULL;
    MXFPackage *temp_package = NULL;
    int i, j, k, ret;

    av_dlog(mxf->fc, "metadata sets count %d\n", mxf->metadata_sets_count);
    /* TODO: handle multiple material packages (OP3x) */
    for (i = 0; i < mxf->packages_count; i++) {
        material_package = mxf_resolve_strong_ref(mxf, &mxf->packages_refs[i], MaterialPackage);
        if (material_package) break;
    }
    if (!material_package) {
        av_log(mxf->fc, AV_LOG_ERROR, "no material package found\n");
        return AVERROR_INVALIDDATA;
    }

    for (i = 0; i < material_package->tracks_count; i++) {
        MXFPackage *source_package = NULL;
        MXFTrack *material_track = NULL;
        MXFTrack *source_track = NULL;
        MXFTrack *temp_track = NULL;
        MXFDescriptor *descriptor = NULL;
        MXFStructuralComponent *component = NULL;
        MXFTimecodeComponent *tccomp = NULL;
        UID *essence_container_ul = NULL;
        const MXFCodecUL *codec_ul = NULL;
        const MXFCodecUL *container_ul = NULL;
        const MXFCodecUL *pix_fmt_ul = NULL;
        AVStream *st;
        AVTimecode tc;
        int flags;

        if (!(material_track = mxf_resolve_strong_ref(mxf, &material_package->tracks_refs[i], Track))) {
            av_log(mxf->fc, AV_LOG_ERROR, "could not resolve material track strong ref\n");
            continue;
        }

        if ((component = mxf_resolve_strong_ref(mxf, &material_track->sequence_ref, TimecodeComponent))) {
        	tccomp = (MXFTimecodeComponent*)component;

            /* -------> Store the start time in the MXFTrack. */
			mxf->overall_start_time_ms = av_rescale(tccomp->start_time * 1000, 1000 + tccomp->drop_frame, tccomp->rounded_timecode_base * 1000);

            flags = tccomp->drop_frame == 1 ? AV_TIMECODE_FLAG_DROPFRAME : 0;
            if (av_timecode_init(&tc, (AVRational){tccomp->rounded_timecode_base, 1}, flags, tccomp->start_time, mxf->fc) == 0) {
                mxf_add_timecode_metadata(&mxf->fc->metadata, "timecode", &tc);
            }
        }

        if (!(material_track->sequence = mxf_resolve_strong_ref(mxf, &material_track->sequence_ref, Sequence))) {
            av_log(mxf->fc, AV_LOG_ERROR, "could not resolve material track sequence strong ref\n");
            continue;
        }

        /* -------> JLB Get the time code component belonging to this track
        			NOTE: According to the MXF specifications: 9.5.3 point 7: 'The value of the Origin Property of all Essence Tracks and all Timecode Tracks in a Material Package shall be zero.'
        			So (I assume) the timecode component offset should account for all time code tracks:
        			Thus: I use the Timecode component of the Timecode track for all tracks in this package. */

        for (j = 0; j < material_track->sequence->structural_components_count; j++) {
            component = mxf_resolve_strong_ref(mxf, &material_track->sequence->structural_components_refs[j], TimecodeComponent);
            if (!component)
                continue;

            tccomp = (MXFTimecodeComponent*)component;

            /* -------> Store the start time in the MXFTrack. */
			mxf->overall_start_time_ms = av_rescale(tccomp->start_time * 1000, 1000 + tccomp->drop_frame, tccomp->rounded_timecode_base * 1000);

            flags = tccomp->drop_frame == 1 ? AV_TIMECODE_FLAG_DROPFRAME : 0;
            if (av_timecode_init(&tc, (AVRational){tccomp->rounded_timecode_base, 1}, flags, tccomp->start_time, mxf->fc) == 0) {
                mxf_add_timecode_metadata(&mxf->fc->metadata, "timecode", &tc);
                break;
            }
        }

        /* TODO: handle multiple source clips */
        for (j = 0; j < material_track->sequence->structural_components_count; j++) {

            component = mxf_resolve_strong_ref(mxf, &material_track->sequence->structural_components_refs[j], SourceClip);
            if (!component)
                continue;

            for (k = 0; k < mxf->packages_count; k++) {
                temp_package = mxf_resolve_strong_ref(mxf, &mxf->packages_refs[k], SourcePackage);
                if (!temp_package)
                    continue;
                if (!memcmp(temp_package->package_uid, component->source_package_uid, 16)) {
                    source_package = temp_package;
                    break;
                }
            }
            if (!source_package) {
                av_dlog(mxf->fc, "material track %d: no corresponding source package found\n", material_track->track_id);
                break;
            }
            for (k = 0; k < source_package->tracks_count; k++) {
                if (!(temp_track = mxf_resolve_strong_ref(mxf, &source_package->tracks_refs[k], Track))) {
                    av_log(mxf->fc, AV_LOG_ERROR, "could not resolve source track strong ref\n");
                    ret = AVERROR_INVALIDDATA;
                    goto fail_and_free;
                }
                if (temp_track->track_id == component->source_track_id) {
                    source_track = temp_track;
                    break;
                }
            }
            if (!source_track) {
                av_log(mxf->fc, AV_LOG_ERROR, "material track %d: no corresponding source track found\n", material_track->track_id);
                break;
            }
        }
        if (!source_track || !component)
            continue;

        if (!(source_track->sequence = mxf_resolve_strong_ref(mxf, &source_track->sequence_ref, Sequence))) {
            av_log(mxf->fc, AV_LOG_ERROR, "could not resolve source track sequence strong ref\n");
            ret = AVERROR_INVALIDDATA;
            goto fail_and_free;
        }

        /* 0001GL00.MXF.A1.mxf_opatom.mxf has the same SourcePackageID as 0001GL.MXF.V1.mxf_opatom.mxf
         * This would result in both files appearing to have two streams. Work around this by sanity checking DataDefinition */
        if (memcmp(material_track->sequence->data_definition_ul, source_track->sequence->data_definition_ul, 16)) {
            av_log(mxf->fc, AV_LOG_ERROR, "material track %d: DataDefinition mismatch\n", material_track->track_id);
            continue;
        }

        st = avformat_new_stream(mxf->fc, NULL);
        if (!st) {
            av_log(mxf->fc, AV_LOG_ERROR, "could not allocate stream\n");
            ret = AVERROR(ENOMEM);
            goto fail_and_free;
        }
        st->id = source_track->track_id;
        st->priv_data = source_track;
        st->duration = component->duration;
        if (st->duration == -1)
            st->duration = AV_NOPTS_VALUE;

        st->start_time = 0; // JLB All start times are zero

        avpriv_set_pts_info(st, 64, material_track->edit_rate.den, material_track->edit_rate.num);

        PRINT_KEY(mxf->fc, "data definition   ul", source_track->sequence->data_definition_ul);
        codec_ul = mxf_get_codec_ul(ff_mxf_data_definition_uls, &source_track->sequence->data_definition_ul);
        st->codec->codec_type = codec_ul->id;

        source_package->descriptor = mxf_resolve_strong_ref(mxf, &source_package->descriptor_ref, AnyType);
        if (source_package->descriptor) {
            if (source_package->descriptor->type == MultipleDescriptor) {
                for (j = 0; j < source_package->descriptor->sub_descriptors_count; j++) {
                    MXFDescriptor *sub_descriptor = mxf_resolve_strong_ref(mxf, &source_package->descriptor->sub_descriptors_refs[j], Descriptor);

                    if (!sub_descriptor) {
                        av_log(mxf->fc, AV_LOG_ERROR, "could not resolve sub descriptor strong ref\n");
                        continue;
                    }
                    if (sub_descriptor->linked_track_id == source_track->track_id) {
                        descriptor = sub_descriptor;
                        break;
                    }
                }
            } else if (source_package->descriptor->type == Descriptor)
                descriptor = source_package->descriptor;
        }
        if (!descriptor) {
            av_log(mxf->fc, AV_LOG_INFO, "source track %d: stream %d, no descriptor found\n", source_track->track_id, st->index);
            continue;
        }
        PRINT_KEY(mxf->fc, "essence codec     ul", descriptor->essence_codec_ul);
        PRINT_KEY(mxf->fc, "essence container ul", descriptor->essence_container_ul);
        essence_container_ul = &descriptor->essence_container_ul;
        /* HACK: replacing the original key with mxf_encrypted_essence_container
         * is not allowed according to s429-6, try to find correct information anyway */
        if (IS_KLV_KEY(essence_container_ul, mxf_encrypted_essence_container)) {
            av_log(mxf->fc, AV_LOG_INFO, "broken encrypted mxf file\n");
            for (k = 0; k < mxf->metadata_sets_count; k++) {
                MXFMetadataSet *metadata = mxf->metadata_sets[k];
                if (metadata->type == CryptoContext) {
                    essence_container_ul = &((MXFCryptoContext *)metadata)->source_container_ul;
                    break;
                }
            }
        }

        /* TODO: drop PictureEssenceCoding and SoundEssenceCompression, only check EssenceContainer */
        codec_ul = mxf_get_codec_ul(ff_mxf_codec_uls, &descriptor->essence_codec_ul);
        st->codec->codec_id = codec_ul->id;
        if (descriptor->extradata) {
            st->codec->extradata = descriptor->extradata;
            st->codec->extradata_size = descriptor->extradata_size;
        }
        if (st->codec->codec_type == AVMEDIA_TYPE_VIDEO) {
            source_track->intra_only = mxf_is_intra_only(descriptor);
            container_ul = mxf_get_codec_ul(mxf_picture_essence_container_uls, essence_container_ul);
            if (st->codec->codec_id == AV_CODEC_ID_NONE)
                st->codec->codec_id = container_ul->id;
            st->codec->width = descriptor->width;

            // JLB Added
            if (st->codec->dtg_active_format == 0)
            	st->codec->dtg_active_format = descriptor->active_format;

            /* Field height, not frame height */
            st->codec->height = descriptor->height;
            switch (descriptor->frame_layout) {
                case SegmentedFrame:
                    /* This one is a weird layout I don't fully understand. */
                    av_log(mxf->fc, AV_LOG_INFO,
                           "SegmentedFrame layout isn't currently supported\n");
                    break;
                case FullFrame:
					st->codec->field_order = AV_FIELD_PROGRESSIVE;
                    break;
                case OneField:
                    /* Every other line is stored and needs to be duplicated. */
                    av_log(mxf->fc, AV_LOG_INFO,
                           "OneField frame layout isn't currently supported\n");
                    break;
                    /* The correct thing to do here is fall through, but by
                     * breaking we might be able to decode some streams at half
                     * the vertical resolution, rather than not al all.
                     * It's also for compatibility with the old behavior. */
                case SeparateFields:
                case MixedFields:
                    /* Turn field height into frame height. */
                    st->codec->height *= 2;
                default:
                    av_log(mxf->fc, AV_LOG_INFO,
                           "Unknown frame layout type: %d\n",
                           descriptor->frame_layout);
            }
            if (st->codec->codec_id == AV_CODEC_ID_RAWVIDEO) {
                st->codec->pix_fmt = descriptor->pix_fmt;
                if (st->codec->pix_fmt == AV_PIX_FMT_NONE) {
                    pix_fmt_ul = mxf_get_codec_ul(ff_mxf_pixel_format_uls,
                                                  &descriptor->essence_codec_ul);
                    st->codec->pix_fmt = pix_fmt_ul->id;
                    if (st->codec->pix_fmt == AV_PIX_FMT_NONE) {
                        /* support files created before RP224v10 by defaulting to UYVY422
                           if subsampling is 4:2:2 and component depth is 8-bit */
                        if (descriptor->horiz_subsampling == 2 &&
                            descriptor->vert_subsampling == 1 &&
                            descriptor->component_depth == 8) {
                            st->codec->pix_fmt = AV_PIX_FMT_UYVY422;
                        }
                    }
                }
            }
            st->need_parsing = AVSTREAM_PARSE_HEADERS;
        } else if (st->codec->codec_type == AVMEDIA_TYPE_AUDIO) {
            container_ul = mxf_get_codec_ul(mxf_sound_essence_container_uls, essence_container_ul);
            if (st->codec->codec_id == AV_CODEC_ID_NONE)
                st->codec->codec_id = container_ul->id;
            st->codec->channels = descriptor->channels;
            st->codec->bits_per_coded_sample = descriptor->bits_per_sample;

            if (descriptor->sample_rate.den > 0) {
                st->codec->sample_rate = descriptor->sample_rate.num / descriptor->sample_rate.den;
                avpriv_set_pts_info(st, 64, descriptor->sample_rate.den, descriptor->sample_rate.num);
            } else {
                av_log(mxf->fc, AV_LOG_WARNING, "invalid sample rate (%d/%d) "
                       "found for stream #%d, time base forced to 1/48000\n",
                       descriptor->sample_rate.num, descriptor->sample_rate.den,
                       st->index);
                avpriv_set_pts_info(st, 64, 1, 48000);
            }

            /* TODO: implement AV_CODEC_ID_RAWAUDIO */
            if (st->codec->codec_id == AV_CODEC_ID_PCM_S16LE) {
                if (descriptor->bits_per_sample > 16 && descriptor->bits_per_sample <= 24)
                    st->codec->codec_id = AV_CODEC_ID_PCM_S24LE;
                else if (descriptor->bits_per_sample == 32)
                    st->codec->codec_id = AV_CODEC_ID_PCM_S32LE;
            } else if (st->codec->codec_id == AV_CODEC_ID_PCM_S16BE) {
                if (descriptor->bits_per_sample > 16 && descriptor->bits_per_sample <= 24)
                    st->codec->codec_id = AV_CODEC_ID_PCM_S24BE;
                else if (descriptor->bits_per_sample == 32)
                    st->codec->codec_id = AV_CODEC_ID_PCM_S32BE;
            } else if (st->codec->codec_id == AV_CODEC_ID_MP2) {
                st->need_parsing = AVSTREAM_PARSE_FULL;
            }
        }
        else if (st->codec->codec_type == AVMEDIA_TYPE_DATA ) {
        	/* JLB: Added extra check to parse the data containers */
        	container_ul = mxf_get_codec_ul(mxf_data_container_uls, essence_container_ul);
			if (st->codec->codec_id == AV_CODEC_ID_NONE) {
				st->codec->codec_id = container_ul->id;
			}
        }

        if (st->codec->codec_type != AVMEDIA_TYPE_DATA && (*essence_container_ul)[15] > 0x01) {
            /* TODO: decode timestamps */
            st->need_parsing = AVSTREAM_PARSE_TIMESTAMPS;
        }
    }

    ret = 0;
fail_and_free:
    return ret;
}

static const MXFMetadataReadTableEntry mxf_metadata_read_table[] = {
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x05,0x01,0x00 }, mxf_read_primer_pack },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x02,0x01,0x00 }, mxf_read_partition_pack },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x02,0x02,0x00 }, mxf_read_partition_pack },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x02,0x03,0x00 }, mxf_read_partition_pack },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x02,0x04,0x00 }, mxf_read_partition_pack },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x03,0x01,0x00 }, mxf_read_partition_pack },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x03,0x02,0x00 }, mxf_read_partition_pack },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x03,0x03,0x00 }, mxf_read_partition_pack },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x03,0x04,0x00 }, mxf_read_partition_pack },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x04,0x02,0x00 }, mxf_read_partition_pack },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x05,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x04,0x04,0x00 }, mxf_read_partition_pack },

    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x18,0x00 }, mxf_read_content_storage, 0, AnyType },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x37,0x00 }, mxf_read_source_package, sizeof(MXFPackage), SourcePackage },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x36,0x00 }, mxf_read_material_package, sizeof(MXFPackage), MaterialPackage },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x0F,0x00 }, mxf_read_sequence, sizeof(MXFSequence), Sequence },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x11,0x00 }, mxf_read_source_clip, sizeof(MXFStructuralComponent), SourceClip },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x44,0x00 }, mxf_read_generic_descriptor, sizeof(MXFDescriptor), MultipleDescriptor },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x42,0x00 }, mxf_read_generic_descriptor, sizeof(MXFDescriptor), Descriptor }, /* Generic Sound */
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x28,0x00 }, mxf_read_generic_descriptor, sizeof(MXFDescriptor), Descriptor }, /* CDCI */
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x29,0x00 }, mxf_read_generic_descriptor, sizeof(MXFDescriptor), Descriptor }, /* RGBA */
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x51,0x00 }, mxf_read_generic_descriptor, sizeof(MXFDescriptor), Descriptor }, /* MPEG 2 Video */
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x48,0x00 }, mxf_read_generic_descriptor, sizeof(MXFDescriptor), Descriptor }, /* Wave */
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x47,0x00 }, mxf_read_generic_descriptor, sizeof(MXFDescriptor), Descriptor }, /* AES3 */

    { { 0x06,0x0e,0x2b,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x5C,0x00 }, mxf_read_generic_descriptor, sizeof(MXFDescriptor), Descriptor }, /* JLB added: ANC Data Descriptor, SMPTE 436 - 7.3 */
    { { 0x06,0x0e,0x2b,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x14,0x00 }, mxf_read_timecode_descriptor_gv, sizeof(MXFTimecodeComponent), TimecodeComponent }, /* JLB added: Read Timecode */
    { { 0x06,0x0e,0x2b,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x23,0x00 }, mxf_read_essence_container_data_gv, sizeof(MXFEssenceContainerData), EssenceContainerData }, /* JLB added: Read Essence Container data*/

    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x3A,0x00 }, mxf_read_track, sizeof(MXFTrack), Track }, /* Static Track */
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x01,0x01,0x01,0x01,0x3B,0x00 }, mxf_read_track, sizeof(MXFTrack), Track }, /* Generic Track */
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x04,0x01,0x02,0x02,0x00,0x00 }, mxf_read_cryptographic_context, sizeof(MXFCryptoContext), CryptoContext },
    { { 0x06,0x0E,0x2B,0x34,0x02,0x53,0x01,0x01,0x0d,0x01,0x02,0x01,0x01,0x10,0x01,0x00 }, mxf_read_index_table_segment, sizeof(MXFIndexTableSegment), IndexTableSegment },
    { { 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 }, NULL, 0, AnyType },
};

static int mxf_read_local_tags(MXFContext *mxf, KLVPacket *klv, MXFMetadataReadFunc *read_child, int ctx_size, enum MXFMetadataSetType type)
{
    AVIOContext *pb = mxf->fc->pb;
    MXFMetadataSet *ctx = ctx_size ? av_mallocz(ctx_size) : mxf;
    uint64_t klv_end = avio_tell(pb) + klv->length;

    if (!ctx)
        return AVERROR(ENOMEM);
    while (avio_tell(pb) + 4 < klv_end && !pb->eof_reached) {
        int ret;
        int tag = avio_rb16(pb);
        int size = avio_rb16(pb); /* KLV specified by 0x53 */
        uint64_t next = avio_tell(pb) + size;
        UID uid = {0};

        av_dlog(mxf->fc, "local tag %#04x size %d\n", tag, size);
        if (!size) { /* ignore empty tag, needed for some files with empty UMID tag */
            av_log(mxf->fc, AV_LOG_ERROR, "local tag %#04x with 0 size\n", tag);
            continue;
        }
        if (tag > 0x7FFF) { /* dynamic tag */
            int i;
            for (i = 0; i < mxf->local_tags_count; i++) {
                int local_tag = AV_RB16(mxf->local_tags+i*18);
                if (local_tag == tag) {
                    memcpy(uid, mxf->local_tags+i*18+2, 16);
                    av_dlog(mxf->fc, "local tag %#04x\n", local_tag);
                    PRINT_KEY(mxf->fc, "uid", uid);
                }
            }
        }
        if (ctx_size && tag == 0x3C0A)
            avio_read(pb, ctx->uid, 16);
        else if ((ret = read_child(ctx, pb, tag, size, uid, -1)) < 0)
            return ret;

        /* Accept the 64k local set limit being exceeded (Avid). Don't accept
         * it extending past the end of the KLV though (zzuf5.mxf). */
        if (avio_tell(pb) > klv_end) {
            av_log(mxf->fc, AV_LOG_ERROR,
                   "local tag %#04x extends past end of local set @ %#"PRIx64"\n",
                   tag, klv->offset);
            return AVERROR_INVALIDDATA;
        } else if (avio_tell(pb) <= next)   /* only seek forward, else this can loop for a long time */
            avio_seek(pb, next, SEEK_SET);
    }
    if (ctx_size) ctx->type = type;

    // JLB added the following block for setting the partition address of the index table
    if (ctx->type == IndexTableSegment) {
    	MXFIndexTableSegment* its = (MXFIndexTableSegment*) ctx;
    	if (its != NULL && mxf->current_partition != NULL){
    		its->this_partition = mxf->current_partition->this_partition;
    	}
    }

    return ctx_size ? mxf_add_metadata_set(mxf, ctx) : 0;
}

/**
 * Seeks to the previous partition, if possible
 * @return <= 0 if we should stop parsing, > 0 if we should keep going
 */
static int mxf_seek_to_previous_partition(MXFContext *mxf)
{
    AVIOContext *pb = mxf->fc->pb;

    if (!mxf->current_partition ||
        mxf->run_in + mxf->current_partition->previous_partition <= mxf->last_forward_tell)
        return 0;   /* we've parsed all partitions */

    /* seek to previous partition */
    avio_seek(pb, mxf->run_in + mxf->current_partition->previous_partition, SEEK_SET);
    mxf->current_partition = NULL;

    av_dlog(mxf->fc, "seeking to previous partition\n");

    return 1;
}

/**
 * Called when essence is encountered
 * @return <= 0 if we should stop parsing, > 0 if we should keep going
 */
static int mxf_parse_handle_essence(MXFContext *mxf)
{
    AVIOContext *pb = mxf->fc->pb;
    int64_t ret;

    if (mxf->parsing_backward) {
        return mxf_seek_to_previous_partition(mxf);
    } else {
        if (!mxf->footer_partition) {
            av_dlog(mxf->fc, "no footer\n");
            return 0;
        }

        av_dlog(mxf->fc, "seeking to footer\n");

        /* remember where we were so we don't end up seeking further back than this */
        mxf->last_forward_tell = avio_tell(pb);

        if (!pb->seekable) {
            av_log(mxf->fc, AV_LOG_INFO, "file is not seekable - not parsing footer\n");
            return -1;
        }

        // JLB Extra check
        if ( mxf->run_in + mxf->footer_partition > avio_size(pb) ) {
        	av_dlog(mxf->fc, "footer outside file, skipping...\n");
        	return 0;
        }

        /* seek to footer partition and parse backward */
        if ((ret = avio_seek(pb, mxf->run_in + mxf->footer_partition, SEEK_SET)) < 0) {
            av_log(mxf->fc, AV_LOG_ERROR, "failed to seek to footer @ 0x%"PRIx64" (%"PRId64") - partial file?\n",
                   mxf->run_in + mxf->footer_partition, ret);
            return ret;
        }

        mxf->current_partition = NULL;
        mxf->parsing_backward = 1;
    }

    return 1;
}

/**
 * Called when the next partition or EOF is encountered
 * @return <= 0 if we should stop parsing, > 0 if we should keep going
 */
static int mxf_parse_handle_partition_or_eof(MXFContext *mxf)
{
    return mxf->parsing_backward ? mxf_seek_to_previous_partition(mxf) : 1;
}

/**
 * Figure out the proper offset and length of the essence container
 * in each partition
 */
static void mxf_compute_essence_containers(MXFContext *mxf)
{
    int x;

    /* everything is already correct */
    if (mxf->op == OPAtom)
        return;

    for (x = 0; x < mxf->partitions_count; x++) {
        MXFPartition *p = &mxf->partitions[x];

        if (!p->body_sid)
            continue;       /* BodySID == 0 -> no essence */

        if (x >= mxf->partitions_count - 1)
            break;          /* last partition - can't compute length (and we don't need to) */

        /* essence container spans to the next partition */
        p->essence_length = mxf->partitions[x+1].this_partition - p->essence_offset;

        if (p->essence_length < 0) {
            /* next ThisPartition < essence_offset */
            p->essence_length = 0;
            av_log(mxf->fc, AV_LOG_ERROR,
                   "partition %i: bad ThisPartition = %"PRIX64"\n",
                   x+1, mxf->partitions[x+1].this_partition);
        }
    }
}

static int64_t round_to_kag(int64_t position, int kag_size)
{
    /* TODO: account for run-in? the spec isn't clear whether KAG should account for it */
    /* NOTE: kag_size may be any integer between 1 - 2^10 */
    int64_t ret = (position / kag_size) * kag_size;
    return ret == position ? ret : ret + kag_size;
}

static inline void compute_partition_essence_offset(AVFormatContext *s,
                                                    MXFContext *mxf,
                                                    KLVPacket *klv)
{
    MXFPartition *cur_part = mxf->current_partition;
    /* for OP1a we compute essence_offset
     * for OPAtom we point essence_offset after the KL
     *     (usually op1a_essence_offset + 20 or 25)
     * TODO: for OP1a we could eliminate this entire if statement, always
     *       stopping parsing at op1a_essence_offset
     *       for OPAtom we still need the actual essence_offset though
     *       (the KL's length can vary)
     */
    int64_t op1a_essence_offset =
        round_to_kag(cur_part->this_partition + cur_part->pack_length,
                     cur_part->kag_size) +
        round_to_kag(cur_part->header_byte_count, cur_part->kag_size) +
        round_to_kag(cur_part->index_byte_count, cur_part->kag_size);

    if (mxf->op == OPAtom) {
        /* point essence_offset to the actual data
         * OPAtom has all the essence in one big KLV
         */
        cur_part->essence_offset = avio_tell(s->pb);
        cur_part->essence_length = klv->length;
    } else {
        /* NOTE: op1a_essence_offset may be less than to klv.offset
         * (C0023S01.mxf)  */
        cur_part->essence_offset = op1a_essence_offset;
    }
}

static int is_pcm(enum AVCodecID codec_id)
{
    /* we only care about "normal" PCM codecs until we get samples */
    return codec_id >= AV_CODEC_ID_PCM_S16LE && codec_id < AV_CODEC_ID_PCM_S24DAUD;
}

/**
 * Deal with the case where for some audio atoms EditUnitByteCount is
 * very small (2, 4..). In those cases we should read more than one
 * sample per call to mxf_read_packet().
 */
static void mxf_handle_small_eubc(AVFormatContext *s)
{
    MXFContext *mxf = s->priv_data;

    /* assuming non-OPAtom == frame wrapped
     * no sane writer would wrap 2 byte PCM packets with 20 byte headers.. */
    if (mxf->op != OPAtom)
        return;

    /* expect PCM with exactly one index table segment and a small (< 32) EUBC */
    if (s->nb_streams != 1                                     ||
        s->streams[0]->codec->codec_type != AVMEDIA_TYPE_AUDIO ||
        !is_pcm(s->streams[0]->codec->codec_id)                ||
        mxf->nb_index_tables != 1                              ||
        mxf->index_tables[0].nb_segments != 1                  ||
        mxf->index_tables[0].segments[0]->edit_unit_byte_count >= 32)
        return;

    /* arbitrarily default to 48 kHz PAL audio frame size */
    /* TODO: We could compute this from the ratio between the audio
     *       and video edit rates for 48 kHz NTSC we could use the
     *       1802-1802-1802-1802-1801 pattern. */
    mxf->edit_units_per_packet = 1920;
}


/*
 * Method to log the date to the av_log
 */
static void av_logDate(void* avcl, int level, int64_t time_in_ms)
{
	int hours, mins, secs, us; // Required for log

	secs = time_in_ms / 1000;
	us = time_in_ms % 1000;
	mins = secs / 60;
	secs %= 60;
	hours = mins / 60;
	mins %= 60;
	av_log(avcl, level, "%02d:%02d:%02d.%02d", hours, mins, secs, us);
}


static int mxf_read_header(AVFormatContext *s)
{
    MXFContext *mxf = s->priv_data;
    KLVPacket klv;
    int64_t essence_offset = 0;
    int ret;
	int64_t readHeaderStartTime = av_gettime();
	int64_t time_in_ms;
    AVTimecode tc;

    mxf->last_forward_tell = INT64_MAX;
    mxf->edit_units_per_packet = 1;
    mxf->first_essence_offset = 0;
    mxf->overall_start_time_ms = 0; // Reset overall start time

    if (!mxf_read_sync(s->pb, mxf_header_partition_pack_key, 14)) {
        av_log(s, AV_LOG_ERROR, "could not find header partition pack key\n");
        return AVERROR_INVALIDDATA;
    }
    avio_seek(s->pb, -14, SEEK_CUR);
	mxf->fc = s;
    mxf->run_in = avio_tell(s->pb);

	// Check if we can find a RIP:
    //ripFound = mxf_locate_RIP_gv(mxf, &klv);
    //avio_seek(s->pb, mxf->run_in, SEEK_SET);


    while (!s->pb->eof_reached) {
        const MXFMetadataReadTableEntry *metadata;

        if (klv_read_packet(&klv, s->pb) < 0) {
            /* EOF - seek to previous partition or stop */
            if(mxf_parse_handle_partition_or_eof(mxf) <= 0)
                break;
            else
                continue;
        }

        PRINT_KEY(s, "read header", klv.key);
        av_dlog(s, "size %"PRIu64" offset %#"PRIx64"\n", klv.length, klv.offset);

        if (IS_KLV_KEY(klv.key, mxf_encrypted_triplet_key) ||
            IS_KLV_KEY(klv.key, mxf_essence_element_key) ||
            IS_KLV_KEY(klv.key, mxf_avid_essence_element_key) ||
            IS_KLV_KEY(klv.key, mxf_system_item_key)) {

            if (!mxf->current_partition) {
                av_log(mxf->fc, AV_LOG_ERROR,
                       "found essence prior to first PartitionPack\n");
                return AVERROR_INVALIDDATA;
            }

            if (!mxf->current_partition->essence_offset) {
                compute_partition_essence_offset(s, mxf, &klv);
            }

            if (!essence_offset)
                essence_offset = klv.offset;

            /* seek to footer, previous partition */
            ret = mxf_parse_handle_essence(mxf);
            if ( ret < 0 )
                break;
            if ( ret > 0 )
            	continue;

            // Else: just try to read the next partition....

        } else if (!memcmp(klv.key, mxf_header_partition_pack_key, 13) && klv.key[13] >= 2 && klv.key[13] <= 4 && mxf->current_partition) {
        	/* next partition pack - keep going, seek to previous partition or stop */
			if(mxf_parse_handle_partition_or_eof(mxf) <= 0)
				break;
			else if (mxf->parsing_backward)
				continue;
			/* we're still parsing forward. proceed to parsing this partition pack */
        }

        for (metadata = mxf_metadata_read_table; metadata->read; metadata++) {
            if (IS_KLV_KEY(klv.key, metadata->key)) {
                int res;
                if (klv.key[5] == 0x53) {
                    res = mxf_read_local_tags(mxf, &klv, metadata->read, metadata->ctx_size, metadata->type);
                } else {
                    uint64_t next = avio_tell(s->pb) + klv.length;
                    res = metadata->read(mxf, s->pb, 0, klv.length, klv.key, klv.offset);

                    /* only seek forward, else this can loop for a long time */
                    if (avio_tell(s->pb) > next) {
                        av_log(s, AV_LOG_ERROR, "read past end of KLV @ %#"PRIx64"\n",
                               klv.offset);
                        return AVERROR_INVALIDDATA;
                    }

                    avio_seek(s->pb, next, SEEK_SET);
                }
                if (res < 0) {
                    av_log(s, AV_LOG_ERROR, "error reading header metadata\n");
                    return res;
                }
                break;
            }
        }
        if (!metadata->read)
            avio_skip(s->pb, klv.length);

        // JLB: quit when too many partitions have been found
        // It is a bit of a guess: after reading the header, the code will start to read the footer, after that, the previous body partition
        // will be read. Normal files only contain those 3 partitions, but some may contain a large number of partitions (even > 200). To prevent
        // stalling too much, quit after three...
        // Read 1 more to include index partitions that might be present
        if (mxf->partitions_count > 4 ||
        	 (essence_offset != 0 && (av_gettime() - readHeaderStartTime) > 200000) ) { // Extra bail out when almost no partitions are found and reading took too long...
        	// When we bail out here, we probably did not yet read the index tables of the last encountered partition (if present),
        	// Make sure the index tables will be read when seeking:
        	if ( mxf->current_partition != NULL )
        		mxf->current_partition->pack_length = -1; // Reset the read flag
        	break;
        }
    }

    // JLB added, read the RIP (when found!)
    // (should be executed before setting the seek pointer to the first essence part since this method will screw-up the seek pointer)
    mxf_read_RIP_gv(mxf);


    /* FIXME avoid seek */
    if (!essence_offset)  {
        av_log(s, AV_LOG_ERROR, "no essence\n");
        return AVERROR_INVALIDDATA;
    }
    // Store the essence offset
    mxf->first_essence_offset = essence_offset;
    // Seek to the correct position
    avio_seek(s->pb, essence_offset, SEEK_SET);

    mxf_compute_essence_containers(mxf);

    /* we need to do this before computing the index tables
     * to be able to fill in zero IndexDurations with st->duration */
    if ((ret = mxf_parse_structural_metadata(mxf)) < 0)
        return ret;


	//
	// JLB: Currently this MXF format is always playing the first package (this codec is still not completely finished...)
	// So get the essence container descriptor of the first package
	for (int i = 0; i < mxf->metadata_sets_count; i++) {
		if (mxf->metadata_sets[i]->type == EssenceContainerData){
			MXFEssenceContainerData* pecd = (MXFEssenceContainerData*) mxf->metadata_sets[i];
			if (memcmp(pecd->linked_package_id, mxf->packages_refs[0], 16)){
				mxf->m_indexSID = pecd->index_sid;
				break;
			}
		}
	}


	//
	// Set the timecode object
	for (int i = 0; i < mxf->metadata_sets_count; i++) {
		if (mxf->metadata_sets[i]->type == TimecodeComponent){
			MXFTimecodeComponent* tccomp = (MXFTimecodeComponent*) mxf->metadata_sets[i];
			if ( tccomp->start_time != 0 ) {
				if (av_timecode_init(&tc, (AVRational){tccomp->rounded_timecode_base, 1}, tccomp->drop_frame == 1 ? AV_TIMECODE_FLAG_DROPFRAME : 0, tccomp->start_time, mxf->fc) == 0) {

					// Store the timecode in the metadata
					mxf_add_timecode_metadata(&mxf->fc->metadata, "timecode", &tc);

					// Set the overall start time
					// Use a trick to get the correct scaling: 1000 + (drop frame) / timebase * 1000
					// Suppose 25 fps: 1000 / 25000 = 1/25
					// Suppose 29.97 fps: 1001 / 30000 = 1/29.97
					mxf->overall_start_time_ms = av_rescale(tccomp->start_time * 1000, 1000 + tccomp->drop_frame, tccomp->rounded_timecode_base * 1000);
				}
			}
		}
	}

	// JLB Generate some logging and set the overall start time to the start time of the first track
	av_log(mxf->fc, AV_LOG_WARNING, "All tracks:\n");
	for (int i = 0; i < mxf->fc->nb_streams; i++) {
		AVStream* st = s->streams[i];

		// Track startTimecode is in frames. Calculate the time:
		time_in_ms = st->start_time * 1000 * av_q2d(st->time_base); // Calculate ms
		av_log(mxf->fc, AV_LOG_WARNING, " * Stream %d[%d]: Start time: %ld frames (", st->id, i, st->start_time);
		av_logDate(0, AV_LOG_WARNING, time_in_ms);
		time_in_ms = st->duration * 1000 * av_q2d(st->time_base); // Calculate ms

		if ( st->duration == AV_NOPTS_VALUE )
			av_log(0, AV_LOG_WARNING, "), duration: n/a (" );
		else
			av_log(0, AV_LOG_WARNING, "), duration: %ld (", st->duration);

		av_logDate(0, AV_LOG_WARNING, time_in_ms);
		av_log(0, AV_LOG_WARNING, ")\n");
	}



	/*
	 	 AFD data can be found on 3 different locations:
			1. In a vbi line -> It will be parsed by the MxfAncDec decoder
			2. In the user data of an mpeg stream. To find this data we decode the packets during the reading of the packets (mxf_read_packet_old)
			   if user data is present another data stream is created, data is passed through to the mpeguserdata decoder.
			3. In a special KLV key in the MXF header. This has already been decoded when we get here. So check if this data has been set,
			   if so, create a new stream (also the mpeguserdata codec) and make sure the AFD data is returned by the mpeguserdata decoder.
		(JLB)
	*/
	for (int i = 0; i < mxf->fc->nb_streams; i++) {
		AVStream* st = s->streams[i];
		if (st->codec->codec_type == AVMEDIA_TYPE_VIDEO) {
			int addExtraStream = 0;
			if (st->codec->dtg_active_format != 0) {
				addExtraStream = s->nb_streams;
				av_log(s, AV_LOG_WARNING, "AFD information found [%d]. Added an extra user data stream [%d] for decoding \n", st->codec->dtg_active_format, addExtraStream);
			}
			if (addExtraStream)	{
				// Create a new data stream
				AVStream* st_user;
				st_user = avformat_new_stream(s, NULL);
				if ( st_user ) {
					st_user->id = addExtraStream; // Add to the end of the streams (add an extra stream)
					st_user->codec->codec_type = AVMEDIA_TYPE_DATA;
					st_user->codec->codec_id = AV_CODEC_ID_MPEG_USERDATA;
					st_user->need_parsing = AVSTREAM_PARSE_FULL;

					// Copy the active format (if present)
					st_user->codec->dtg_active_format = st->codec->dtg_active_format;

					// Since we will copy the video packets to generate the data stream,
					// the pts info should be same as the video stream
					st_user->time_base = st->time_base;
				}
			}
			break;
		}
	}


	// Set the current dts to the start time of the file
	ff_update_cur_dts(s, s->streams[0], 0);

    mxf_handle_small_eubc(s);

    // JLB Added profiling
    av_log(mxf->fc, AV_LOG_WARNING, "Read header in %ld usec (%d partitions)\n", av_gettime() - readHeaderStartTime, mxf->partitions_count);

    return 0;
}


/**
 * Sets mxf->current_edit_unit based on what offset we're currently at.
 * @return next_ofs if OK, <0 on error
 */
static int64_t mxf_set_current_edit_unit(MXFContext *mxf, int64_t current_offset)
{
    int64_t last_ofs = -1, next_ofs = -1;
    MXFIndexTable *t = &mxf->index_tables[0];

    /* this is called from the OP1a demuxing logic, which means there
     * may be no index tables */
    if (mxf->nb_index_tables <= 0)
        return -1;

    /* find mxf->current_edit_unit so that the next edit unit starts ahead
     * of current_offset */
    while (mxf->current_edit_unit >= 0) {
        if (mxf_edit_unit_absolute_offset(mxf, t, mxf->current_edit_unit + 1,
                                          NULL, &next_ofs, 0) < 0)
            return -1;

        if (next_ofs <= last_ofs) {
            /* large next_ofs didn't change or current_edit_unit wrapped
             * around this fixes the infinite loop on zzuf3.mxf */
            av_log(mxf->fc, AV_LOG_ERROR,
                   "next_ofs didn't change. not deriving packet timestamps\n");
            return -1;
        }

        if (next_ofs > current_offset)
            break;

        last_ofs = next_ofs;
        mxf->current_edit_unit++;
    }

    /* not checking mxf->current_edit_unit >= t->nb_ptses here since CBR files
     * may lack IndexEntryArrays */
    if (mxf->current_edit_unit < 0)
        return -1;

    return next_ofs;
}

static int mxf_read_packet_old(AVFormatContext *s, AVPacket *pkt)
{
    KLVPacket klv;
    MXFContext *mxf = s->priv_data;

    while (!s->pb->eof_reached) {
        if (klv_read_packet(&klv, s->pb) < 0)
            return s->pb->eof_reached ? AVERROR_EOF : AVERROR_INVALIDDATA;

        PRINT_KEY(s, "read packet", klv.key);
        av_dlog(s, "size %"PRIu64" offset %#"PRIx64"\n", klv.length, klv.offset);
        if (IS_KLV_KEY(klv.key, mxf_encrypted_triplet_key)) {
            int res = mxf_decrypt_triplet(s, pkt, &klv);
            if (res < 0) {
                av_log(s, AV_LOG_ERROR, "invalid encoded triplet\n");
                return AVERROR_INVALIDDATA;
            }
            return 0;
        }

        if (IS_KLV_KEY(klv.key, mxf_essence_element_key) ||
            IS_KLV_KEY(klv.key, mxf_avid_essence_element_key)) {
            int index = mxf_get_stream_index(s, &klv);
            int64_t next_ofs, next_klv;
            AVStream *st;
            MXFTrack *track;

            if (index < 0) {
                av_log(s, AV_LOG_ERROR, "error getting stream index %d\n", AV_RB32(klv.key+12));
                goto skip;
            }

            st = s->streams[index];
            track = st->priv_data;

            if (s->streams[index]->discard == AVDISCARD_ALL)
                goto skip;

            next_klv = avio_tell(s->pb) + klv.length;
            next_ofs = mxf_set_current_edit_unit(mxf, klv.offset);

            if (next_ofs >= 0 && next_klv > next_ofs) {
                /* if this check is hit then it's possible OPAtom was treated
                 * as OP1a truncate the packet since it's probably very large
                 * (>2 GiB is common) */
                av_log_ask_for_sample(s,
                                      "KLV for edit unit %i extends into next "
                                      "edit unit - OPAtom misinterpreted as "
                                      "OP1a?\n",
                                      mxf->current_edit_unit);
                klv.length = next_ofs - avio_tell(s->pb);
            }

            /* check for 8 channels AES3 element */
            if (klv.key[12] == 0x06 && klv.key[13] == 0x01 && klv.key[14] == 0x10) {
                if (mxf_get_d10_aes3_packet(s->pb, s->streams[index], pkt, klv.length) < 0) {
                    av_log(s, AV_LOG_ERROR, "error reading D-10 aes3 frame\n");
                    return AVERROR_INVALIDDATA;
                }
            } else {
                int ret = av_get_packet(s->pb, pkt, klv.length);
                if (ret < 0)
                    return ret;
            }
            pkt->stream_index = index;
            pkt->pos = klv.offset;

            if (s->streams[index]->codec->codec_type == AVMEDIA_TYPE_VIDEO && next_ofs >= 0) {
                /* mxf->current_edit_unit good - see if we have an
                 * index table to derive timestamps from */
                MXFIndexTable *t = &mxf->index_tables[0];

                if (mxf->nb_index_tables >= 1 &&
                    mxf->current_edit_unit < t->nb_ptses) {
                    pkt->dts = mxf->current_edit_unit + t->first_dts;
                    pkt->pts = t->ptses[mxf->current_edit_unit];
                } else if (track->intra_only) {
                    /* intra-only -> PTS = EditUnit.
                     * let utils.c figure out DTS since it can be
                     * < PTS if low_delay = 0 (Sony IMX30) */
                    pkt->pts = mxf->current_edit_unit;
                }
            }

			/* --------> If this packet contains mpeg video, we want to know if
			 	 	 	 the mpeg video contains user data with AFD data.
			 	 	 	 So parse the data (skip from header to header), when DTG1 user data
			 	 	 	 is found, create a new data stream. Make sure the packets are only scanned
			 	 	 	 the first x packets...
						 (JLB)
			 */
			if (s->streams[index]->codec->codec_id == AV_CODEC_ID_MPEG2VIDEO){
				/*
				 * -------> Only execute this code when we are probing using av_find_stream_info! */
				if (s->streams[index]->info != 0){
					AVStream* st_user;
					int foundUserData = 0;
					int start_code = 0;

					/* -------> If we already have an extra mpeg_user_data data stream, do not check anymore. */
					for(int n = 0; n < s->nb_streams; n++) {
						if (s->streams[n]->codec->codec_id == AV_CODEC_ID_MPEG_USERDATA){
							foundUserData = 1;
							break;
						}
					}

					if (!foundUserData) {
						/* -------> No data mpeg user data stream present, check for user data.
									Loop through all start codes present in this packet. */
						const uint8_t* buf_ptr = pkt->data;
						const uint8_t* buf_end = pkt->data + pkt->size;
						for(;;) {
							/* Find next start code */
							buf_ptr = avpriv_mpv_find_start_code( buf_ptr, buf_end, &start_code );
							if (start_code == 0x000001b2){
								/* Is this the AFD start code?? */
								if (buf_ptr[0] == 'D' && buf_ptr[1] == 'T' && buf_ptr[2] == 'G' && buf_ptr[3] == '1'){
									av_log(s, AV_LOG_ERROR, "AFD User data found in MXF packet (DTG1 user id).\n");

									// Create a new stream to parse the data
									st_user = avformat_new_stream(s, NULL);
									if ( st_user ) {
										st_user->id = s->nb_streams; // Add to the end of the streams (add an extra stream)
										st_user->codec->codec_type = AVMEDIA_TYPE_DATA;
										st_user->codec->codec_id = AV_CODEC_ID_MPEG_USERDATA;
										st_user->need_parsing = AVSTREAM_PARSE_FULL;

										// Since we will copy the video packets to generate the data stream,
										// the pts info should be same as the video stream
										st_user->time_base = s->streams[index]->time_base;
									}
									break;
								}
							}
							if (buf_ptr >= buf_end){
								break; /* -------> Start code not found, simply ignore this packet */
							}
						}
					}
				}
			}

            // JLB Added
            //if (s->streams[index]->first_dts != AV_NOPTS_VALUE)
            //	pkt->pts = ;
            //pkt->dts = st->cur_dts;
            //pkt->dts = AV_NOPTS_VALUE;//s->streams[index]->first_dts++;
            //pkt->pts = AV_NOPTS_VALUE;

            if (st->cur_dts != 0){
            	pkt->dts = st->cur_dts;
            	pkt->pts = AV_NOPTS_VALUE;
            }

            /* seek for truncated packets */
            avio_seek(s->pb, next_klv, SEEK_SET);

            return 0;
        }
        else
        skip:
            avio_skip(s->pb, klv.length);
    }
    return AVERROR_EOF;
}

static int mxf_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    MXFContext *mxf = s->priv_data;
    int ret, size;
    int64_t ret64, pos, next_pos;
    int64_t partition_essence_end = -1;
    AVStream *st;
    int edit_units;
    int64_t	sample_time_request = 0;

    if (mxf->op != OPAtom)
        return mxf_read_packet_old(s, pkt);

    /* OPAtom - clip wrapped demuxing */
    /* NOTE: mxf_read_header() makes sure nb_index_tables > 0 for OPAtom */
    st = s->streams[0];

    if (mxf->current_edit_unit >= st->duration)
        return AVERROR_EOF;

    edit_units = FFMIN(mxf->edit_units_per_packet, st->duration - mxf->current_edit_unit);


    // Find the position
    pos = mxf_find_offset(s, 0, mxf->current_edit_unit, AVSEEK_FLAG_ANY, &sample_time_request, &partition_essence_end, 0);

    /* compute size by finding the next edit unit or the end of the essence container
     * not pretty, but it works */
    next_pos = mxf_find_offset(s, 0, mxf->current_edit_unit + edit_units, AVSEEK_FLAG_ANY, NULL, NULL, 0);
    if (partition_essence_end < 0) {
        av_log(s, AV_LOG_ERROR, "unable to compute the size of the last packet\n");
        return AVERROR_INVALIDDATA;
    }
    if (next_pos > partition_essence_end)
    	next_pos = partition_essence_end;

    if (pos > avio_size( s->pb )) {
    	return AVERROR_EOF;
    }

    if ((size = next_pos - pos) <= 0) {
        av_log(s, AV_LOG_ERROR, "bad size: %i\n", size);
        return AVERROR_INVALIDDATA;
    }

    if ((ret64 = avio_seek(s->pb, pos, SEEK_SET)) < 0) {
        return ret64;
    }

	if ((ret = av_get_packet(s->pb, pkt, size)) != size)
		return ret < 0 ? ret : AVERROR_EOF;

    if (st->codec->codec_type == AVMEDIA_TYPE_VIDEO &&  sample_time_request >= 0) {
        pkt->dts = sample_time_request;
        pkt->pts = sample_time_request;
    }

    pkt->stream_index = 0;
    mxf->current_edit_unit += edit_units;

    return 0;
}


static int mxf_read_close(AVFormatContext *s)
{
    MXFContext *mxf = s->priv_data;
    MXFIndexTableSegment *seg;
    int i;

    av_freep(&mxf->packages_refs);

    for (i = 0; i < s->nb_streams; i++)
        s->streams[i]->priv_data = NULL;

    for (i = 0; i < mxf->metadata_sets_count; i++) {
        switch (mxf->metadata_sets[i]->type) {
        case MultipleDescriptor:
            av_freep(&((MXFDescriptor *)mxf->metadata_sets[i])->sub_descriptors_refs);
            break;
        case Sequence:
            av_freep(&((MXFSequence *)mxf->metadata_sets[i])->structural_components_refs);
            break;
        case SourcePackage:
        case MaterialPackage:
            av_freep(&((MXFPackage *)mxf->metadata_sets[i])->tracks_refs);
            break;
        case IndexTableSegment:
            seg = (MXFIndexTableSegment *)mxf->metadata_sets[i];
            av_freep(&seg->temporal_offset_entries);
            av_freep(&seg->flag_entries);
            av_freep(&seg->stream_offset_entries);
            av_freep(&seg->delta_pos_table_entries );
            av_freep(&seg->key_frame_offset_entries );
            break;
        default:
            break;
        }
        av_freep(&mxf->metadata_sets[i]);
    }
    av_freep(&mxf->partitions);
    av_freep(&mxf->metadata_sets);
    av_freep(&mxf->aesc);
    av_freep(&mxf->local_tags);

    for (i = 0; i < mxf->nb_index_tables; i++) {
        av_freep(&mxf->index_tables[i].segments);
        av_freep(&mxf->index_tables[i].ptses);
        av_freep(&mxf->index_tables[i].fake_index);
    }
    av_freep(&mxf->index_tables);

    return 0;
}

static int mxf_probe(AVProbeData *p) {
    uint8_t *bufp = p->buf;
    uint8_t *end = p->buf + p->buf_size;

    if (p->buf_size < sizeof(mxf_header_partition_pack_key))
        return 0;

    /* Must skip Run-In Sequence and search for MXF header partition pack key SMPTE 377M 5.5 */
    end -= sizeof(mxf_header_partition_pack_key);
    for (; bufp < end; bufp++) {
        if (IS_KLV_KEY(bufp, mxf_header_partition_pack_key))
            return AVPROBE_SCORE_MAX;
    }
    return 0;
}


/*
 * JLB New method to read a complete partition header
 * Used inside our own seek method to read in a partition header. It will also read in any index tables
 * that are defined in the specified partition.
 */
static int mxf_read_partition_header_gv(MXFContext *mxf, AVIOContext *pb, int64_t pos)
{
    MXFPartition 	*partition;
    UID 			op;
    int		 		idx;
    int64_t			ret = 0;
    KLVPacket		klv;
    int64_t			nextklv;

    if ((ret = avio_seek(pb, pos, SEEK_SET)) < 0)
    	return ret;

    if (klv_read_packet(&klv, pb) < 0)
    	return ret;
    nextklv = avio_tell(pb) + klv.length;

    // Is this a partition pack header??
    if (!memcmp(klv.key, mxf_header_partition_pack_key, 13) && klv.key[13] >= 2 && klv.key[13] <= 4) {
    	// Yes, we have an offset, now locate this partition in our partition table
		for( idx = 0; idx < mxf->partitions_count; idx++ ) {
			if (mxf->partitions[idx].this_partition == pos){
				// Yes, found!
				partition = &(mxf->partitions[idx]);

				// Fill in all details
				switch(klv.key[13]) {
				case 2:
					partition->type = Header;
					break;
				case 3:
					partition->type = BodyPartition;
					break;
				case 4:
					partition->type = Footer;
					break;
				default:
					av_log(mxf->fc, AV_LOG_ERROR, "unknown partition type %i\n", klv.key[13]);
					return AVERROR_INVALIDDATA;
				}

				avio_skip(pb, 4);
				partition->pack_length = klv.length;
			    partition->closed = partition->type == Footer || !(klv.key[14] & 1);
			    partition->complete = klv.key[14] > 2;
			    partition->kag_size = avio_rb32(pb);
			    partition->this_partition = avio_rb64(pb);
			    partition->previous_partition = avio_rb64(pb);
			    avio_skip(pb, 8); // footer_partition pointer
			    partition->header_byte_count = avio_rb64(pb);
			    partition->index_byte_count = avio_rb64(pb);
			    partition->index_sid = avio_rb32(pb);
			    partition->body_offset = avio_rb64(pb); // Actually the body offset
			    partition->body_sid = avio_rb32(pb);
			    partition->essence_offset = 0;
			    partition->essence_length = 0;
			    avio_read(pb, op, sizeof(UID));
			    avio_skip(pb, 4); // nb_essence_containers

			    //av_log(mxf->fc, AV_LOG_INFO,
			    //        "PartitionPack: ThisPartition = 0x%"PRIX64", PreviousPartition = 0x%"PRIX64", IndexSID = %i, BodySID = %i\n",
			    //        partition->this_partition,
			    //        partition->previous_partition,
			    //        partition->index_sid, partition->body_sid);

			    if (partition->kag_size <= 0 || partition->kag_size > (1 << 20)) {
					av_log(mxf->fc, AV_LOG_WARNING, "invalid KAGSize %i - guessing ", partition->kag_size);
					if (mxf->op == OPSonyOpt)
						partition->kag_size = 512;
					else
						partition->kag_size = 1;
					av_log(mxf->fc, AV_LOG_WARNING, "%i\n", partition->kag_size);
				}

			    // Set this partition to the current
			    mxf->current_partition = partition;

			    avio_seek(pb, nextklv, SEEK_SET);

			    // Now loop through all other KLV packets in the data
			    while (!pb->eof_reached) {
					if (klv_read_packet(&klv, pb) < 0)
						break;

					if (IS_KLV_KEY(klv.key, mxf_index_table))
					{
						// Yes, Index table found, make sure we return a code that the index table is found,
						// now continue so the index table is processed by mxf_read_local_tags
						// (Note: this method will set the file pointer to the next klv)
						ret = mxf_read_local_tags(mxf, &klv, mxf_read_index_table_segment, sizeof(MXFIndexTableSegment), IndexTableSegment);
					} else if (IS_KLV_KEY(klv.key, mxf_encrypted_triplet_key) ||
						IS_KLV_KEY(klv.key, mxf_essence_element_key) ||
						IS_KLV_KEY(klv.key, mxf_avid_essence_element_key) ||
						IS_KLV_KEY(klv.key, mxf_system_item_key)) {

						// Data found, compute the essence offset
						if (!mxf->current_partition->essence_offset) {
							compute_partition_essence_offset(mxf->fc, mxf, &klv);
						}

						// Compute the real partition essence length
						if (idx < mxf->partitions_count - 1)
							partition->essence_length = mxf->partitions[idx + 1].this_partition - partition->essence_offset;

						break; // QUIT after the essence offset has been found (mostly after the index table)
					}
					else if (!memcmp(klv.key, mxf_header_partition_pack_key, 13))  {
						break; // QUIT Next partition pack (no essence found so far...)
					}
					else {
						// Jump to the next KLV
						avio_skip(pb, klv.length);
					}
				}
			}
		}
    }
    return ret;
}


/*
 * Find the offset in an mxf file
 *
 * This method will do a calculated guess in what partition the data might be found,
 * it will read in the partition header (only if it has not already been read) and load it's index table segments (if present).
 * The next step is to scan all index segments (in memory) to try to locate the frame we are looking for in the
 * tables. If present, it could be that it points to another (mostly previous) partition (see Sony Long-GOP MPEG-2 specifications in MXF Book page 360).
 * By using the body_offset one can determine if a frame address belongs to a certain partition. If the correct partition is found, calculate the start
 * of the frame an set the seek pointer to the that address. If the partition is not found yet, keep looping through all partitions.
 */
static int64_t mxf_find_offset(AVFormatContext *s, int stream_index, int64_t sample_time, int flags, int64_t* psample_time_request, int64_t* pend_of_essencecontainer, int showInfoLogging)
{
    AVStream*		st 	= s->streams[stream_index];
    MXFContext* 	mxf = s->priv_data;
    int64_t 		seekPartition = 0;
    int32_t 		seekStep = 1;
    int32_t 		seekDir = 1;
    int32_t			seekDivider = 0;
    int64_t 		framesPerPartition = 0;
    int64_t 		streamoffset = 0;
    int64_t 		seekPos = 0;
    int64_t 		originalSeekPos = 0;
	int32_t			quitNextLoop = 0;
	int64_t			sample_time_request = sample_time;
	int32_t			foundOffset = 0;

	// Guess the initial partition that should be read
	if (mxf->partitions_count > 1)
		framesPerPartition = st->duration / (mxf->partitions_count - 1); // Skip the header
	else
		framesPerPartition = st->duration;
	seekPartition = (sample_time / framesPerPartition) + 1; // Add header


	//
	// Loop until we find the correct stream offset
	streamoffset = -1;
	seekDir = 0; // Meaning we don't have a clue...
	seekDivider = -3;
	do {
		 // Sanity checks
		if (seekPartition < 0) seekPartition = 0;
		if (seekPartition >= mxf->partitions_count) seekPartition = mxf->partitions_count - 1;

		// Read the header of this partition if not yet in memory
		// (The 'pack_length' is set when the partition header is read)
		if (mxf->partitions[seekPartition].pack_length == -1)
			 mxf_read_partition_header_gv( mxf, s->pb, mxf->partitions[seekPartition].this_partition);


		// Now check if the requested frame number is in the list of segments
		// Note that this list is hold in memory for as long as this clip lives (default libav functionality)!
		// When another partition that contains a IndexTable is read from disk (using mxf_read_partition_header_gv), it's Index table is added to the meta-data-sets.
		// So we keep reading partitions until the correct index table is found
		// (Since the new metadata_sets are added at the back of the list, start at the back... )
		for(int i = mxf->metadata_sets_count - 1; i >= 0; i--){
			if (mxf->metadata_sets[i]->type == IndexTableSegment)
			{
				MXFIndexTableSegment* pts = (MXFIndexTableSegment*) mxf->metadata_sets[i];
				if (pts != NULL && pts->index_sid == mxf->m_indexSID){ // Is this the correct index tables? Since multiple index tables are possible in a single MXF!
					if (pts->index_duration == 0){
						// Probably fixed size units... (I don't know what else to do here...)
						seekPos = sample_time;
						streamoffset = pts->edit_unit_byte_count * seekPos; // Just jump...
						seekPartition = 1; // Data is probably located in the first body partition
						sample_time_request = pts->index_start_position + seekPos;

						// 'Found' data, quit looping
						break;
					}
					else if (sample_time >= pts->index_start_position &&
							 sample_time < pts->index_start_position + pts->index_duration) {

						// Yes, this is the correct segment
						seekPos = sample_time - pts->index_start_position;

						if (pts->nb_index_entries == 2 * pts->index_duration + 1)
							seekPos *= 2;     /* Avid index JLB: Don't know how to test this...*/

						if (pts->edit_unit_byte_count != 0) {
							// By definition: fixes size units
							streamoffset = pts->edit_unit_byte_count * seekPos;
						}
						else {
							if (seekPos < pts->nb_index_entries){
								// Subtract any temporal offsets

								originalSeekPos = seekPos;
								if (flags & AVSEEK_FLAG_ANY) {
									// Normal seek behavior
									seekPos = originalSeekPos + pts->temporal_offset_entries[originalSeekPos];
								}
								else{
									// Seek the matching I frame
									seekPos = originalSeekPos + pts->key_frame_offset_entries[originalSeekPos];
								}

								// Recalculate the new position (required for setting the pts)
								sample_time_request = pts->index_start_position + seekPos;

								// This should be the one!
								streamoffset = pts->stream_offset_entries[seekPos];

								// Get the partition number of this index table
					    		for(int n = 0; n < mxf->partitions_count; n++){
					    			if (mxf->partitions[n].this_partition == pts->this_partition){
					    				seekPartition = n;
					    				break;
					    			}
					    		}
							}
						}

						// Found the data so quit looping
						break;
					}

					// If an index was found in the current seek partition, determine what the next step should be when binary searching
					if (mxf->partitions[seekPartition].this_partition == pts->this_partition) {
						seekDir = (sample_time < pts->index_start_position) ? -1: 1;
					}
				}
			}
		}

		// The previous jump was actually out of bounds
		if (quitNextLoop == 1)
			break;

		if (streamoffset == -1) {
			if (seekDivider == -3){
				// First time, I guess the guess was wrong. Next time, try the next partition (this is done to increase performance)
				seekDivider = -2;
				if (seekDir == 0) // Still no clue
					seekPartition += 1;
				else
					seekPartition += seekDir; // Clue! use the seekDir
			}
			else if (seekDivider == -2){
				// First time, I guess the next wrong as well. Next time, try the previous partition (this is done to increase performance)
				seekDivider = -1;
				if (seekDir == 0)
					seekPartition -= 2; // Still no clue (no index tables found), try a few partions back
				else
					seekPartition += seekDir; // Clue, use the seekDir
			}
			else if (seekDivider == -1){
				// Second time, I guess the next partition was wrong as well. Next time try binary search
				seekDir = 1; // Search forward, hopefully we will encounter some index tables (maybe in the footer??)
				seekDivider = 1; // Normal pattern again
				seekPartition = mxf->partitions_count / 2;
			}
			else {
				// Not found, try another partition (divide and conquer algo:)
				seekDivider *= 2;
				seekStep = (int32_t) ceil((double) mxf->partitions_count / seekDivider);
				if (seekStep <= 1)	{
					seekStep = 1; // Limit to steps of 1
					quitNextLoop = 1; // BUT: quit the next time after we get at steps of 1!
				}
				seekPartition = seekPartition + seekDir * seekStep;
			}
		}
	} while (streamoffset == -1);


	if (streamoffset != -1)
	{
		// Now we have the correct stream offset, but we don't know yet in what partition the data is stored...

		// Check if the requested stream position is in this partition, if not, check the previous partition:
		// SONY Long-GOP MPEG-2 Mapping states that the index is stored in the NEXT partition, so when we have the index look in the previous partition to find the data
		// If still not found, use the same trick as before, binary search
		// Start with current partition (the one with the requested index table)
		seekDivider = -1; // Special case: make sure we start by looking at the previous item (in the first step), then binary search
		seekDir = 1;
		quitNextLoop = 0;

		do
		{
			if (seekPartition < 0) seekPartition = 0;
			if (seekPartition >= mxf->partitions_count) seekPartition = mxf->partitions_count - 1;

			// Now we have the correct partition, or the partition cannot be found and we use the first (body) partition

			// Load the partition (if not already loaded)
			if (mxf->partitions[seekPartition].pack_length == -1)
				mxf_read_partition_header_gv( mxf, s->pb, mxf->partitions[seekPartition].this_partition);

			// Is the offset inside this partition??
			if (streamoffset >= mxf->partitions[seekPartition].body_offset && mxf->partitions[seekPartition].type != Footer){

				// If the essence length is not known this is probably because there is no RIP
				// and no next partition header present, so assume the length is equal to the length
				// of the whole file minus this offset
				if (mxf->partitions[seekPartition].essence_length <= 0) {
					mxf->partitions[seekPartition].essence_length = avio_size( s->pb ) - mxf->partitions[seekPartition].body_offset;
				}

				// Stop searching if the requested offset is in this partition:
				if (streamoffset < mxf->partitions[seekPartition].body_offset + mxf->partitions[seekPartition].essence_length) {

					// Now we have the correct partition, or the partition cannot be found and we use the first (body) partition
					foundOffset = 1;
					break;
				}
			}


			// Determine the jump direction
			if ( mxf->partitions[seekPartition].essence_length > 0 ) // If there is some essence in this partition, the body_offset should contain a valid value
				seekDir = (streamoffset < mxf->partitions[seekPartition].body_offset) ? -1 : 1;
			else
				seekDir = seekDir * -1; // Unknown position, just jump around. Start with SONY spec.: that is start by checking the previous partition

			if (quitNextLoop)
				break;

			// Calculate the next seek partition
			if (seekDivider < 0 ){
				seekPartition = seekPartition + seekDir;
				seekDivider = 1; // Normal pattern again
			}
			else {
				seekDivider *= 2;
				seekStep = (int32_t) ceil((double) mxf->partitions_count / seekDivider);
				if (seekStep < 1)	{
					seekStep = 1; // Limit to steps of 1
					quitNextLoop = 1; // BUT: quit the next time after we get at steps of 1!
				}
				seekPartition = seekPartition + seekDir * seekStep;
			}
		} while (1);


	}

	if ( foundOffset ) {

		// When we have encountered a valid offset...
		if ( seekPartition >= 0 && seekPartition < mxf->partitions_count ) {

			if ( showInfoLogging != 0 ) {
				av_log(s, AV_LOG_INFO, "   - required stream offset %ld found in partition: %ld\n", streamoffset, seekPartition);
			}

			// Subtract the body-offset:
			streamoffset -= mxf->partitions[seekPartition].body_offset;

			// And add the real-absolute essence offset
			streamoffset += mxf->partitions[seekPartition].essence_offset;

			// Return the end of the partition offset if required
			if (pend_of_essencecontainer != NULL)
				*pend_of_essencecontainer = mxf->partitions[seekPartition].essence_offset +  mxf->partitions[seekPartition].essence_length;

			// Some logging
			if ( showInfoLogging != 0 ) {
				av_log(s, AV_LOG_INFO, "   - current body offset: %ld, file offset set to: %ld\n", mxf->partitions[seekPartition].body_offset, streamoffset);
			}
		}
	}
	else {
		av_log(s, AV_LOG_ERROR, "   - Unable to find the requested search location! Seeking using the bitrate (%d)\n", s->bit_rate);

		if ( !st->codec->bit_rate ) {
			av_log(s, AV_LOG_ERROR, "   - Unknown bitrate, cannot guess frame offset, defaulting to start of file!\n");
			streamoffset = 0;
			sample_time_request = 0; // Starttime will be added later
		}
		else {
			streamoffset = av_rescale( st->codec->bit_rate * sample_time, st->time_base.num, st->time_base.den );
		}

        // Add the first essence offset (when found)
        streamoffset = mxf->first_essence_offset + streamoffset >> 3;
	}

	// Return the sample time request
	if (psample_time_request != NULL)
	{
		 // And add the start time (if present)
		if (st->start_time != AV_NOPTS_VALUE)
			sample_time_request += st->start_time;
		*psample_time_request = sample_time_request;
	}


	return streamoffset;
}


/*
 * JLB New seek method
 * This method will do a calculated guess in what partition the data might be found,
 * it will read in the partition header (only if it has not already been read) and load it's index table segments (if present).
 * The next step is to scan all index segments (in memory) to try to find if the frame we are looking for is present in the
 * tables. If present, it could be that it points to another (mostly previous) partition (see Sony Long-GOP MPEG-2 specifications in MXF Book page 360).
 * By using the body_offset one can determine if a frame address belongs to a certain partition. If the correct partition is found, calculate the start
 * of the frame an set the seek pointer to the that address. If the partition is not found yet, keep looping through all paritions.
 */
static int mxf_read_seek(AVFormatContext *s, int stream_index, int64_t sample_time, int flags)
{
    AVStream*		st 	= s->streams[stream_index];
    AVStream*		stTemp = NULL;
    MXFContext* 	mxf = s->priv_data;
    int64_t 		streamoffset = 0;
    int64_t 		ret = 0;
	int64_t 		seekStartTime = av_gettime();
	int64_t			sample_time_request = 0;
	int64_t			start_time = 0;

	// Subtract the stream start offset when absolute time is requested
	sample_time_request = sample_time;
	if (flags & AVSEEK_FLAG_ABSOLUTE_TIME)
	{
		start_time = mxf->overall_start_time_ms; // in msec
		sample_time -= av_rescale(start_time, st->time_base.den, st->time_base.num) / 1000;
		av_log(s, AV_LOG_INFO, "MxfSeek, sample_time_request %ld corrected to sample_time %ld\n", sample_time_request, sample_time );
	}
	else
		av_log(s, AV_LOG_INFO, "MxfSeek, sample_time_request %ld\n", sample_time_request);

	// Let all streams use the same start_time
	for(int i = 0; i < mxf->fc->nb_streams; i++){
		stTemp = mxf->fc->streams[i];
		if ( stTemp != NULL)
			stTemp->start_time = av_rescale(start_time, stTemp->time_base.den, stTemp->time_base.num) / 1000;
	}


	// clamp above zero, else ff_index_search_timestamp() returns negative
	// this also means we allow seeking before the start
	sample_time = FFMAX(sample_time, 0);
    if (sample_time < 0)
    	sample_time = 0;
	av_log(s, AV_LOG_INFO, "SeekGV started, seeking frame: %ld\n", sample_time);

	//
	// Find the correct offset in the data
	// (The sample_time_request is set to the exact clip value)
	streamoffset = mxf_find_offset(s, stream_index, sample_time, flags, &sample_time_request, NULL, 1);

	// Perform extra sanity checks
	if ( streamoffset < 0 || streamoffset > avio_size( s->pb ) ) {
		av_log(s, AV_LOG_ERROR, "Found stream offset falls outside the file! (%ld). Starting at the start\n", streamoffset);

		// RESET to the start of the first essence
		streamoffset = mxf->first_essence_offset;
		sample_time_request = st->start_time;
	}

	// Set the reference (not required, but might come in handy later on...)
	st->reference_dts = sample_time_request;

	// Store the new edit unit (currently working on THIS unit)
	mxf->current_edit_unit = sample_time_request;

	// Seek to the right posit ion
	if ((ret = avio_seek(s->pb, streamoffset, SEEK_SET)) < 0)
	  	return ret;

	// Set the DTS
	ff_update_cur_dts(s, st, sample_time_request);


    // Added profiling
    av_log(mxf->fc, AV_LOG_WARNING, "[Total seek time: %ld usec]\n", av_gettime() - seekStartTime);
    return 0;
}


AVInputFormat ff_mxf_demuxer = {
    .name           = "mxf",
    .long_name      = NULL_IF_CONFIG_SMALL("MXF (Material eXchange Format)"),
    .priv_data_size = sizeof(MXFContext),
    .read_probe     = mxf_probe,
    .read_header    = mxf_read_header,
    .read_packet    = mxf_read_packet,
    .read_close     = mxf_read_close,
    .read_seek      = mxf_read_seek,
};
