#!/bin/bash

# Use case insensitive matching within 'PLATFORM' and 'GLOBAL' sections.
shopt -s nocasematch

# ----------------------------------------------------------------------------------------
#                                         PLATFORM
# ----------------------------------------------------------------------------------------

# Detect the platform relying on 'uname' output. On Windows a 'MINGW32*' output signifies MinGW|MSYS presence with
# 'nexFFMpeg' and 'regexp' being built with it. An output of 'Windows32|Windows64' implies MinGW is missing, precompiled
# 'nexFFMpeg' is used and 'regexp' is expected pre-installed at MSVC standard location.
case $(uname | tr -d \\r) in
	linux) PLATFORM="linux" ;;
	darwin) PLATFORM="osx" ;;
	windows32) MINGW="no"; PLATFORM="Win32" ;;
	windows64) MINGW="no"; PLATFORM="x64" ;;
	# For MinGW, assume 32 bit toolchain able to produce both 32 and 64 bit binaries. Build ones matching host arch.
	mingw32*)	MINGW="yes";
				# MSYS has GNU AWK. 'Platform ID' of 0x20 means 32 bit, surprisingly for 64 bit it's '0x10'.
				# TODO: check for Intel64 or AMD64 by the occasion and use correspondent '-favor:' with MSVC?
				[ $(reg query "HKLM\Hardware\Description\System\CentralProcessor\0" -v "Platform ID" | awk 'NR==3 {print $4}') = "0x20" ] && { PLATFORM="Win32"; MINGWTarget="-m32"; } || { PLATFORM="x64"; MINGWTarget="-m64"; }
				# MinGW toolchain prefix (if crosscompiler include the dash as well, empty otherwise).
				MINGWPrefix="x86_64-w64-mingw32-"
				;;
	*)
		echo "Couldn't reliably detect a build platform, bailing out!"
		exit
		;;
esac


# An array of system names. Systems are processed in this order. Linux is a production platform.
SystemNameList=( 'nexlibav' );

# ----------------------------------------------------------------------------------------
#                                         GLOBAL
# ----------------------------------------------------------------------------------------

# The name of the main system in the group folder. Must also be part of SystemNameList.
SystemMain='nexlibav'

# The global qmake debug/release mode flag. By default we want release level.
[ "x${QmakeReleaseDebug}" = "x" ] && QmakeReleaseDebug=release


# Produce release/debug selector for MSVC, must be 'Release' or 'Debug', exactly.
case ${QmakeReleaseDebug} in
	release) MSVCBuildType="Release" ;;
	debug) MSVCBuildType="Debug" ;;
	*) MSVCBuildType="Release" ;;
esac

# Global make command.
if [ -z "$SystemMakeCmd" ]; then
	if [ $PLATFORM = "Win32" -o  $PLATFORM = "x64" ]; then
		# 'nmake' doesn't spawn multiple jobs, MSVC compiler does, with /MP switch 
		SystemMakeCmd='nmake'
	else
		SystemMakeCmd='make -j 8'
	fi
fi


# Revert to case sensitive matching as a precaution (i.e. restore default setting).
shopt -u nocasematch

# ----------------------------------------------------------------------------------------
#                                           SVN                                          
# ----------------------------------------------------------------------------------------

# The SVN properties of each of the systems in SystemNameList.. 
# These are arrays with following components:
#							 0: svn server	  1: svn folder			2: svn version

SystemSvn_nexlibav=(         svn://svnsrv	  ext/nexlibav		    trunk)

# ----------------------------------------------------------------------------------------
#                                          CLEAN
# ----------------------------------------------------------------------------------------


# The clean command for each of the systems in SystemNameList.
# The command is invoked from the group folder's root directory.

SystemClean_nexlibav='(  cd nexlibav;	 ${SystemMakeCmd} distclean; )'

# ----------------------------------------------------------------------------------------
#                                        CONFIGURE
# ----------------------------------------------------------------------------------------

# The configure command for each of the systems in SystemNameList.
# The command is invoked from the group folder's root directory.

if [ ${QmakeReleaseDebug} = "debug" ]; then
	SystemConfigure_nexlibav='(  cd nexlibav;  qmake CONFIG+=${QmakeReleaseDebug} make-programs.pro; ${SystemMakeCmd} distclean; ./configure --disable-optimizations --disable-symver --arch=x86_64 --disable-bzlib --disable-network --disable-avplay --enable-static --enable-pthreads; echo 'BUILD_DOC=' >> config.mak )'
else
	SystemConfigure_nexlibav='(  cd nexlibav;  qmake CONFIG+=${QmakeReleaseDebug} make-programs.pro; ${SystemMakeCmd} distclean; ./configure --disable-symver --arch=x86_64 --disable-bzlib --disable-network --disable-avplay --enable-static --enable-pthreads; echo 'BUILD_DOC=' >> config.mak )'
fi
# ----------------------------------------------------------------------------------------
#                                          MAKE
# ----------------------------------------------------------------------------------------

SystemMake_nexlibav='(	 cd nexlibav;  ${SystemMakeCmd} || exit 1; ${SystemMakeCmd} -f Makefile-programs || exit 1; )'

# ----------------------------------------------------------------------------------------
#                                          ALIAS
# ----------------------------------------------------------------------------------------

# Translation table from short system name to formal system name.
# And to keep things simple in the scripts, also from formal -> formal name.

SystemAlias_libav=nexlibav
SystemAlias_nexlibav=nexlibav

# ----------------------------------------------------------------------------------------
#                             ACCESSOR FUNCTIONS.  DO NOT CHANGE!     
# ----------------------------------------------------------------------------------------


SysName()
{
	echo $1
}
	
FormalSysName()
{
	local formalName="SystemAlias_$1"
	eval echo \${$formalName}
}
	
SysNameUpper()
{
	echo $1 | tr [:lower:] [:upper:]
}

SysSvnServer()
{
	local formalName="SystemAlias_$1"
	local ident="SystemSvn_${!formalName}"
	eval echo \${$ident[0]}
}
	
SysSvnFolder()
{
	local formalName="SystemAlias_$1"
	local ident="SystemSvn_${!formalName}"
	eval echo \${$ident[1]}
}
	
SysSvnVersion()
{
	local formalName="SystemAlias_$1"
	local ident="SystemSvn_${!formalName}"
	eval echo \${$ident[2]}
}
	
SysClean()
{
	local formalName="SystemAlias_$1"
	local ident="SystemClean_${!formalName}"
	echo ${!ident}
}

SysConfigure()
{
	local formalName="SystemAlias_$1"
	local ident="SystemConfigure_${!formalName}"
	echo ${!ident}
}

SysMake()
{
	local formalName="SystemAlias_$1"
	local ident="SystemMake_${!formalName}"
	echo ${!ident}
}

Help()
{
    if [ $# -gt 0 ]; then
       if [ "$1" = "-h" ]; then
    	   echo " "
		   echo "usage: ${ScriptName} [qmakemode] [sysname...]"
		   echo "       qmakemode   one of [debug|release], and defaults to release."
		   echo "       sysname...  list of system names, in either full (ostoolkit) or short (otk) form,"
		   echo "                   and defaults to all systems in the group."
		   echo " "
		   exit 0
	   fi
	fi
}



